/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Dialog "Options"
********************************************************************/
#include "pch.h"
#include "resource.h"

static BOOL CALLBACK DgOptionsProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void Opt_Init (HWND hDlg);
static void Opt_DialogToVars (HWND hDlg);

extern HINSTANCE  g_hInst;   
extern COptions   g_Options;

// Write variable to dialog's edit/text window
bool _cdecl okVarToDialog (HWND hwDlg, INT_PTR Id, LPCWSTR szFmt, ...);

// Read variable from dialog's edit window
bool okDialogToVar (HWND hwDlg, INT_PTR Id, INT_PTR ns, LPCWSTR szFmt, LPVOID pVar);

//-----------------------------------------------
BOOL Dg_Options (HWND hwParent, BOOL bSave)
{
  COptions opt2 = g_Options;
  INT_PTR ret = ::DialogBox( g_hInst, MAKEINTRESOURCEW(IDD_OPTIONS), hwParent, (DLGPROC)DgOptionsProc );
  if (ret == IDOK){
    if (bSave){
      g_Options.Save();
    }
    return true;
  }else{
    // restore
    g_Options = opt2;
  }
  return false;
}


//-----------------------------------------------
static BOOL CALLBACK DgOptionsProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  RECT  rect;
  POINT p;

  switch (uMsg) {
    case WM_INITDIALOG:
      ::GetWindowRect( hDlg, &rect );
      p.x = (::GetSystemMetrics(SM_CXSCREEN)-(rect.right-rect.left))>>1;
      p.y = (::GetSystemMetrics(SM_CYSCREEN)-(rect.bottom-rect.top))>>1;
      ::SetWindowPos( hDlg, HWND_TOP, p.x, p.y, 0,0, SWP_NOSIZE );
      Opt_Init( hDlg );
      return TRUE;

    case WM_CLOSE:
      ::EndDialog( hDlg, 0 );
      return TRUE;

    case WM_COMMAND:
      switch (LOWORD(wParam)) {

        case IDOK:
          Opt_DialogToVars( hDlg );
          ::EndDialog( hDlg, IDOK );
          return TRUE;

        case IDCANCEL:
          ::EndDialog( hDlg, 0 );
          return TRUE;
      }
      break;
  }
  return FALSE;
}


//-----------------------------------------------
void Opt_Init (HWND hDlg)
{
  int i=0;
  HWND hCBox = ::GetDlgItem( hDlg, ID_CB_FILEVER );
  ::SendMessage( hCBox, CB_RESETCONTENT, 0, 0 );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R10" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R11-12" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R13" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R14" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R15 (2000)" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R18 (2004)" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R21 (2007)" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R24 (2010)" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R27 (2013)" );
  ::SendMessage( hCBox, CB_ADDSTRING, 0, (LPARAM)L"AutoCAD R32 (2018)" );
  switch( g_Options.m_Exp_FileVer ){
    case 13: i=0; break;  // Release 10 
    case 16: i=1; break;  // Release 11-12 
    case 19: i=2; break;  // Release 13 
    case 21: i=3; break;  // Release 14 
    case 23: i=4; break;  // R15 (2000) 
    case 25: i=5; break;  // R18 (2004) 
    case 27: i=6; break;  // R21 (2007)
    case 29: i=7; break;  // R24 (2010)
    case 31: i=8; break;  // R27 (2013)
    case 33: i=9; break;  // R32 (2018)
  }
  ::SendMessage( hCBox, CB_SETCURSEL, i, 0 );
  okVarToDialog( hDlg, ID_EDIT_DXFPREC, L"%d", g_Options.m_Exp_DxfPrec );
  ::CheckDlgButton( hDlg, ID_CHK_DIMBLOCK, (g_Options.m_Imp_bDimAsBlock)? BST_CHECKED : BST_UNCHECKED );
  ::CheckDlgButton( hDlg, ID_CHK_HATCHBLOCK, (g_Options.GetImp_HatchAsBlock())? BST_CHECKED : BST_UNCHECKED );
}


//-----------------------------------------------
void Opt_DialogToVars (HWND hDlg)
{
  LRESULT ival;
  HWND hCBox = ::GetDlgItem( hDlg, ID_CB_FILEVER );

  ival = ::SendMessage( hCBox, CB_GETCURSEL, 0, 0 );
  switch( ival ){
    case 0: g_Options.m_Exp_FileVer = 13; break;  // Release 10 
    case 1: g_Options.m_Exp_FileVer = 16; break;  // Release 11-12 
    case 2: g_Options.m_Exp_FileVer = 19; break;  // Release 13 
    case 3: g_Options.m_Exp_FileVer = 21; break;  // Release 14 
    case 4: g_Options.m_Exp_FileVer = 23; break;  // R15 (2000) 
    case 5: g_Options.m_Exp_FileVer = 25; break;  // R18 (2004) 
    case 6: g_Options.m_Exp_FileVer = 27; break;  // R21 (2007)
    case 7: g_Options.m_Exp_FileVer = 29; break;  // R24 (2010)
    case 8: g_Options.m_Exp_FileVer = 31; break;  // R27 (2013)
    case 9: g_Options.m_Exp_FileVer = 33; break;  // R32 (2018)
  }
  okDialogToVar( hDlg, ID_EDIT_DXFPREC, 5, L"%d", &ival );
  if (0 <= ival && ival <= 16){
    g_Options.m_Exp_DxfPrec = (int)ival;
  }
  ival = ::IsDlgButtonChecked( hDlg, ID_CHK_DIMBLOCK );
  g_Options.m_Imp_bDimAsBlock = (ival == BST_CHECKED)? true : false;
  ival = ::IsDlgButtonChecked( hDlg, ID_CHK_HATCHBLOCK );
  g_Options.SetImp_HatchAsBlock( (ival == BST_CHECKED)? true : false );
}


#define LENDLGBUF 32032   // buffer size for dialog variable

//-----------------------------------------------
// Write variable to dialog's edit/text window
//-----------------------------------------------
bool _cdecl okVarToDialog (HWND dgItem, INT_PTR Id, LPCWSTR szFmt, ...)
{
  bool res = false;
  wchar_t* szBuf = new wchar_t[LENDLGBUF];
  if (szBuf){
    memset( szBuf, 0, sizeof(wchar_t)*LENDLGBUF );
    va_list  argptr;	 // pointer to list of arguments
    va_start( argptr, szFmt );
    if (vswprintf( szBuf, szFmt, argptr ) > 0){
      res = true;
    }
    ::SetDlgItemText( dgItem, (int)Id, szBuf );
    va_end( argptr );
    delete[] szBuf;
  }
  return res;
}

//-----------------------------------------------
// Read variable from dialog's edit window
//-----------------------------------------------
bool okDialogToVar (HWND hDlg, INT_PTR Id, INT_PTR ns, LPCWSTR szFmt, LPVOID pVar)
{
  bool     res = false;
  INT_PTR  n;
  wchar_t* szBuf = new wchar_t[ns+8];
  if (szBuf){
    memset( szBuf, 0, sizeof(wchar_t)*ns );
    ::GetDlgItemText( hDlg, (int)Id, szBuf, (int)(ns+1) );
    if (szFmt[1] == L's'){
      wcscpy( (wchar_t*)pVar, szBuf );
      res = true;
    }else{
      n = swscanf( szBuf, szFmt, pVar );
      if (n > 0){
        res = true;
      }else{
        n = swscanf( L"0", szFmt, pVar );
        res = false;
      }
    }
    delete[] szBuf;
  }
  return res;
}

