/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Simple functions
*********************************************************************/
#include "pch.h"

LPCWSTR GetFileDir  (LPCWSTR szFileName, bool bIncludeBackslash=true);
LPCWSTR GetFileName (LPCWSTR szFileName, bool bWithExtention=true);

//-----------------------------------------------
// Seach for file
//  szFileName - original filename (can be with path)
//  szFoundName - found filename or NULL
//   0 - found by original filename
//   1 - found in current directory
//   2 - found in application directory
//   3 - found in Windows  directory
//   4 - found in Windows System directory
//   5 - found in the 'szSearchDir'
//   6 - found in application SUBdirectory
//   7 - found in directory from PATH variable
//  -1 - file NOT FOUND
//-----------------------------------------------
INT_PTR okSearchFile (LPCWSTR szFileName, wchar_t* szFoundName, LPCWSTR szSearchDir)
{
  wchar_t          szBuf[320], szBuf2[256], szFn[256], szDir[256], szAppDir[256];
  wchar_t          szEnv[512];
  wchar_t*         szE;
  wchar_t*         psz;
  INT_PTR          i, n, len;
  WIN32_FIND_DATAW fdata, fdata2;
  HANDLE           hff, hff2;

  if (szFileName == NULL){
    return -1;
  }
  if (szFileName[0] == 0){
    return -1;
  }
  if (szFoundName){
    szFoundName[0] = 0;
  }
  hff = ::FindFirstFile( szFileName, &fdata );
  if (hff != INVALID_HANDLE_VALUE){
    ::FindClose( hff );
    if (szFoundName){
      wcscpy( szDir, GetFileDir( szFileName, false ) );
      if (szDir[0]){
        swprintf( szFoundName, L"%s\\%s", szDir, fdata.cFileName );
      }else{
        wcscpy( szFoundName, fdata.cFileName );
      }
    }
    return 0;  // file is found by original filename
  }

  wcscpy( szFn, GetFileName(szFileName) );   // Get filename witout path
  szAppDir[0] = 0;
  if (szSearchDir && szSearchDir[0]){
    n = 5;
  }else{
    n = 4;
  }
  for (i=1; i<=n; ++i){
    switch(i){
      case 1:  // Search in the current directory
        ::GetCurrentDirectory( 256, szDir );
        break;
      case 2:  // Search in application directory
        ::GetModuleFileName( NULL, szAppDir, 255 );
        psz = wcsrchr( szAppDir, L'\\' );
        if (psz){
          psz[0] = 0;
        }
        wcscpy( szDir, szAppDir );
        break;
      case 3:  // Search in Windows  directory
        len = ::GetWindowsDirectory( szDir, 256 );
        break;
      case 4:  // Search in Windows System directory
        ::GetSystemDirectory( szDir, 256 );
        break;
      case 5:
        wcscpy( szDir, szSearchDir );
        break;
    }
    swprintf( szBuf, L"%s\\%s", szDir, szFn );
    hff = ::FindFirstFile( szBuf, &fdata );
    if (hff != INVALID_HANDLE_VALUE){
      ::FindClose( hff );
      if (szFoundName){
        wcscpy( szFoundName, szBuf );
      }
      return i;    // file is found
    }
  }

  // Search in application subdirectory
  swprintf( szBuf, L"%s\\*.*", szAppDir );
  hff = ::FindFirstFile( szBuf, &fdata );
  if (hff != INVALID_HANDLE_VALUE){
    for(;;){ //while(1){
      if (fdata.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY){
        swprintf( szBuf2, L"%s\\%s\\%s", szAppDir, fdata.cFileName, szFn );
        hff2 = ::FindFirstFile( szBuf2, &fdata2 );
        if (hff2 != INVALID_HANDLE_VALUE){
          ::FindClose( hff2 );
          ::FindClose( hff );
          if (szFoundName){
            wcscpy( szFoundName, szBuf2 );
          }
          return 6;    // file is found in application subdirectory
        }
      }
      if (::FindNextFile( hff,&fdata )==FALSE){
        break;
      }
    }
    ::FindClose( hff );
  }

  // Search by PATH variables
  szE = _wgetenv( L"PATH" );
  if (szE){
    wcsncpy( szEnv, szE, 510 );
    szEnv[510] = 0;
    psz = wcstok( szEnv, L";" );
    while (psz) {
      swprintf( szBuf, L"%s\\%s", psz, szFn );
      hff = ::FindFirstFile( szBuf, &fdata );
      if (hff != INVALID_HANDLE_VALUE){
        ::FindClose( hff );
        if (szFoundName){
          wcscpy( szFoundName, szBuf );
        }
        return 7;  // file is found in directory from PATH variable
      }
      psz = wcstok( NULL, L";" );
    }
  }

  if (szFoundName){
    szFoundName[0] = 0;
  }
  return -1;   // file not found
}

//-----------------------------------------------
// Get file directory from full pathname
// The output directory includes leading drive symbol
// and trailing backslash (\)
// If bIncludeSlash is FALSE then trailing backslash will not be included
//-----------------------------------------------
LPCWSTR GetFileDir (LPCWSTR szFileName, bool bIncludeBackslash)
{
  static WCHAR szBuf[256];
  WCHAR*  ch = L"\\//";
  size_t  i, len;

  wcscpy( szBuf, szFileName );
  len = wcslen( szBuf );
  for (i=len-1; i>0; --i){
    if (szBuf[i]==ch[0] || szBuf[i]==ch[1]){
      if (bIncludeBackslash){
        szBuf[i+1] = 0;
      }else{
        szBuf[i] = 0;
      }
      return szBuf;
    }
  }
  szBuf[0] = 0;
  return szBuf;
}

//-----------------------------------------------
// Get file name with/without extention from full pathname
//-----------------------------------------------
LPCWSTR GetFileName (LPCWSTR szFileName, bool bWithExtention)
{
  static WCHAR szBuf[300];
  WCHAR*  ch = L"\\.";
  WCHAR*  pDot = NULL;
  size_t  i, j, len;

  if (wcschr( szFileName, ch[0] )==NULL){
    // divider not found 
    wcscpy( szBuf, szFileName );
    if (bWithExtention == false){
      // delete extention
      j = 0;
      len = wcslen( szBuf );
      for (i=len-1; i>=0 && j<5; i--){
        if (szBuf[i] == ch[1]){
          szBuf[i] = 0;
          break;
        }
        ++j;
      }
    }
    return szBuf;
  }
  len = wcslen( szFileName );
  for (i=len-1; i>0; --i){
    if (szFileName[i] == ch[0]){
      j=0;
      ++i;
      while( szFileName[i] != 0 ){
        if (szFileName[i]==ch[1]){
          pDot = szBuf + j;
        }
        szBuf[j] = szFileName[i];
        ++j; ++i;
      }
      szBuf[j] = 0;
      if (bWithExtention == false){
        // delete extention
        if (pDot){
          pDot[0] = 0;
        }
      }
      return szBuf;
    }
  }
  szBuf[0]=0;
  return szBuf;
}


//-----------------------------------------------
// Get file name with/without extention from full pathname
//-----------------------------------------------
LPCWSTR okGetFileName (LPCWSTR szFileName, bool bWithExtention)
{
  static WCHAR szBuf[300];
  WCHAR*   ch = L"/\\.";
  WCHAR*   pDot = NULL;
  INT_PTR  i, j, len;

  if (wcschr( szFileName, ch[0] )==NULL){
    if (wcschr( szFileName, ch[1] )==NULL){
      // divider not found 
      wcscpy( szBuf, szFileName );
      if (bWithExtention == false){
        // delete extention
        j = 0;
        len = wcslen( szBuf );
        for (i=len-1; i>=0 && j<5; --i){
          if (szBuf[i] == ch[2]){
            szBuf[i] = 0;
            break;
          }
          ++j;
        }
      }
      return szBuf;
    }
  }
  len = wcslen( szFileName );
  for (i=len-1; i>=0; --i){
    if ((szFileName[i] == ch[0]) || (szFileName[i] == ch[1])){
      j=0;
      ++i;
      while( szFileName[i] != 0 ){
        if (szFileName[i]==ch[2]){
          pDot = szBuf + j;
        }
        szBuf[j] = szFileName[i];
        ++j;
        ++i;
      }
      szBuf[j] = 0;
      if (bWithExtention == false){
        // delete extention
        if (pDot){
          pDot[0] = 0;
        }
      }
      return szBuf;
    }
  }
  szBuf[0]=0;
  return szBuf;
}

//-----------------------------------------------
// Get file extention from full pathname
//-----------------------------------------------
LPCWSTR okGetFileExt (LPCWSTR szFileName, WCHAR* szExt)
{
  static WCHAR szBuf[8];
  WCHAR*  ch = L".";
  INT_PTR i, j, len;
  
  szBuf[0] = 0;
  len = (int)(wcslen( szFileName ));
  if (len > 2){
    for (i=len-1; i>=0; --i){
      if (szFileName[i] == ch[0]){
        j=0;
        ++i;
        while( szFileName[i]!=0  && j<5){
          szBuf[j] = szFileName[i];
          ++j;
          ++i;
        }
        szBuf[j] = 0;
        _wcslwr( szBuf );
        if (szExt){
          wcscpy( szExt, szBuf );
        }
        return szBuf;
      }
    }
  }
  if (szExt){
    szExt[0] = 0;
  }
  return 0;
}

//-----------------------------------------------
// Get file directory from full pathname
// The output directory includes leading drive symbol
// and trailing backslash (\)
// If bIncludeSlash is FALSE then trailing backslash will not be included
//-----------------------------------------------
LPCWSTR okGetFileDir (LPCWSTR szFileName, bool bIncludeBackslash)
{
  static WCHAR szBuf[260];
  WCHAR*  ch = L"\\//";
  INT_PTR i, len;

  wcsncpy( szBuf, szFileName, 258 );
  szBuf[258] = 0;
  len = wcslen( szBuf );
  if (len > 1){
    for (i=len-1; i>=0; --i){
      if (szBuf[i]==ch[0] || szBuf[i]==ch[1]){
        if (bIncludeBackslash){
          szBuf[i+1] = 0;
        }else{
          szBuf[i] = 0;
        }
        return szBuf;
      }
    }
  }
  szBuf[0] = 0;
  return szBuf;
}




//-----------------------------------------------
void okDblToStr (double Val, WCHAR* szBuf, UINT_PTR MaxDec, bool bRemainPoint) 
{
  okNumToStr( Val, szBuf, MaxDec, true, bRemainPoint ); 
}

/********************************************************************
* Convert numerical value to string
*   Val - input value
*   szBuf - output string
*   MaxDec - max number of decimal digits, max - 8
*   bTrimZero - if true then last zeros will be deleted
*********************************************************************/
void okNumToStr (double Val, WCHAR* szBuf, UINT_PTR MaxDec, bool bTrimZero, bool bRemainPoint, UINT_PTR Separator, UINT_PTR TypeChar)
{
  WCHAR    szFmt[16];
  INT_PTR  i, n;

  if (MaxDec==0){
    if (Val>=0){
      i = (INT_PTR)(Val + 0.5);
    }else{
      i = (INT_PTR)(Val - 0.5);
    }
#if _WIN64
    swprintf(szBuf, L"%I64d", i);
#else
    swprintf(szBuf, L"%d", i);
#endif
  }else{
    if (MaxDec > 8){
      MaxDec = 8;
    }
#if _WIN64
    swprintf( szFmt, L"%%.%I64d%c", MaxDec, (int)TypeChar );
#else
    swprintf(szFmt, L"%%.%d%c", MaxDec, TypeChar);
#endif
    swprintf( szBuf, szFmt, Val );
    if (bTrimZero){
      n = wcslen( szBuf );
      for (i=n-1; i>0; --i){
        if (szBuf[i] == L'0'){
          szBuf[i] = 0;
        }else{
          if (szBuf[i] == L'.'){
            if (bRemainPoint){
              // remain point and 1 digit
              szBuf[i+1] = L'0';
              szBuf[i+2] = 0;
            }else{
              szBuf[i] = 0;
            }
          }
          break;
        }
      }
    }else{
      
    }
    // replace separator
    i = 0;
    while( szBuf[i] ){
      if (szBuf[i] == L'.' || szBuf[i] == L','){
        szBuf[i] = (WCHAR)Separator;
      }
      ++i;
    }
  }
}

//-----------------------------------------------
// Delete last blank in the specified text string
//-----------------------------------------------
void okDelLastBlanks (WCHAR* szText)
{
  size_t i = wcslen(szText) - 1;
  for (; i>0; --i){
    if (szText[i] == L' '){
      szText[i] = 0;
    }else{
      break;
    }
  }
}

//-----------------------------------------------
int LWeight_DwgToLcd (OdDb::LineWeight LwDwg)
{
  int LwLcd = 0;
  if (OdDb::LineWeight::kLnWt000<=LwDwg && LwDwg<=OdDb::LineWeight::kLnWt211){
    LwLcd = LwDwg * 10;
  }else{
    LwLcd = LwDwg;
  }
  return LwLcd;
}

//-----------------------------------------------
OdDb::LineWeight LWeight_LcdToDwg (int LwLcd)
{
  OdDb::LineWeight LwDwg = OdDb::LineWeight::kLnWtByLwDefault;
  if (LwLcd == -1){
    LwDwg = OdDb::LineWeight::kLnWtByLayer;
  }else
  if (LwLcd == -2){
    LwDwg = OdDb::LineWeight::kLnWtByBlock;
  }else
  if (LwLcd == -3){
    LwDwg = OdDb::LineWeight::kLnWtByLwDefault;
  }else
  if (0<=LwLcd && LwLcd<=25){
    LwDwg = OdDb::LineWeight::kLnWt000;
  }else
  if (26<=LwLcd && LwLcd<=70){
    LwDwg = OdDb::LineWeight::kLnWt005;
  }else
  if (71<=LwLcd && LwLcd<=110){
    LwDwg = OdDb::LineWeight::kLnWt009;
  }else
  if (111<=LwLcd && LwLcd<=140){
    LwDwg = OdDb::LineWeight::kLnWt013;
  }else
  if (141<=LwLcd && LwLcd<=165){
    LwDwg = OdDb::LineWeight::kLnWt015;
  }else
  if (166<=LwLcd && LwLcd<=190){
    LwDwg = OdDb::LineWeight::kLnWt018;
  }else
  if (191<=LwLcd && LwLcd<=225){
    LwDwg = OdDb::LineWeight::kLnWt020;
  }else
  if (226<=LwLcd && LwLcd<=275){
    LwDwg = OdDb::LineWeight::kLnWt025;
  }else
  if (276<=LwLcd && LwLcd<=325){
    LwDwg = OdDb::LineWeight::kLnWt030;
  }else
  if (326<=LwLcd && LwLcd<=375){
    LwDwg = OdDb::LineWeight::kLnWt035;
  }else
  if (376<=LwLcd && LwLcd<=450){
    LwDwg = OdDb::LineWeight::kLnWt040;
  }else
  if (451<=LwLcd && LwLcd<=515){
    LwDwg = OdDb::LineWeight::kLnWt050;
  }else
  if (516<=LwLcd && LwLcd<=565){
    LwDwg = OdDb::LineWeight::kLnWt053;
  }else
  if (566<=LwLcd && LwLcd<=650){
    LwDwg = OdDb::LineWeight::kLnWt060;
  }else
  if (651<=LwLcd && LwLcd<=750){
    LwDwg = OdDb::LineWeight::kLnWt070;
  }else
  if (751<=LwLcd && LwLcd<=850){
    LwDwg = OdDb::LineWeight::kLnWt080;
  }else
  if (851<=LwLcd && LwLcd<=950){
    LwDwg = OdDb::LineWeight::kLnWt090;
  }else
  if (951<=LwLcd && LwLcd<=1030){
    LwDwg = OdDb::LineWeight::kLnWt100;
  }else
  if (1031<=LwLcd && LwLcd<=1130){
    LwDwg = OdDb::LineWeight::kLnWt106;
  }else
  if (1131<=LwLcd && LwLcd<=1300){
    LwDwg = OdDb::LineWeight::kLnWt120;
  }else
  if (1301<=LwLcd && LwLcd<=1490){
    LwDwg = OdDb::LineWeight::kLnWt140;
  }else
  if (1491<=LwLcd && LwLcd<=1790){
    LwDwg = OdDb::LineWeight::kLnWt158;
  }else
  if (1791<=LwLcd && LwLcd<=2060){
    LwDwg = OdDb::LineWeight::kLnWt200;
  }else{
    LwDwg = OdDb::LineWeight::kLnWt211;
  }
  return LwDwg;
}
