/********************************************************************
* Project: Geom.Lib
* Copyright (C) 2009-2021, Oleg Kolbaskin
* All rights reserved.
*
* Class CGeEllipse implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include "Geom.h"

#define MIN_ARCANG 0.001   // min arc angle

//-----------------------------------------------
CGeEllipse::CGeEllipse () 
{
  m_R1 = m_R2 = 1.0; 
  m_AngArc = m_Rotation = m_AngStart = 0.0; 
  m_bSector = false; 
  m_bNormalized = false;
}

//-----------------------------------------------
CGeEllipse::CGeEllipse (const CGeEllipse& Ell) 
{
  Set(Ell);
}

//-----------------------------------------------
CGeEllipse::CGeEllipse (const CGePoint& Cen, double R1, double R2, 
                        double RotAng, double Ang0, double AngArc, bool bSec) 
{
  m_bNormalized = false;
  Set( Cen, R1, R2, RotAng, Ang0, AngArc, bSec); 
}

//-----------------------------------------------
CGeEllipse::CGeEllipse (double x, double y, double R1, double R2, 
                        double RotAng, double Ang0, double AngArc, bool bSec) 
{
  m_bNormalized = false;
  Set( CGePoint(x,y), R1, R2, RotAng, Ang0, AngArc, bSec); 
}


//-----------------------------------------------
CGeRect CGeEllipse::GeExtents () const
{
  CGeRect Rect;
  GetGeExtents( &Rect );
  return Rect; 
}

//-----------------------------------------------
void CGeEllipse::GetGeExtents (CGeRect* pRect) const
{
  CGePoint Pnt[34];
  UINT_PTR i, npt;

  npt = GetPointsE( Pnt, 32 );
  pRect->Set( Pnt[0], Pnt[1] );
  for (i=2; i<npt; ++i){
    pRect->UpdateExtentsBy( Pnt[i] );
  }
}

//-----------------------------------------------
void CGeEllipse::Set (const CGeEllipse& Ell) 
{
  m_Center     = Ell.m_Center; 
  m_R1         = Ell.m_R1; 
  m_R2         = Ell.m_R2; 
  m_Rotation   = Ell.m_Rotation; 
  m_AngStart   = Ell.m_AngStart; 
  m_AngArc     = Ell.m_AngArc; 
  m_bSector    = Ell.m_bSector;
  m_bNormalized= Ell.m_bNormalized;
}

//-----------------------------------------------
bool CGeEllipse::Set (const CGePoint& Pnt, double R1, double R2, double RotAng, double Ang0, double AngArc, bool bSec) 
{
  m_Center = Pnt; 
  m_Rotation = RotAng; 
  m_AngStart = Ang0; 
  SetArcAngle( AngArc );
  m_bSector = bSec;
  if (SetRadius1( R1 )){
    if (SetRadius2( R2 )){
      return true;
    }
  }
  return false;
//  return SetRadius1( R1, R2 );
}

//-----------------------------------------------
bool CGeEllipse::Set2 (const CGePoint& Center, double R1, double R2, 
                       double RotAngle, double Ang0, double AngArc, bool bSec)
{
  m_Center = Center; 
  m_Rotation = RotAngle; 
  m_AngStart = Ang0; 
  m_AngArc = AngArc;
  m_bSector = bSec;
  m_bNormalized = false;
  if (R1 > 0.0){
    m_R1 = R1;
    if (R2 > 0.0){
      m_R2 = R2;
      return true;
    }
  }
  return false;
}

/*
//-----------------------------------------------
bool CGeEllipse::SetRadius1 (double Radius, double R2, bool bCanFlip, bool* pbFlipped)
{
  if (pbFlipped){
    *pbFlipped = false;
  }
  if (Radius > 0.){
    if (0.0<R2 && R2<=Radius){
      m_R1 = Radius;
      m_R2 = R2;
    }else{
      if (Radius > m_R2){
        m_R1 = Radius;
      }else{
        if (bCanFlip){
          m_R1 = m_R2;
          m_R2 = Radius;
          m_Rotation = geNormalizeAngle( m_Rotation + GE_DEG90 );
          if (IsArc()){
            m_AngStart = geNormalizeAngle( m_AngStart - GE_DEG90 );
          }
          if (pbFlipped){
            *pbFlipped = true;
          }
        }else{
          m_R1 = m_R2;
        }
      }
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeEllipse::SetRadius2 (double Radius, bool bCanFlip, bool* pbFlipped)
{
  if (pbFlipped){
    *pbFlipped = false;
  }
  if (Radius > 0.0){
    if (Radius < m_R1){
      m_R2 = Radius;
    }else{
      if (bCanFlip){
        m_R2 = m_R1;
        m_R1 = Radius;
        m_Rotation = geNormalizeAngle( m_Rotation + GE_DEG90 );
        if (IsArc()){
          m_AngStart = geNormalizeAngle( m_AngStart - GE_DEG90 );
        }
        if (pbFlipped){
          *pbFlipped = true;
        }
      }else{
        m_R2 = m_R1;
      }
    }
    return true;
  }
  return false;
}
*/

//-----------------------------------------------
bool CGeEllipse::SetRadius12 (double R1, double R2)
{
  if ((R1 > 0.0) && (R2 > 0.0)){
    if (m_bNormalized){
      if (R1 >= R2){
        m_R1 = R1;
        m_R2 = R2;
      }else{
        m_R1 = R2;
        m_R2 = R1;
      }
    }else{
      m_R1 = R1;
      m_R2 = R2;
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeEllipse::SetRadius1 (double R)
{
  if (R > 0.0){
    m_R1 = R;
    if (m_bNormalized){
      if (m_R1 < m_R2){
        m_R1 = m_R2;
      }
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeEllipse::SetRadius2 (double R)
{
  if (R > 0.0){
    m_R2 = R;
    if (m_bNormalized){
      if (m_R2 > m_R1){
        m_R2 = m_R1;
      }
    }
    return true;
  }
  return false;
/*
    if (R <= m_R1){
      m_R2 = R;
      return true;
    }else{
      if (m_bNormalized){
      }else{
        m_R2 = R;
        return true;
      }
    }
  }
  return false;
*/
}

// set minor radius (m_R2) by ratio
//-----------------------------------------------
bool CGeEllipse::SetRatio (double Rat)
{
  if (Rat < 0.0){
    return false;
  }
  if ((Rat == 0.0) || (Rat == 1.0)){
    m_R2 = m_R1;
    return true;
  }
  if (m_bNormalized){
    if (Rat < 1.0){
      m_R2 = m_R1 * Rat;
      return true;
    }
    return false;
  }
  m_R2 = m_R1 * Rat;
  return true;
}

//-----------------------------------------------
void CGeEllipse::SetRotation (double Angle)
{
  m_Rotation = geNormalizeAngle( Angle );
}

//-----------------------------------------------
void CGeEllipse::SetStartAngle (double Angle)
{
  m_AngStart = geNormalizeAngle( Angle );
}


//-----------------------------------------------
void CGeEllipse::SetArcAngle (double Angle)
{
  double ang;
  const double MinAng = 0.000001;

  ang = fabs(Angle);
  if (ang < MinAng){
    m_AngArc = 0.0;
    m_AngStart = 0.0;
    m_bSector = false;
  }else{
    if (ang <= GE_2PI){
      m_AngArc = Angle;
    }else{
      while( ang > GE_2PI){
        ang -= GE_2PI;
      }
      m_AngArc = (Angle > 0.0)? ang : -ang;
    }
    ang = fabs(m_AngArc);
    if (ang < MinAng){
      m_AngArc = 0.0;
      m_AngStart = 0.0;
      m_bSector = false;
    }else
    if (ang > (GE_2PI - MinAng)){
      m_AngArc = 0.0;
      m_AngStart = 0.0;
      m_bSector = false;
    }
  }
}

// set start & arc angles by parameters (from AutoCAD)
//-----------------------------------------------
void CGeEllipse::SetParams (double StartPrm, double EndPrm)
{
  double Rat, Ang0, Ang1;

  Rat = Ratio();
  ParamToAngle( StartPrm, &Ang0, Rat );
  ParamToAngle( EndPrm, &Ang1, Rat );
  if (Ang1 > Ang0){
    m_AngStart = Ang0;
    m_AngArc = Ang1 - Ang0;
  }else{
    m_AngStart = Ang0;
    m_AngArc = GE_2PI + Ang1 - Ang0;
  }
}

//-----------------------------------------------
void CGeEllipse::ParamToAngle (double Param, double* pAngle, double RatioR2R1)
{
  CGePoint Pt;

  Pt.Set( cos( Param ), sin( Param ) );
  Pt.y *= RatioR2R1;
  *pAngle = Pt.Angle();
}

//-----------------------------------------------
void CGeEllipse::AngleToParam (double Angle, double* pParam, double RatioR2R1)
{
  CGePoint Pt;

  Pt.Set( cos( Angle ), sin( Angle ) );
  Pt.y /= RatioR2R1;
  *pParam = Pt.Angle();
}


//-----------------------------------------------
double CGeEllipse::EndAngle () const
{
  double ang;
  ang = m_AngStart + m_AngArc;
  // normalize to the range 0..2PI
  if (ang < 0.){
    ang += GE_2PI;
  }else{
    if (ang > GE_2PI){
      ang -= GE_2PI;
    }
  }
  return ang;
}

//-----------------------------------------------
void CGeEllipse::AnglePoint (double Angle, CGePoint* pPnt) const
{
  double Ex, dist, x, y;
  double Rh2, Rv2;
  double cosRot, sinRot, cosAng, sinAng;

  Rh2 = m_R1 * m_R1;
  Rv2 = m_R2 * m_R2;
  cosRot = cos( m_Rotation );
  sinRot = sin( m_Rotation );
  cosAng = cos( Angle );
  sinAng = sin( Angle );
  Ex = 1.0 - (Rv2 / Rh2);
  dist = sqrt( Rv2 / (1.0 - (Ex * cosAng * cosAng)) );
  x = dist * cosAng;
  y = dist * sinAng;
  pPnt->x = m_Center.x + (x*cosRot) - (y*sinRot);
  pPnt->y = m_Center.y + (x*sinRot) + (y*cosRot);
}

//-----------------------------------------------
CGePoint CGeEllipse::AnglePoint (double Angle) const 
{
  CGePoint Pt;

  AnglePoint( Angle, &Pt );
  return Pt;
}

//-----------------------------------------------
CGePoint CGeEllipse::ParamPoint (double Param) const
{
  CGePoint Pt;
  double x, y, cosRot, sinRot;

  x = m_R1 * cos(Param);
  y = (m_R1 * sin(Param)) * Ratio();
  cosRot = cos( m_Rotation );
  sinRot = sin( m_Rotation );
  Pt.x = m_Center.x + x*cosRot - y*sinRot;
  Pt.y = m_Center.y + x*sinRot + y*cosRot;
  return Pt;
}

//-----------------------------------------------
double CGeEllipse::Length () const
{
  CGePoint Pt0, Pt1;
  double   Len, Ex, cosAng, sinAng, R, a;
  double   Rh2 = m_R1 * m_R1;
  double   Rv2 = m_R2 * m_R2;
  double   Ang = 0.0; //m_AngStart;
  double   MaxAng = fabs(m_AngArc);
  double   Step = GE_DEG1/2.0;    // 0.5 degree
  bool     bExit = false;
  UINT_PTR i=0;   

  if (IsFull()){
    Len = GE_PI * ( 3.*(m_R1+m_R2) - sqrt( (3.*m_R1+m_R2) * (m_R1+3.*m_R2) ) );
  }else{
    Len = 0.0;
    for ( ; ; ){ //while( 1 ){
      a = Ang + m_AngStart;
      cosAng = cos( a );
      sinAng = sin( a );
      Ex = 1.0 - Rv2 / Rh2;
      R = sqrt( Rv2 / (1.0 - Ex * cosAng * cosAng) );
      Pt1.x = R * cosAng;
      Pt1.y = R * sinAng;
      if (i > 0){
        Len += Pt0.Distance( Pt1 );
      }
      if (bExit){
        break;
      }
      Ang+=Step;
      if (Ang >= MaxAng){
        Ang = MaxAng;
        bExit = true;
      }
      Pt0 = Pt1;
      ++i;
    }
  }
  return Len;
}


//-----------------------------------------------
double CGeEllipse::Area () const
{
  double ar;

  if (m_AngArc == 0.0){
    ar = GE_PI * m_R1 * m_R2;
  }else{
    ar = 0.0;
  }
  return ar;
}

//-----------------------------------------------
void CGeEllipse::GetPointsOld (CGePoint* Pnt, UINT_PTR NumPoints) const
{
  double   AngStep, Ang, cosRot, sinRot, x, y;
  double   StartPrm, EndPrm, ArcAng, EndAng;
  double   ratio = Ratio();
  bool     bRotate = false;
  UINT_PTR i;
  CGePoint* pPnt = Pnt;

  if (NumPoints < 2){
    Pnt[0] = Pnt[1] = StartPoint();
    Pnt[1].x += 1.0;
    return;
  }
  if (m_Rotation){
    bRotate = true;
    cosRot = cos( m_Rotation );
    sinRot = sin( m_Rotation );
  }else{
    cosRot = 1.0;
    sinRot = 0.0;
  }
  if (IsFull()){
    Ang = 0.0;
    AngStep = GE_2PI / NumPoints;  // angle step
    for (i=0; i<NumPoints; ++i){
      x = m_R1 * cos( Ang );
      y = m_R1 * sin( Ang ) * ratio;
      if (bRotate){
        pPnt->x = m_Center.x + x*cosRot - y*sinRot;
        pPnt->y = m_Center.y + x*sinRot + y*cosRot;
      }else{
        pPnt->x = m_Center.x + x;
        pPnt->y = m_Center.y + y;
      }
      Ang += AngStep;
      ++pPnt;
    }
  }else{
    if (m_bSector){
      --NumPoints;  // remain one point for center
    }
    AngleToParam( m_AngStart, &StartPrm, ratio );
    EndAng = EndAngle();
    AngleToParam( EndAng, &EndPrm, ratio );
    if (m_AngArc >= 0.0){
      if (EndPrm > StartPrm){
        ArcAng = EndPrm - StartPrm;
      }else{
        ArcAng = GE_2PI + EndPrm - StartPrm;
      }
    }else{
      if (EndPrm > StartPrm){
        ArcAng = EndPrm - StartPrm - GE_2PI;
      }else{
        ArcAng = EndPrm - StartPrm;
      }
    }
    *pPnt = ParamPoint( StartPrm ); //StartPoint();
    ++pPnt;
    Ang = StartPrm;
    AngStep = ArcAng / (NumPoints - 1);  // angle step
    for (i=1; i<NumPoints; ++i){
      Ang += AngStep;
      x = m_R1 * cos( Ang );
      y = m_R1 * sin( Ang ) * ratio;
      if (bRotate){
        pPnt->x = m_Center.x + x*cosRot - y*sinRot;
        pPnt->y = m_Center.y + x*sinRot + y*cosRot;
      }else{
        pPnt->x = m_Center.x + x;
        pPnt->y = m_Center.y + y;
      }
      ++pPnt;
    }
    if (m_bSector){
      *pPnt = m_Center;
    }
  }
}

//-----------------------------------------------
void CGeEllipse::GetPolylineOld (CGePolyline* pPline, UINT_PTR NumPoints) const
{
  bool      bClosed;
  CGePoint* pPnt;
  
  if (NumPoints < 3){
    NumPoints = 3;
  }
  pPnt = new CGePoint[NumPoints];
  GetPointsOld( pPnt, NumPoints );
  if (IsFull()){
    bClosed = true;
  }else{
    bClosed = IsSector();
  }
  pPline->SetVers( pPnt, NumPoints, bClosed );
  delete[] pPnt;
}

//-----------------------------------------------
UINT_PTR CGeEllipse::GetPoints (CGePoint* Pnt, double AngStep, UINT_PTR MaxNumPoints) const
{
  UINT_PTR i = 0;
  double   MaxAng = 0.0;
  double   Ang, Ang2, cosRot, sinRot, x, y;
  double   StartPrm, EndPrm, ArcAng, EndAng;
  double   ratio = Ratio();
  bool     bRotate = false;
  bool     bCW = (m_AngArc < 0.0)?true:false;
  CGePoint* pPnt = Pnt;

  if (m_Rotation){
    bRotate = true;
    cosRot = cos( m_Rotation );
    sinRot = sin( m_Rotation );
  }else{
    cosRot = 1.0;
    sinRot = 0.0;
  }
  if (IsFull()){
    // full ellipse
    MaxAng = GE_DEG360 - 0.01;
    Ang = 0.0;
    for ( ; ; ){ //while( 1 ){
      x = m_R1 * cos( Ang );
      y = m_R1 * sin( Ang ) * ratio;
      if (bRotate){
        pPnt->x = m_Center.x + x*cosRot - y*sinRot;
        pPnt->y = m_Center.y + x*sinRot + y*cosRot;
      }else{
        pPnt->x = m_Center.x + x;
        pPnt->y = m_Center.y + y;
      }
      ++pPnt;
      ++i;
      Ang += AngStep;
      if (Ang > MaxAng){
        break;
      }
      if (i >= MaxNumPoints){
        break;
      }
    }
  }else{
    // arc
    AngleToParam( m_AngStart, &StartPrm, ratio );
    EndAng = EndAngle();
    AngleToParam( EndAng, &EndPrm, ratio );
    if (m_AngArc >= 0.0){
      if (EndPrm > StartPrm){
        ArcAng = EndPrm - StartPrm;
      }else{
        ArcAng = GE_2PI + EndPrm - StartPrm;
      }
    }else{
      if (EndPrm > StartPrm){
        ArcAng = EndPrm - StartPrm - GE_2PI;
      }else{
        ArcAng = EndPrm - StartPrm;
      }
    }
    *pPnt = ParamPoint( StartPrm );
    ++pPnt;
    MaxAng = fabs(ArcAng);
    MaxAng -= MIN_ARCANG / 2.0;
    Ang = 0.0;
    for ( ; ; ){ 
      ++i;
      Ang += AngStep;
      if (Ang > MaxAng){
        *pPnt = EndPoint();
        ++pPnt;
        ++i;
        break;
      }
      if (i >= MaxNumPoints){
        return i;
      }
      if (bCW){
        Ang2 = StartPrm - Ang;
      }else{
        Ang2 = StartPrm + Ang;
      }
      x = m_R1 * cos( Ang2 );
      y = m_R1 * sin( Ang2 ) * ratio;
      if (bRotate){
        pPnt->x = m_Center.x + x*cosRot - y*sinRot;
        pPnt->y = m_Center.y + x*sinRot + y*cosRot;
      }else{
        pPnt->x = m_Center.x + x;
        pPnt->y = m_Center.y + y;
      }
      ++pPnt;
    }
    if (m_bSector){
      if (i < MaxNumPoints){
        *pPnt = m_Center;
        ++pPnt;
        ++i;
      }
    }
  }
  return i;
}

// Retrieves on-arc points
// MaxNumPoints - number of points for 360 degrees
// can be only MaxNumPoints = 16, 32, 64, 128, 256, 512
//-----------------------------------------------
UINT_PTR CGeEllipse::GetPointsE (CGePoint* Pnt, UINT_PTR MaxNumPoints) const
{
  CArcPrm  Prm;
  UINT_PTR nPnts;
  double   StartPrm, EndPrm, ArcAng, EndAng;

  switch( MaxNumPoints ){
    case 16:  Prm.m_Resolution = GE_ARCRES_16; break;
    case 32:  Prm.m_Resolution = GE_ARCRES_32; break;
    case 64:  Prm.m_Resolution = GE_ARCRES_64; break;
    case 128: Prm.m_Resolution = GE_ARCRES_128; break;
    case 256: Prm.m_Resolution = GE_ARCRES_256; break;
    case 512: Prm.m_Resolution = GE_ARCRES_512; break;
    default: 
      Prm.m_Resolution = GE_ARCRES_64; 
      break;
  }

  Prm.m_Center = m_Center;
  Prm.m_Radius = m_R1;
  Prm.m_Ratio = Ratio();
  AngleToParam( m_AngStart, &StartPrm, Prm.m_Ratio );
  EndAng = EndAngle();
  AngleToParam( EndAng, &EndPrm, Prm.m_Ratio );
  Prm.m_Ang0 = StartPrm;
  Prm.m_AngRot = m_Rotation;
  if (m_AngArc >= 0.0){
    if (EndPrm > StartPrm){
      ArcAng = EndPrm - StartPrm;
    }else{
      ArcAng = GE_2PI + EndPrm - StartPrm;
    }
    Prm.m_bDirCW = false;
    Prm.m_AngArc = ArcAng;
  }else{
    if (EndPrm > StartPrm){
      ArcAng = EndPrm - StartPrm - GE_2PI;
    }else{
      ArcAng = EndPrm - StartPrm;
    }
    Prm.m_bDirCW = true;
    Prm.m_AngArc = -ArcAng;
  }
  nPnts = geFastArc( Prm, Pnt );
  if (m_bSector){
    Pnt[nPnts] = m_Center;
    ++nPnts;
  }
  return nPnts;
}

//-----------------------------------------------
UINT_PTR CGeEllipse::GetPolylineE (CGePolyline* pPline, UINT_PTR MaxNumPoints) const
{
  CGePoint Pnt[516];
  UINT_PTR nPnts;
  bool bClosed;

  if (MaxNumPoints < 16){
    MaxNumPoints = 16;
  }else{
    if (MaxNumPoints > 512){
      MaxNumPoints = 512;
    }
  }
  nPnts = GetPointsE( Pnt, MaxNumPoints );
  if (IsFull() || IsSector()){
    bClosed = true;
  }else{
    bClosed = false;
  }
  pPline->SetVers( Pnt, nPnts, bClosed );
  return nPnts;
}


/*
// Retrieves arc point by angle
//-----------------------------------------------
void CGeEllipse::GetPointByAngle (double Angle, CGePoint* pPnt) const 
{
  double Ex, dist, x, y;
  double Rh2 = m_R1 * m_R1;
  double Rv2 = m_R2 * m_R2;
  double cosRot = cos( m_Rotation );
  double sinRot = sin( m_Rotation );
  double cosAng = cos( Angle );
  double sinAng = sin( Angle );

  Ex = 1.0 - (Rv2 / Rh2);
  dist = sqrt( Rv2 / (1.0 - (Ex * cosAng * cosAng)) );
  x = dist * cosAng;
  y = dist * sinAng;
  pPnt->x = m_Center.x + (x*cosRot) - (y*sinRot);
  pPnt->y = m_Center.y + (x*sinRot) + (y*cosRot);
}
*/



// Returns a distanse along arc from startpoint
//-----------------------------------------------
double CGeEllipse::GetDistByPoint (const CGePoint& Pnt, CGePoint* pArcPt) const
{
  double Dist, Ang;
  CGePoint StartPt;

  StartPt = StartPoint();
  if (m_AngArc > 0.0){
    Ang = StartPt.RightAngle( m_Center, Pnt );
  }else{
    Ang = StartPt.LeftAngle( m_Center, Pnt );
  }
  Dist = m_R1 * Ang;
  if (pArcPt){
    Ang = m_Center.Angle( Pnt );
    *pArcPt = m_Center.PolarBy( Ang, m_R1 );
  }
  return Dist;  // !!! not exact value (used for trim)
}

// Retrieves 4 points: 0,1 - major axis; 2,3-minor axis
//-----------------------------------------------
void CGeEllipse::GetAxisPoints (CGePoint* Pnt) const
{
  CGePoint Pt;
  int    i;
  double x[] = {m_R1, -m_R1,  0.0,   0.0};
  double y[] = { 0.0,   0.0, m_R2, -m_R2};
  double cosRot = cos( m_Rotation );
  double sinRot = sin( m_Rotation );

  for (i=0; i<4; ++i){
    Pt.x = x[i]*cosRot - y[i]*sinRot;
    Pt.y = x[i]*sinRot + y[i]*cosRot;
    Pnt[i].x = m_Center.x + Pt.x;
    Pnt[i].y = m_Center.y + Pt.y;
  }
}


//-----------------------------------------------
bool CGeEllipse::Contains (const CGePoint& PickPt) const
{
  CGePoint Pnt[65];
  UINT_PTR nPnts, Ret;

  nPnts = GetPointsE( Pnt, 64 );
  if (nPnts){
    Ret = geContainsPoint( Pnt, nPnts, PickPt );
    if (Ret == 1){
      return true;
    }
  }
  return false;
}


//-----------------------------------------------
bool CGeEllipse::Contains (const CGeRect& /*Rect*/) const
{
/*
  if (Contains( rect.LefBot() )){
    if (Contains( rect.LefTop() )){
      if (Contains( rect.RigBot() )){
        if (Contains( rect.RigTop() )){
          return true;
        }
      }
    }
  }
*/
  return false;
}

//-----------------------------------------------
bool CGeEllipse::IsClosed () const
{
  if (m_AngArc == 0.0){
    return true;
  }
  if (m_bSector){
    return true;
  }
  return false;
}


// test is the point are placed inside of the arc included angle
//-----------------------------------------------
bool CGeEllipse::IsArcPoint (const CGePoint& Pnt) const
{
  CGePoint spt = StartPoint();
  CGePoint Pt0, Pt2, cpt;
  double  ang2;

  if (IsArc()){
    Pt2 = Pnt;
    Pt0.Set( spt.x, spt.y );
    cpt.Set( m_Center.x, m_Center.y );
    if (m_AngArc > 0){
      ang2 = Pt0.RightAngle( cpt, Pt2 );
      if (ang2 < m_AngArc){
        return true;
      }
    }else{
      ang2 = Pt0.LeftAngle( cpt, Pt2 );
      if (ang2 < -m_AngArc){
        return true;
      }
    }
    return false;
  }
  return true;
}

#define NPTCROSS 64
//-----------------------------------------------
bool CGeEllipse::Crosses (const CGeRect& Rect) const
{
  CGePoint Pnt[NPTCROSS+2];
  UINT_PTR i, n;
  CGeLine  Line;

  n = GetPointsE( Pnt, NPTCROSS );
  for (i=1; i<n; ++i){
    Line.Set( Pnt[i-1], Pnt[i] );
    if (Rect.Crosses( Line )){
      return true;
    }
  }
  if (m_bSector){
    Line.Set( Pnt[n-1], Pnt[0] );
    if (Rect.Crosses( Line )){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
void CGeEllipse::Move (const CGePoint& Off)
{
  m_Center.Move( Off.x, Off.y );
}

//-----------------------------------------------
void CGeEllipse::Rotate (const CGePoint& Pnt0, double Angle)
{
  m_Center.Rotate( Pnt0, Angle );
  m_Rotation = geNormalizeAngle( m_Rotation + Angle );
}

//-----------------------------------------------
void CGeEllipse::Scale (const CGePoint& Pnt0, double ScaleX, double ScaleY)
{
  double asx, asy;
  if ((ScaleX != 0.0) && (ScaleY != 0.0)){
    m_Center.Scale( Pnt0, ScaleX, ScaleY );
    if (ScaleX == ScaleY){
      if (ScaleX < 0.0){
        ScaleX = -ScaleX;
      }
      m_R1 *= ScaleX;
      m_R2 *= ScaleX;
    }else{
      asx = fabs(ScaleX);
      asy = fabs(ScaleY);
      if (asx < asy){
        m_R1 *= asx;
        m_R2 *= asx;
      }else{
        m_R1 *= asy;
        m_R2 *= asy;
      }
    }
  }
}

//-----------------------------------------------
void CGeEllipse::Mirror (const CGeLine& Line)
{
  CGePoint Pt0;
  CGePoint PtR1 = m_Center.PolarBy( m_Rotation, m_R1 );
  if (IsFull()){
    m_Center.Mirror( Line );
    PtR1.Mirror( Line );
    m_Rotation = m_Center.Angle( PtR1 );
  }else{
    Pt0 = StartPoint();
    m_Center.Mirror( Line );
    PtR1.Mirror( Line );
    Pt0.Mirror( Line );
    m_AngStart = PtR1.RightAngle( m_Center, Pt0 );
    m_Rotation = m_Center.Angle( PtR1 );
    m_AngArc = -m_AngArc;
  }
}

//-----------------------------------------------
void CGeEllipse::IncreaseArcLength (double Delta)
{
  double AngDelta = Delta / ((m_R1+m_R2)/2.0);
  double AngArc = fabs(m_AngArc) + AngDelta + AngDelta;
  if (IsArc() && (AngArc < GE_2PI)){
    if (m_AngArc < 0.0){
      SetStartAngle( m_AngStart + AngDelta );
      SetArcAngle( -AngArc );
    }else{
      SetStartAngle( m_AngStart - AngDelta );
      SetArcAngle( AngArc );
    }
  }
}

//-----------------------------------------------
void CGeEllipse::Save (BYTE* pBuf) const
{
  int sz;
  BYTE* ptr;

  ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( ptr, &m_Center, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( ptr, &m_R1, sz );  ptr += sz;
  memcpy( ptr, &m_R2, sz );  ptr += sz;
  memcpy( ptr, &m_Rotation, sz );  ptr += sz;
  memcpy( ptr, &m_AngStart, sz );  ptr += sz;
  memcpy( ptr, &m_AngArc, sz );  ptr += sz;
  sz = sizeof(bool);
  memcpy( ptr, &m_bSector, sz );  ptr += sz;
}


//-----------------------------------------------
void CGeEllipse::Restore (const BYTE* pBuf)
{
  int sz;
  const BYTE* ptr;

  ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( &m_Center, ptr, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( &m_R1, ptr, sz );  ptr += sz;
  memcpy( &m_R2, ptr, sz );  ptr += sz;
  memcpy( &m_Rotation, ptr, sz );  ptr += sz;
  memcpy( &m_AngStart, ptr, sz );  ptr += sz;
  memcpy( &m_AngArc, ptr, sz );  ptr += sz;
  sz = sizeof(bool);
  memcpy( &m_bSector, ptr, sz );  ptr += sz;
}
