/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
*********************************************************************/
#include "pch.h"

extern int      g_NumEnts;
extern COptions g_Options;

HANDLE Read_Line   (const OdDbLinePtr pOdLine, HANDLE hLcDb, HANDLE hBlock);
void   Read_Arc    (const OdDbArcPtr pOdArc,  HANDLE hLcDb, HANDLE hBlock);
void   Read_Circle (const OdDbCirclePtr pOdCirc, HANDLE hLcDb, HANDLE hBlock);
void   Read_Surf   (OdDbSurfacePtr pOdSurf, HANDLE hLcDb, HANDLE hBlock);
void   Read_Body   (OdDbBodyPtr pOdBody, HANDLE hLcDb, HANDLE hBlock);
void   Read_Region (OdDbRegionPtr pRgn, HANDLE hLcDb, HANDLE hBlock);

void OD_ms2ps (OdDbViewportPtr pVp, OdGeMatrix3d &ResultMat);
void OD_ps2ms (OdDbViewportPtr pVp, OdGeMatrix3d &ResultMat);


//-----------------------------------------------
void Read_EntBase (const OdDbEntityPtr pOdEnt, HANDLE hEnt, HANDLE hLcDb)
{
  OdCmColor OdColor;
  OdDbObjectId idOdObj;
  WCHAR     szColor[32];
  HANDLE    hLayer, hLtype;
  int       ival;

  // identifier
  idOdObj = pOdEnt->objectId();
  SetObjHandle( idOdObj, hEnt, LC_PROP_ENT_ID );
#ifdef _DEBUG
  ival = lcPropGetInt( hEnt, LC_PROP_ENT_ID );
  if (ival == 38239){
    ival = ival;
  }
#endif
  // layer
  idOdObj = pOdEnt->layerId();
  hLayer = DrwGetObjectByHandle( idOdObj, hLcDb, LC_OBJ_LAYER );
  lcPropPutHandle( hEnt, LC_PROP_ENT_LAYER,  hLayer );
  // linetype
  idOdObj = pOdEnt->linetypeId();
  hLtype = DrwGetObjectByHandle( idOdObj, hLcDb, LC_OBJ_LINETYPE );
  lcPropPutHandle( hEnt, LC_PROP_ENT_LINETYPE, hLtype );
  // linetype scale
  lcPropPutFloat( hEnt, LC_PROP_ENT_LTSCALE, pOdEnt->linetypeScale() ); 
  // color
/*
  wcscpy( szColor, L"ByLayer" );
  OdColor = pOdEnt->color();
  switch( OdColor.colorMethod() ){
//    case OdCmEntityColor::kByLayer:   // *Color* of the *layer* on which it resides.
//      lcPropPutInt( 0, LC_PROP_G_COLOR, LC_COLOR_INDEX );
//      Color = LC_COLOR_BYLAYER;
//      break;
    case OdCmEntityColor::kByBlock:    // *Color* of the block reference in which it is contained.
      wcscpy( szColor, L"ByBlock" );
      break;
    case OdCmEntityColor::kByColor:    // *Color* specified by RGB value.
      swprintf( szColor, L"%d,%d,%d", OdColor.red(), OdColor.green(), OdColor.blue() );
      break;
    case OdCmEntityColor::kByACI:      // *Color* specified by an index (ACI) into a *color* table.
      swprintf( szColor, L"%d", OdColor.colorIndex() );
      break;
    case OdCmEntityColor::kByPen:      // *Color* specified by an index into a pen *color* table.
    case OdCmEntityColor::kForeground: // Editor foreground *color*.
    case OdCmEntityColor::kNone:       // No *color*.
//      lcPropPutInt( 0, LC_PROP_G_COLOR, LC_COLOR_INDEX );
//      Color = LC_COLOR_BYBLOCK;
      wcscpy( szColor, L"ByBlock" );
      break;
  }
*/
  OdColor = pOdEnt->color();
  if (OdColorToLcad( OdColor, szColor ) == false){
    wcscpy( szColor, L"ByLayer" );
  }
  lcPropPutStr( hEnt, LC_PROP_ENT_COLOR, szColor );
//  lcPropPutStr( hEnt, LC_PROP_ENT_BCOLOR, szColor );
  lcPropPutStr( hEnt, LC_PROP_ENT_FCOLOR, szColor );
/*
  if (Color.colorMethod() == OdCmEntityColor::kByACI){
    lcPropPutInt( 0, LC_PROP_G_COLOR, LC_COLOR_INDEX );
    lcPropPutInt( hEnt, LC_PROP_ENT_COLOR, Color.colorIndex() );
    lcPropPutInt( hEnt, LC_PROP_ENT_BCOLOR, Color.colorIndex() );
  }else{
    lcPropPutInt( 0, LC_PROP_G_COLOR, LC_COLOR_RGB );
    rgb = RGB( Color.red(), Color.green(), Color.blue() );
    lcPropPutInt( hEnt, LC_PROP_ENT_COLOR, rgb );
    lcPropPutInt( hEnt, LC_PROP_ENT_BCOLOR, rgb );
  }
*/
  // lineweight
//  lcPropPutInt( hEnt, LC_PROP_ENT_LWEIGHT, pOdEnt->lineWeight() );
  ival = pOdEnt->lineWeight();
  if (ival > 0){
    ival *= 10;
  }
  lcPropPutInt( hEnt, LC_PROP_ENT_LWIDTH, ival );
  // hidden
  if (pOdEnt->visibility() == OdDb::kInvisible){
    lcPropPutBool( hEnt, LC_PROP_ENT_HIDDEN, true );
  }
  ++g_NumEnts;
}

//-----------------------------------------------
bool OdColorToLcad (const OdCmColor& OdColor, WCHAR* szColor)
{
  szColor[0] = 0;
  switch( OdColor.colorMethod() ){
    case OdCmEntityColor::kByLayer:   // *Color* of the *layer* on which it resides.
      wcscpy( szColor, L"ByLayer" );
      break;
    case OdCmEntityColor::kByBlock:    // *Color* of the block reference in which it is contained.
      wcscpy( szColor, L"ByBlock" );
      break;
    case OdCmEntityColor::kByColor:    // *Color* specified by RGB value.
      swprintf( szColor, L"%d,%d,%d", OdColor.red(), OdColor.green(), OdColor.blue() );
      break;
    case OdCmEntityColor::kByACI:      // *Color* specified by an index (ACI) into a *color* table.
      swprintf( szColor, L"%d", OdColor.colorIndex() );
      break;
    case OdCmEntityColor::kForeground: // Editor foreground *color*.
      wcscpy( szColor, L"7" );
      break;
    case OdCmEntityColor::kByPen:      // *Color* specified by an index into a pen *color* table.
    case OdCmEntityColor::kNone:       // No *color*.
      wcscpy( szColor, L"ByBlock" );
      break;
    default:
      return false;
  }
  return true;
}

//-----------------------------------------------
void Read_Point (const OdDbPointPtr pOdPoint, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE hPoint;
  BOOL   b;
  double X = pOdPoint->position().x;
  double Y = pOdPoint->position().y;
  double Z = pOdPoint->position().z;
  b = lcPropGetBool( hBlock, LC_PROP_BLOCK_LAYOUT );
  if (b){
    hPoint = lcBlockAddPoint( hBlock, X, Y );
  }else{
    b = lcPropGetBool( hBlock, LC_PROP_BLOCK_DIM );
    if (b){
      hPoint = lcBlockAddPoint2( hBlock, X, Y, LC_POINT_PIXEL, 0 );
    }else{
      b = lcPropGetBool( hBlock, LC_PROP_BLOCK_HATCH ) ;
      if (b){
        hPoint = lcBlockAddPoint2( hBlock, X, Y, LC_POINT_PIXEL, 0 );
      }else{
        hPoint = lcBlockAddPoint( hBlock, X, Y );
      }
    }
  }
  if (hPoint){
    lcPropPutFloat( hPoint, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdPoint, hPoint, hLcDb ); //, hBlock );
  }
}



//-----------------------------------------------
void Read_Image (const OdDbRasterImagePtr pOdImageRef, HANDLE hLcDb, HANDLE hBlock)
{
  WCHAR  szFileName[256], szFoundFName[256], szSearchDir[256], szDrwFname[256], szName[128];
  INT_PTR  res;
  double Lef, Bot, W, H;
  HANDLE hImage, hImageRef;
  OdDbRasterImageDefPtr pOdImage;
  OdGeExtents3d         Ext;

  pOdImage = pOdImageRef->imageDefId().openObject();
  if (pOdImage.isNull() == false){
    wcscpy( szFileName, (LPCWSTR)pOdImage->sourceFileName().c_str() );
    wcscpy( szName, okGetFileName( szFileName, false ));  // without extention
    hImage = lcDrwGetObjectByName( hLcDb, LC_OBJ_IMAGE, szName );
    if (hImage == NULL){
      // add image definition
      wcscpy( szDrwFname, lcPropGetStr( hLcDb, LC_PROP_DRW_FILENAME ));
      wcscpy( szSearchDir, okGetFileDir( szDrwFname ) );
      res = okSearchFile( szFileName, szFoundFName, szSearchDir );
      if (res >= 0){
        hImage = lcDrwAddImage( hLcDb, szName, szFoundFName ); 
        if (hImage){
          SetObjHandle( pOdImage->objectId(), hImage, LC_PROP_TABLE_ID );
        }
      }
    }
    if (hImage){
      pOdImageRef->getGeomExtents( Ext );
      Lef = Ext.minPoint().x;
      Bot = Ext.minPoint().y;
      W = Ext.maxPoint().x - Lef;
      H = Ext.maxPoint().y - Bot;
      hImageRef = lcBlockAddImageRef( hBlock, hImage, Lef, Bot, W, H, false );
      if (hImageRef){
        Read_EntBase( pOdImageRef, hImageRef, hLcDb );
      }
    }
  }
}



//-----------------------------------------------
void Read_Shape (const OdDbShapePtr pOdShape, HANDLE hLcDb, HANDLE hBlock)
{
}

//-----------------------------------------------
void Read_Solid (const OdDbSolidPtr pOdSolid, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d odpt;
  int         i;
//  double      X, Y, Z;
  double    PtX[4], PtY[4], Delta;
  HANDLE      hPline, hLtype;
  OdGeVector3d Norm = pOdSolid->normal();

  hPline = lcBlockAddPolyline( hBlock, LC_PLFIT_BULGE, true, true );
  if (hPline){
    Read_EntBase( pOdSolid, hPline, hLcDb ); //, hBlock );
    for (i=0; i<4; ++i){
      pOdSolid->getPointAt( i, odpt );
      PtX[i] = odpt.x;
      PtY[i] = odpt.y;
    }
    Delta = 0.00001;
    if ((PtX[0]-Delta <= PtX[3]) && (PtX[3] <= PtX[0]+Delta) && 
        (PtY[0]-Delta <= PtY[3]) && (PtY[3] <= PtY[0]+Delta)) 
    {
      lcPlineAddVer( hPline, NULL, PtX[0], PtY[0] );
      lcPlineAddVer( hPline, NULL, PtX[1], PtY[1] );
      lcPlineAddVer( hPline, NULL, PtX[2], PtY[2] );
    }else{
      if ((PtX[2]-Delta <= PtX[3]) && (PtX[3] <= PtX[2]+Delta) && 
          (PtY[2]-Delta <= PtY[3]) && (PtY[3] <= PtY[2]+Delta)) 
      {
        lcPlineAddVer( hPline, NULL, PtX[0], PtY[0] );
        lcPlineAddVer( hPline, NULL, PtX[1], PtY[1] );
        lcPlineAddVer( hPline, NULL, PtX[2], PtY[2] );
      }else{
        lcPlineAddVer( hPline, NULL, PtX[0], PtY[0] );
        lcPlineAddVer( hPline, NULL, PtX[1], PtY[1] );
        lcPlineAddVer( hPline, NULL, PtX[3], PtY[3] );
        lcPlineAddVer( hPline, NULL, PtX[2], PtY[2] );
      }
    }
    hLtype = lcPropGetHandle( hLcDb, LC_PROP_DRW_LINETYPE_CONT );
    lcPropPutHandle( hPline, LC_PROP_ENT_LINETYPE, hLtype );
    lcPropPutFloat( hPline, LC_PROP_ENT_LWIDTH, 0.0 );
//    lcPropPutStr( hPline, LC_PROP_ENT_FILLING, L"Solid" );
    lcPropPutBool( hPline, LC_PROP_ENT_FILLED, true );
    lcPropPutBool( hPline, LC_PROP_PLINE_ODSOLID, true );
/*
    X = pOdSolid->normal().x;
    Y = pOdSolid->normal().y;
    Z = pOdSolid->normal().z;
    if (X!=0. || Y!=0. || Z!= 1.){
      lcEntPutNormal( hPline, X, Y, Z );
    }
*/
//    lcEntUpdate( hPline );
  }
}

//-----------------------------------------------
void Read_Trace (const OdDbTracePtr pOdTrace, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE      hPline;
  OdGePoint3d odpt;
//  OdGeVector3d Norm = pOdTrace->normal();

  hPline = lcBlockAddPolyline( hBlock, LC_PLFIT_BULGE, true, true );
  if (hPline){
    Read_EntBase( pOdTrace, hPline, hLcDb );
    pOdTrace->getPointAt( 0, odpt );
    lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
    pOdTrace->getPointAt( 1, odpt );
    lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
    pOdTrace->getPointAt( 3, odpt );
    lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
    pOdTrace->getPointAt( 2, odpt );
    lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
  }
}

//-----------------------------------------------
void Read_Face3D (const OdDbFacePtr pOdFace, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt[4];
  HANDLE      hFace;
  int         invis[4] = {LC_FACE_EDGE1HIDDEN, LC_FACE_EDGE2HIDDEN, LC_FACE_EDGE3HIDDEN, LC_FACE_EDGE4HIDDEN};
  int         Flags = 0;
  int         i;

  for (i=0; i<4; ++i){
    if (pOdFace->isEdgeVisibleAt( i ) == false){
      Flags |= invis[i];
    }
    pOdFace->getVertexAt( i, Pt[i] );
  }
  if ((Pt[2].x == Pt[3].x) && (Pt[2].y == Pt[3].y)){
    hFace = lcBlockAddFace( hBlock, Flags, Pt[0].x, Pt[0].y, Pt[0].z, Pt[1].x, Pt[1].y, Pt[1].z, Pt[2].x, Pt[2].y, Pt[2].z );
  }else{
    hFace = lcBlockAddFace4( hBlock, Flags, Pt[0].x, Pt[0].y, Pt[0].z, Pt[1].x, Pt[1].y, Pt[1].z, Pt[2].x, Pt[2].y, Pt[2].z, Pt[3].x, Pt[3].y, Pt[3].z );
  }
  if (hFace){
    Read_EntBase( pOdFace, hFace, hLcDb ); //, hBlock );
  }
}

//-----------------------------------------------
void Read_FaceMesh (const OdDbPolyFaceMeshPtr pOdPFmesh, HANDLE hLcDb, HANDLE hBlock)
{
}

//-----------------------------------------------
void Read_PgonMesh (const OdDbPolygonMeshPtr pOdMesh, HANDLE hLcDb, HANDLE hBlock)
{
}

//-----------------------------------------------
void Read_3dSolid (OdDb3dSolidPtr pOdSolid, HANDLE hLcDb, HANDLE hBlock)
{
  OdBrBrep Brep;
  OdGeExtents3d odExtents;
  OdRxObjectPtr pOdEntity;
  OdDbSurfacePtr  pOdSurf;
  OdDbBodyPtr     pOdBody;
  OdDbRegionPtr   pOdReg;
//  OdDbLinePtr   pOdLine;
//  OdDbArcPtr    pOdArc;
//  OdDbCirclePtr pOdCircle;
  OdRxObjectPtrArray pObjs;
  OdResult Res;
  int i, n, n1=0, n2=0;

  pOdSolid->getGeomExtents( odExtents );
  pOdSolid->brep( Brep );
//  Brep.getLineContainment(

//  OdDbEntityPtr pEntity = pEntIter->entity(OdDb::kForWrite);
  Res = pOdSolid->explode(pObjs);
  if (Res == eOk){
    n = pObjs.length();
    if (n > 0){
      for (i=0; i<n; ++i){
        pOdEntity = pObjs.getAt( i );
        if (pOdEntity->isKindOf( OdDbSurface::desc() )){      // SURFACE
          pOdSurf = OdDbSurface::cast( pOdEntity );
          Read_Surf( pOdSurf, hLcDb, hBlock );
          ++n1;
        }else
        if (pOdEntity->isKindOf( OdDbBody::desc() )){       // BODY
          pOdBody = OdDbBody::cast( pOdEntity );
          Read_Body( pOdBody, hLcDb, hBlock );
        }else
        if (pOdEntity->isKindOf( OdDbRegion::desc() )){    // REGIONS
          pOdReg = OdDbRegion::cast( pOdEntity );
          Read_Region( pOdReg, hLcDb, hBlock );
          ++n2;
        }
      }
    }
  }
  hLcDb = 0;
}

//-----------------------------------------------
void Read_Surf (OdDbSurfacePtr pOdSurf, HANDLE hLcDb, HANDLE hBlock)
{
}

//-----------------------------------------------
void Read_Body (OdDbBodyPtr pOdBody, HANDLE hLcDb, HANDLE hBlock)
{
}


//-----------------------------------------------
void Read_Viewport (const OdDbViewportPtr pOdVport, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d       ViewTarget, PsPoint, MsPoint;
  OdGePoint2d       ViewCenter;
  OdDbObjectIdArray odLayerIds;
  OdDbObjectId      odId;
  OdChar            szOdBuf[256];
  OdGeMatrix3d      Mat_ms2ps;
  WCHAR   szVportHandle[32];
  WCHAR*  psz;
  double  Xc, Yc, Lef, Bot, W, H, H2, Scale, Angle;
  int     i, n;
  HANDLE  hVport=0;
  HANDLE  hLayer, hView=0;

#ifdef _DEBUG
  WCHAR  szBName[128], szLName[128], szOdHandleB[32], szOdHandleL[32], szId[16];
  wcscpy( szBName, lcPropGetStr( hBlock, LC_PROP_BLOCK_NAME ));
  wcscpy( szLName, lcPropGetStr( hBlock, LC_PROP_BLOCK_LAYOUTNAME ));
  wcscpy( szId, lcPropGetStr( hBlock, LC_PROP_TABLE_ID ));
  wcscpy( szOdHandleB, lcPropGetStr( hBlock, LC_PROP_TABLE_ODA_HANDLE ));
  wcscpy( szOdHandleL, lcPropGetStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE ));
#endif

  pOdVport->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
  psz = (WCHAR*)szOdBuf;
  wcscpy( szVportHandle, lcPropGetStr( hBlock, LC_PROP_BLOCK_ODA_VPORTHANDLE ) );
  if (wcsicmp(psz, szVportHandle) == 0){
    // this is layout overall viewport, ignore it
    return;
  }

  // viewport's position on Paper space
  Xc = pOdVport->centerPoint().x;
  Yc = pOdVport->centerPoint().y;
  H = pOdVport->height();
  W = pOdVport->width();
  Lef = Xc - W/2.0;
  Bot = Yc - H/2.0;

  hVport = lcBlockAddViewport( hBlock, Lef, Bot, W, H, 0,0, 0, 0 );
  if (hVport){
    Read_EntBase( pOdVport, hVport, hLcDb );

    ViewCenter = pOdVport->viewCenter();
    ViewTarget = pOdVport->viewTarget();
    H2 = pOdVport->viewHeight();
    Scale = H2 / H;
    Angle = pOdVport->twistAngle();
//    pOdVport->getUcs( UcsOrigin, UcsXAxis, UcsYAxis );
    PsPoint.x = Xc;
    PsPoint.y = Yc;
    PsPoint.z = 0.0;
    OD_ms2ps( pOdVport, Mat_ms2ps );
    MsPoint = Mat_ms2ps.inverse() * PsPoint;  // ps2ms
    lcVportSetView( hVport, MsPoint.x, MsPoint.y, Scale, Angle );


    // frozen layers at the viewport
//    lcVportFrolClear( hVport );
    pOdVport->getFrozenLayerList( odLayerIds );
    n = odLayerIds.length();
    if (n > 0){
      for (i=0; i<n; ++i){
        odId = odLayerIds.getAt( i );
        hLayer = DrwGetObjectByHandle( odId, hLcDb, LC_OBJ_LAYER );
        if (hLayer){
//          lcVportFrolAdd( hVport, hLayer);
        }
      }
    }
  }
}


//-----------------------------------------------
void Read_Wipeout (const OdDbWipeoutPtr pOdWipe, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3dArray Vers;
  OdGePoint3d OdPt3d;
  HANDLE   hPline, hVer, hDrw; //, hView;
  int      i, n;
//  COLORREF Color;

  hPline = lcBlockAddPolyline( hBlock, LC_PLFIT_NONE, true, true );
  if (hPline){
    Read_EntBase( pOdWipe, hPline, hLcDb ); 
    lcPropPutBool( hPline, LC_PROP_ENT_WIPEOUT, true );
//    lcPropPutInt( hPline, LC_PROP_ENT_FCOLORI, LC_COLOR_WIPEOUT );
    hDrw = lcPropGetHandle( hBlock, LC_PROP_BLOCK_DRW );
/*
    hView = lcDrwGetViewByBlock( hDrw, hBlock );
    if (hView){
      Color = lcPropGetInt( hView, LC_PROP_VIEW_COLORBACK );
      lcPropPutInt( hPline, LC_PROP_ENT_COLORT, Color );
    }
*/
  }else{
    return;
  }
  pOdWipe->getVertices( Vers );
  n = Vers.length();
  for (i=0; i<n; ++i){
    OdPt3d = Vers.getAt( i );
    hVer = lcPlineAddVer( hPline, NULL, OdPt3d.x, OdPt3d.y );
  }
}


//-----------------------------------------------
void Read_SortEnts (OdDbSortentsTablePtr pOdSortEnts, HANDLE hLcDb, HANDLE hBlock)
{
  int nEnts, i;
//  HANDLE hEnt;
  COkID* idEnt = NULL;
  COkID* pid = NULL;
//  COkID* idEnt2 = NULL;
  HandlePair Pair;
  OdDbHandle Handle1, Handle2;
  HandlePairsArray HandlePairs;
  pOdSortEnts->getAbsoluteDrawOrder( HandlePairs );
  nEnts = HandlePairs.length();
//  if (1 < nEnts){
  if ((1 < nEnts) && (nEnts < 32000)){
    idEnt = new COkID[nEnts+nEnts];
    pid = idEnt;
//    HandlePairs.first();
    for (i=0; i<nEnts; ++i){
      Pair = HandlePairs.getAt( i ); 
      Handle1 = Pair.second.getHandle();
      Handle2 = Pair.first;
      pid->SetHex( (LPCWSTR)(Handle1.ascii()) );
      ++pid;
      pid->SetHex( (LPCWSTR)(Handle2.ascii()) );
      ++pid;
/*
      hEnt = lcBlockGetEntByIDH( hBlock, Handle.ascii() );
      if (hEnt){
        Handle2 = Pair.first;
        lcPropPutStr( hEnt, LC_PROP_ENT_PRIORITY, Handle2.ascii() );
      }
*/
    }
//    lcBlockSortEnts( hBlock, false, NULL );
     lcBlockSortEnts2( hBlock, (HANDLE)idEnt, nEnts );
    delete[] idEnt;
  }
}




#define LC_MAX_HPDASH        8       // max num of dashes in hatch pattern definition

// definition of line of hatch pattern
struct LC_HPLINE {
  double Angle;     // line angle
  double x0;        // pattern-through point, indicating a point through which the pattern passes.
  double y0;
  double dx;        // pattern offset
  double dy;
  USHORT nDash;     // number of dash length items for this line.
  double Dash[LC_MAX_HPDASH];    // lengths of a dashes, negative length means space between dashes
  double DashLength;    // sum of absolute lengths of dashes
};

#define MAXGENVERS   1024
#define MAXPATHVERS  15000

//-----------------------------------------------
void Read_Hatch (const OdDbHatchPtr pOdHatch, HANDLE hLcDb, HANDLE hBlock)
{
  static int ih = 1;
  WCHAR szBuf[32];
  OdDbEntityPtr      pOdEntity;
  OdDbLinePtr        pOdLine;
  OdDbSolidPtr       pOdSolid;
  OdDbPointPtr       pOdPoint;
  OdRxObjectPtr      rxptr;
  OdRxObjectPtrArray arr;
  OdGePoint2d        odpt;
  HANDLE             hHatBlock;
  HANDLE             hHatBlockRef;
  CGePoint           Pt0, Pt1, Pt2;
  COkID              id;
  int                i, n, nSeeds;
  OdGeVector3d Norm = pOdHatch->normal();

  if (g_Options.GetImp_HatchAsBlock()){
    pOdHatch->explode( arr );
    n = arr.length();
    if (n > 0){
      swprintf( szBuf, L"*H%d", ++ih );
      hHatBlock = lcDrwAddBlock( hLcDb, szBuf, 0,0 );
      if (hHatBlock){
        for (i=0; i<n; ++i){
          rxptr = arr.getAt( i );
          if (rxptr->isKindOf( OdDbLine::desc() )){         // LINE
            pOdLine = OdDbLine::cast( rxptr );
            Read_Line( pOdLine, hLcDb, hHatBlock );
          }else{
            if (rxptr->isKindOf( OdDbSolid::desc() )){      // SOLID
              pOdSolid = OdDbSolid::cast( rxptr );
              Read_Solid( pOdSolid, hLcDb, hHatBlock );
            }else{
              n = n;
/*
              if (rxptr->isKindOf( OdDbPoint::desc() )){      // POINT
                pOdPoint = OdDbPoint::cast( rxptr );
                Read_Point( pOdPoint, pLcDb, pHatBlock );
              }else{
                n=n;
              }
*/
            }
          }
        }
/*
        pHatBlock->SetHidden();
        pHatBlock->SetUfScaling( LC_TRUE );      // 1 - Block references can be scaled only uniformly.
        pHatBlock->SetUnits( LC_UNIT_UNDEFINED );
*/
        nSeeds = pOdHatch->numSeedPoints();
        if (nSeeds > 0){
          odpt = pOdHatch->getSeedPointAt( 0 );
          Pt0.Set( odpt.x, odpt.y );
        }else{
          Pt0.Set( 0.0, 0.0 );
        }
        lcPropPutFloat( hHatBlock, LC_PROP_BLOCK_X, Pt0.x );
        lcPropPutFloat( hHatBlock, LC_PROP_BLOCK_Y, Pt0.y );
//        pHatBlock->SetBasePoint( pt0 );   // Origin point

        // insert hatch block
        hHatBlockRef = lcBlockAddBlockRef( hBlock, hHatBlock, Pt0.x, Pt0.y, 1., 0. );
        if (hHatBlockRef){
          Read_EntBase( pOdHatch, hHatBlockRef, hLcDb ); //, hBlock );
//          pBlockRef->SetID( id );
//          pBlockRef->SetRefBlock( pHatBlock, LC_FALSE ); //pHatBlock->ID(), LC_FALSE );
//          pBlockRef->SetPosition( pt0 );
        }
      }
    }
  }else{
OdDbObjectId idOdObj;
static OdChar szHex[32];
    HANDLE           hHatch;
    WCHAR            szName[256];
    int              i, j, k, nv, nb, nDashes;
    int              PathType, PixelSizePresent=0;
    bool             bCW; //isccw;
    double           dval, bulge, Angle;
    double           delta = 0.0001;
    double           LineAngle, baseX, baseY, offsetX, offsetY;
    double           StartAngle, EndAngle, ArcAngle, Rmax, Rmin, weight;
    int              nSplineDegree, nSplineKnots, nSplineCtrlPts;
    bool             bRational;
    CGePoint         Pt;
    CGe3dPoint       Normal;
    OdInt32          PathFlag;
    OdGeDoubleArray  dashes;
    OdGePoint2dArray vertices;
    OdGeDoubleArray  bulges;
    EdgeArray        edges;
    OdGePoint2d      odpt, odpt1, odpt2;
    OdGeVector2d     odvec;
    OdGeCurve2d*     pCurve = NULL;
    OdGeLinearEnt2d* pLine = NULL;
    OdGeEllipArc2d*  pEll = NULL;
    OdGeCircArc2d*   pArc = NULL;
    OdGeNurbCurve2d* pNurb = NULL;
    LC_HPLINE        PatLine;
    HANDLE           hDbPline = NULL;
    HANDLE           hDbEllipse = NULL;
    CGePoint*        BufPathVers = NULL;
    CGePoint*        pPathVer = NULL;
    int              nPathVers = 0;
    CGePoint*        BufGenVers = NULL;
    CGePoint*        pGVer = NULL;
    UINT_PTR         nGenVers = 0;
    UINT_PTR         m;
    CGeArc           Arc;
    CGeCircle        Circ;
    CGeEllipse       Ell;
    UINT             Quality = 20;
    bool    bSolidFill = pOdHatch->isSolidFill();
    int     nPaths = pOdHatch->numLoops();
    int     nLines = pOdHatch->numPatternDefinitions();
    int     nSeeds = pOdHatch->numSeedPoints();
    double  PatAngle = pOdHatch->patternAngle();
    double  PatScale = pOdHatch->patternScale();
    double  nx = pOdHatch->normal().x;
    double  ny = pOdHatch->normal().y;
    double  nz = pOdHatch->normal().z;
    double  Z = pOdHatch->elevation();
    CGe3dPoint  Pt3d, Xaxis, Yaxis, Zaxis;
    int     nEntPaths;
#ifdef _DEBUG
//    CGePoint PtBuf[256];
#endif

    if (nPaths == 0){
      return;
    }
    Zaxis.Set( nx, ny, nz );
    Zaxis.GetArbitraryAxes( &Xaxis, &Yaxis );
    memset( szName, 0, sizeof(szName) );
    wcsncpy( szName, (LPCWSTR)(pOdHatch->patternName().c_str()), 255 );
    hHatch = lcBlockAddHatch( hBlock, L"", L"", 1.0, 0 );
    if (hHatch){
      BufPathVers = new CGePoint[MAXPATHVERS];
      memset( BufPathVers, 0, sizeof(CGePoint)*MAXPATHVERS );
      nPathVers = 0;
      BufGenVers = new CGePoint[MAXGENVERS];
      memset( BufGenVers, 0, sizeof(CGePoint)*MAXGENVERS );
      nGenVers = 0;
      Read_EntBase( pOdHatch, hHatch, hLcDb ); //, hBlock );
      idOdObj = pOdHatch->objectId();
//  memset( szHex, 0, sizeof(szHex) );
      idOdObj.getHandle().getIntoAsciiBuffer( szHex );
      if (wcsicmp((LPCWSTR)szHex,L"EFAA2F")==0){
        szHex[30] = 0;
      }
      if (nx!=0. || ny!=0. || nz!= 1.){
//        lcEntPutNormal( hHatch, nx, ny, nz );
      }
      lcPropPutFloat( hHatch, LC_PROP_ENT_Z, Z );
      lcPropPutFloat( hHatch, LC_PROP_HATCH_ANGLE, PatAngle );
      lcPropPutFloat( hHatch, LC_PROP_HATCH_SCALE, PatScale );
      lcPropPutStr( hHatch, LC_PROP_HATCH_NAME, szName );
      if (bSolidFill){
        lcPropPutBool( hHatch, LC_PROP_HATCH_SOLID, true );
      }
      lcPropPutBool( hHatch, LC_PROP_HATCH_ASSOC, false );
/*
      pHatch->put_Associative( _pOdHatch->associative() );
      pHatch->put_Style( _pOdHatch->hatchStyle() );
      pHatch->put_PatternType( _pOdHatch->patternType() );
      pHatch->put_Double( _pOdHatch->patternDouble() );
      pHatch->put_PixelSize( _pOdHatch->pixelSize() );
*/
      PixelSizePresent = 0;
      // read paths
      lcHatchBoundStart( hHatch );
      nEntPaths = 0;
      for (i=0; i<nPaths; ++i){
        memset( BufPathVers, 0, sizeof(CGePoint)*MAXPATHVERS );
        nPathVers = 0;
        pPathVer = BufPathVers;
        PathFlag = pOdHatch->loopTypeAt( i );
        if (PathFlag & OdDbHatch::kPolyline){
          pOdHatch->getLoopAt( i, vertices, bulges );
          nv = vertices.logicalLength();
          nb = bulges.logicalLength();
          if (nb == 0){
            // linear polyline
            for (j=0; j<nv; ++j) {
              odpt = vertices.getAt( j );
              Pt3d.Set( odpt.x, odpt.y, Z );
              Pt3d.Transform( Xaxis, Yaxis, Zaxis );
              pPathVer->Set( Pt3d.x, Pt3d.y );
              ++pPathVer;
              ++nPathVers;
            }
          }else{
            // polyline with bulges
            bulge = 0.0;
            for (j=0; j<nv; ++j){
              odpt = vertices.getAt( j );
              Pt2.Set( odpt.x, odpt.y );
              if (bulge){
                if (Arc.SetSBE( Pt1, bulge, Pt2 )){
                  ArcAngle = Arc.ArcAngle() * LC_RAD_TO_DEG;
                  Angle = fabs(ArcAngle);
                  if (Angle < 10.0){
                    Quality = 256;
                  }else{
                    Quality = 128;
                  }
                  nGenVers = Arc.GetPointsA( BufGenVers, Quality );
                  --nGenVers;
                  pGVer = BufGenVers;
                  for (m=0; m<nGenVers; ++m){
                    Pt3d.Set( pGVer->x, pGVer->y, Z );
                    Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                    ++pGVer;
                    pPathVer->Set( Pt3d.x, Pt3d.y );
                    ++pPathVer;
                    ++nPathVers;
                  }
                }else{
                  Pt3d.Set( Pt1.x, Pt1.y, Z );
                  Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                  pPathVer->Set( Pt3d.x, Pt3d.y );
                  ++pPathVer;
                  ++nPathVers;
                }
              }else{
                if (j == 0){
                  Pt0 = Pt2;
                }else{
                  Pt3d.Set( Pt1.x, Pt1.y, Z );
                  Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                  pPathVer->Set( Pt3d.x, Pt3d.y );
                  ++pPathVer;
                  ++nPathVers;
                }
              }
              bulge = bulges.getAt( j );
              Pt1 = Pt2;
            }
            // last vertex
            if (bulge){
              Pt2 = Pt0;
              Arc.SetSBE( Pt1, bulge, Pt2 );
              ArcAngle = Arc.ArcAngle() * LC_RAD_TO_DEG;
              Angle = fabs(ArcAngle);
              if (Angle < 10.0){
                Quality = 256;
              }else{
                Quality = 128;
              }
              nGenVers = Arc.GetPointsA( BufGenVers, Quality );
              nGenVers--;
              pGVer = BufGenVers;
              for (m=0; m<nGenVers; ++m){
                Pt3d.Set( pGVer->x, pGVer->y, Z );
                Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                ++pGVer;
                pPathVer->Set( Pt3d.x, Pt3d.y );
                ++pPathVer;
                ++nPathVers;
              }
            }else{
              Pt3d.Set( Pt1.x, Pt1.y, Z );
              Pt3d.Transform( Xaxis, Yaxis, Zaxis );
              pPathVer->Set( Pt3d.x, Pt3d.y );
              ++pPathVer;
              ++nPathVers;
            }
          }
          if (nPathVers >= 3){
            Pt0 = BufPathVers[0];
            Pt1 = BufPathVers[nPathVers-1];
//            if (Pt0.IsEqual(Pt1)){
              pPathVer = BufPathVers;
              for (j=0; j<nPathVers; ++j){
                lcHatchBoundPoint( hHatch, pPathVer->x, pPathVer->y );
                ++pPathVer;
              }
              lcHatchBoundEndLoop( hHatch );
              ++nEntPaths;
//            }
          }

        }else{

          pOdHatch->getLoopAt( i, edges );
          nb = edges.logicalLength();
          for (j=0; j<nb; ++j){
            pCurve = edges.getAt( j );
            PathType = pCurve->type();
            switch( PathType ){
              case OdGe::kLineSeg2d:    // LINE
                pLine = (OdGeLinearEnt2d*)pCurve;
                pLine->hasStartPoint( odpt );
                odpt = pLine->evalPoint( 0 );
                Pt3d.Set( odpt.x, odpt.y, Z );
                Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                pPathVer->Set( Pt3d.x, Pt3d.y );
                ++pPathVer;
                ++nPathVers;
                odpt = pLine->evalPoint( 1 );
                Pt3d.Set( odpt.x, odpt.y, Z );
                Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                pPathVer->Set( Pt3d.x, Pt3d.y );
                ++pPathVer;
                ++nPathVers;
                break;

              case OdGe::kCircArc2d:    // CIRCULAR ARC
                pArc = (OdGeCircArc2d*)pCurve;
                odpt  = pArc->center();
                odpt1 = pArc->startPoint();
                odpt2 = pArc->endPoint();
                if (odpt2.isEqualTo( odpt1 )){
                  // this is a circle
                  dval = odpt.distanceTo( odpt1 );
                  Circ.Set( CGePoint(odpt.x,odpt.y), dval );
                  nGenVers = 128;
                  Circ.GetPoints( BufGenVers, nGenVers );
                  BufGenVers[nGenVers] = BufGenVers[0];
                  ++nGenVers;
                }else{
                  bCW = pArc->isClockWise();
                  Arc.SetCSE( CGePoint(odpt.x,odpt.y), CGePoint(odpt1.x,odpt1.y), CGePoint(odpt2.x,odpt2.y), bCW );
                  ArcAngle = Arc.AngleArc();
                  Angle = fabs(ArcAngle) * LC_RAD_TO_DEG;
                  if (Angle < 10.0){
                    Quality = 256;
                  }else{
                    Quality = 128;
                  }
                  nGenVers = Arc.GetPointsA( BufGenVers, Quality, true );
                }
                pGVer = BufGenVers;
                for (m=0; m<nGenVers; ++m){
                  Pt3d.Set( pGVer->x, pGVer->y, Z );
                  Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                  ++pGVer;
                  pPathVer->Set( Pt3d.x, Pt3d.y );
                  ++pPathVer;
                  ++nPathVers;
                }
                break;

              case OdGe::kEllipArc2d:   // ELLIPTICAL ARC
                pEll = (OdGeEllipArc2d*)pCurve;
                odpt  = pEll->center();
                odvec = pEll->majorAxis();
                Rmax  = pEll->majorRadius();
                Rmin  = pEll->minorRadius();
                StartAngle = pEll->startAng();
                EndAngle = pEll->endAng();
                bCW = pEll->isClockWise();
                dval = Rmax / Rmin;
                Pt.Set( odvec.x, odvec.y );
                Ell.Set( CGePoint(odpt.x,odpt.y), Rmax, Rmin, Pt.Angle() );
                if (bCW){
                  Ell.SetStartAngle( GetAcadEllipseAngle( -StartAngle, dval ));
                  EndAngle = GetAcadEllipseAngle( -EndAngle, dval );
                }else{
                  Ell.SetStartAngle( GetAcadEllipseAngle( StartAngle, dval ));
                  EndAngle = GetAcadEllipseAngle( EndAngle, dval );
                }
                if (EndAngle > Ell.StartAngle()){
                  Ell.SetArcAngle( EndAngle - Ell.StartAngle() );
                }else{
                  Ell.SetArcAngle( LC_DEG360 + EndAngle - Ell.StartAngle() );
                }
                if (bCW){
                  Ell.SetArcAngle( Ell.ArcAngle() - LC_DEG360 );
                }
                if (Ell.IsFull()){
                  Quality = 128;
                }else{
                  ArcAngle = Ell.ArcAngle() * LC_RAD_TO_DEG;
                  Angle = fabs(ArcAngle);
                  if (Angle < 10.0){
                    Quality = 256;
                  }else{
                    Quality = 128;
                  }
                }
                nGenVers = Ell.GetPointsE( BufGenVers, Quality );
                pGVer = BufGenVers;
                for (m=0; m<nGenVers; ++m){
                  Pt3d.Set( pGVer->x, pGVer->y, Z );
                  Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                  ++pGVer;
                  pPathVer->Set( Pt3d.x, Pt3d.y );
                  ++pPathVer;
                  ++nPathVers;
                }
                break;

              case OdGe::kNurbCurve2d:    // SPLINE
                pNurb = (OdGeNurbCurve2d*)pCurve;
                nSplineDegree = pNurb->degree();
                nSplineKnots = pNurb->numKnots();
                nSplineCtrlPts = pNurb->numControlPoints();
                bRational = pNurb->isRational();
//                nVertex = 0;
//                nKnot = 0;
                if (nSplineKnots > 0){
///                  pDbPline->AllocKnots( nSplineKnots );
                  for (k=0; k<nSplineKnots; ++k) {
///                    pDbPline->SetKnot( k, pNurb->knotAt( k ) );
//                  if (nKnot < MAX_POLYVERTEX){
//                    Knot[nKnot] = pNurb->knotAt( k );
//                    ++nKnot;
//                  }
                  }
                }
                for (k=0; k<nSplineCtrlPts; ++k) {
//                  memset( Vertex + nVertex, 0, sizeof(POLYVERTEX) );
                  odpt = pNurb->controlPointAt( k );
                  if (bRational){
                    weight = pNurb->weightAt( k );
                  }else{
                    weight = 1.0;
                  }
                  Pt3d.Set( odpt.x, odpt.y, Z );
                  Pt3d.Transform( Xaxis, Yaxis, Zaxis );
                  pPathVer->Set( Pt3d.x, Pt3d.y );
                  ++pPathVer;
                  ++nPathVers;

///                  pVer = pDbPline->AddVertex( CGePoint(odpt.x,odpt.y) );
///                  pVer->SetWeight( weight );
                }
//                pPline = new CEntPolyline( this, 0, PLINE_3D );
//                pPline->put_Knots( Knot, nKnot );
///                pDbPline->SetClosed( pNurb->isClosed() );
///                pDbPline->UpdateExtents();
                if (bSolidFill){
                  Quality = 30;
                }else{
                  Quality = 20;
                }
//                n = 0;
/*
                if (nSplineDegree == 2){
                  pDbPline->SetFitType( LC_PLFIT_QUAD ); 
                  pDbPline->Gen( &Pline, Quality );
//                    GenBSpline2( Quality, GenVer+nGenVer, &n, MAX_GENVER-nGenVer);
                }else{
                  pDbPline->SetFitType( LC_PLFIT_QUAD ); 
                  pDbPline->SetDegree( nSplineDegree );
                  pDbPline->SetRational( pNurb->isRational() );
                  pDbPline->Gen( &Pline, Quality );
*/
/*
                  pPline->put_Degree( nSplineDegree );
                  pPline->put_Rational( pNurb->isRational() );
                  pPline->GenBSpline( Quality, GenVer+nGenVer, &n, MAX_GENVER-nGenVer);
*/
///                }
//                nGenVer += n;
//                delete pPline;
//                pPline = NULL;
                break;
            }
          }
//          memset( PtBuf, 0, sizeof(PtBuf) );
          if (nPathVers >= 3){
            Pt0 = BufPathVers[0];
            Pt1 = BufPathVers[nPathVers-1];
            if (Pt0.IsEqual(Pt1)){
              pPathVer = BufPathVers;
              for (j=0; j<nPathVers; ++j){
                lcHatchBoundPoint( hHatch, pPathVer->x, pPathVer->y );
//                PtBuf[j] = *pPathVer;
                ++pPathVer;
              }
              lcHatchBoundEndLoop( hHatch );
              ++nEntPaths;
            }
          }
        }
      }
      lcHatchBoundEnd( hHatch );
      delete[] BufPathVers;
      BufPathVers = NULL;
      delete[] BufGenVers;
      BufGenVers = NULL;
      if (nEntPaths == 0){
        nEntPaths = 0;
      }

      // read data of hatch pattern
      if (bSolidFill == false){
        lcHatchPatStart( hHatch );
        memset( &PatLine, 0, sizeof(LC_HPLINE) );
        for (i=0; i<nLines; ++i){
          pOdHatch->getPatternDefinitionAt( i, LineAngle, baseX, baseY, offsetX, offsetY, dashes );  
          nDashes = dashes.logicalLength();
          for (j=0; j<nDashes; ++j){
            dval = dashes.getAt( j );
            if (j < LC_MAX_HPDASH){
              dval /= PatScale;  // read below about scale correction
              PatLine.Dash[j] = dval;  
              PatLine.DashLength += fabs( dval );
            }
          } 
          PatLine.nDash = (nDashes<=LC_MAX_HPDASH)? nDashes : LC_MAX_HPDASH;

          // AutoCAD hatch pattern data are stored in drawing corrected upon scale 
          // and rotation angle. LiteCAD hatch algorithm need original pattern data, 
          // as it stored in *.PAT file,  therefore we must the data rid off from 
          // scale and rotation
          PatLine.dx = offsetX * cos(-LineAngle) - offsetY * sin(-LineAngle);
          PatLine.dy = offsetX * sin(-LineAngle) + offsetY * cos(-LineAngle);
          PatLine.dx /= PatScale;
          PatLine.dy /= PatScale;
          PatLine.x0 = baseX * cos(-PatAngle) - baseY * sin(-PatAngle);
          PatLine.y0 = baseX * sin(-PatAngle) + baseY * cos(-PatAngle);
          PatLine.x0 /= PatScale;
          PatLine.y0 /= PatScale;
          PatLine.Angle = LineAngle - PatAngle;
          lcHatchPatLine( hHatch, PatLine.Angle, PatLine.x0, PatLine.y0, PatLine.dx, PatLine.dy, PatLine.nDash, 
                          PatLine.Dash[0], PatLine.Dash[1], PatLine.Dash[2], PatLine.Dash[3], 
                          PatLine.Dash[4], PatLine.Dash[5], PatLine.Dash[6], PatLine.Dash[7] );
        }
        lcHatchPatEnd( hHatch );
      }
/*
      // read seed points
      CGePoint* SeedPt = new CGePoint[nSeeds];
      memset( SeedPt, 0, sizeof(CGePoint) * nSeeds );
      for (i=0; i<nSeeds; ++i){
        odpt = pOdHatch->getSeedPointAt( i );
        SeedPt[i].x = odpt.x;
        SeedPt[i].y = odpt.y;
      }
      pHatch->SetSeedPoints( SeedPt, nSeeds );
      delete[] SeedPt;
*/
///      pHatch->Update();
    }
  }
}


//-----------------------------------------------
void Read_Region (OdDbRegionPtr pOdRegion, HANDLE hLcDb, HANDLE hBlock)
{
  int i, n;
  OdRxObjectPtr pOdEntity;
  OdDbLinePtr   pOdLine;
  OdDbArcPtr    pOdArc;
  OdDbCirclePtr pOdCircle;
  OdResult      Ret;
  OdRxObjectPtrArray EntitySet;

  Ret = pOdRegion->explode( EntitySet );
  if (Ret == eCannotExplodeEntity){
    return;
  }
  n = EntitySet.length();
  if (n > 0){
    for (i=0; i<n; ++i){
      pOdEntity = EntitySet.getAt( i );
      if (pOdEntity->isKindOf( OdDbLine::desc() )){      // LINE
        pOdLine = OdDbLine::cast( pOdEntity );
        Read_Line( pOdLine, hLcDb, hBlock );
      }else
      if (pOdEntity->isKindOf( OdDbArc::desc() )){       // ARC
        pOdArc = OdDbArc::cast( pOdEntity );
        Read_Arc( pOdArc, hLcDb, hBlock );
      }else
      if (pOdEntity->isKindOf( OdDbCircle::desc() )){    // CIRCLE
        pOdCircle = OdDbCircle::cast( pOdEntity );
        Read_Circle( pOdCircle, hLcDb, hBlock );
      }
    }
  }

/*
  OdWrFileBuf ow( L"d:\\RegionAcisOut.acis" );
  pOdRegion->acisOut( &ow , kAfTypeASCII );
*/
}


//-----------------------------------------------
// https://forum.opendesign.com/showthread.php?7052-OdDbViewport-transformation-from-Model-Space-to-Paper-Space&highlight=OD_ms2ps
// Thread: OdDbViewport transformation from Model Space to Paper Space
// 
// I use the following code, hope this helps :
// Cheers
// Jason Anderssen
//-----------------------------------------------
void OD_ms2ps (OdDbViewportPtr pVp, OdGeMatrix3d &ResultMat) 
{
  // first get all the data
  OdGeVector3d viewDirection = pVp->viewDirection();
  OdGePoint2d  centre2d = pVp->viewCenter();
  OdGePoint3d  viewCenter = OdGePoint3d( centre2d.x, centre2d.y, 0);
  OdGePoint3d  viewTarget  = pVp->viewTarget ();
  double       twistAngle  = -pVp->twistAngle();
  OdGePoint3d  centerPoint = pVp->centerPoint();
  double       viewHeight  = pVp->viewHeight();
  double       height = pVp->height();
  double       width  = pVp->width();
  double       scaling = viewHeight / height;
  double       lensLength = pVp->lensLength();
  OdGeVector3d xAxis, yAxis, zAxis;
  OdGeMatrix3d dcs2wcs; // display coordinate system (DCS) to world coordinate system (WCS)
  OdGeMatrix3d ps2Dcs;  // paperspace to DCS
  OdGeMatrix3d matCoords;
  OdGeMatrix3d perspMat;

  // prepare the transformation
  zAxis = viewDirection.normal();
  xAxis = OdGeVector3d::kZAxis.crossProduct( viewDirection );

  if (!xAxis.isZeroLength()){
    xAxis.normalize();
    yAxis = zAxis.crossProduct( xAxis );
  }else{
    if (zAxis.z < 0 ){
      xAxis = -OdGeVector3d::kXAxis;
      yAxis = OdGeVector3d::kYAxis;
      zAxis = -OdGeVector3d::kZAxis;
    }else{
      xAxis = OdGeVector3d::kXAxis;
      yAxis = OdGeVector3d::kYAxis;
      zAxis = OdGeVector3d::kZAxis;
    }
  }

  // First initialise with a transformation to centerPoint
  ps2Dcs = OdGeMatrix3d::translation( OdGePoint3d::kOrigin - centerPoint );

  // then scale for the view
  ps2Dcs = ps2Dcs * OdGeMatrix3d::scaling( scaling, centerPoint );

  // then adjust to the viewCenter
  dcs2wcs = OdGeMatrix3d::translation( viewCenter - OdGePoint3d::kOrigin );

  // Then transform for the view direction
  matCoords.setCoordSystem( OdGePoint3d::kOrigin, xAxis, yAxis, zAxis);

  dcs2wcs = matCoords * dcs2wcs;

  // Then adjust for the viewTarget
  dcs2wcs = OdGeMatrix3d::translation( viewTarget - OdGePoint3d::kOrigin ) * dcs2wcs;

  // Then the twist angle
  dcs2wcs = OdGeMatrix3d::rotation(twistAngle, zAxis, viewTarget ) * dcs2wcs;

  if (pVp->isPerspectiveOn()){
    // we do special perspective handling
    double viewSize = viewHeight;
    double aspectRatio = width / height;

    double adjustFactor = 1.0 / 42.0;
    double adjustedLensLength = viewSize * lensLength * sqrt ( 1.0 +
    aspectRatio * aspectRatio ) * adjustFactor;

    double eyeDistance = viewDirection.length();
    double lensDistance = eyeDistance - adjustedLensLength;

    double ed = eyeDistance;
    double ll = adjustedLensLength;
    double l = lensDistance;

    perspMat.entry[2][2] = (ll - l ) / ll;
    perspMat.entry[2][3] = l * ( ed - ll ) / ll;
    perspMat.entry[3][2] = -1.0 / ll;
    perspMat.entry[3][3] = ed / ll;
  }

  ResultMat = ps2Dcs.inverse() * perspMat * dcs2wcs.inverse();
}

// it gets a viewport and returns a transformation matrix,
// with which we can move a paperspace coordinate to the model space
// as looked from the given viewport
//-----------------------------------------------
void OD_ps2ms (OdDbViewportPtr pVp, OdGeMatrix3d &ResultMat) 
{
  // Keep life simple, just invert the other way
  OdGeMatrix3d mat;
  OD_ms2ps( pVp, mat );
  ResultMat = mat.inverse();
}
