/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
* Arc, Circle, Ellipse
*********************************************************************/
#include "pch.h"


//-----------------------------------------------
void Read_Circle (const OdDbCirclePtr pOdCirc, HANDLE hLcDb, HANDLE hBlock)
{
  CGeCircle   Cir;
  CGePolyline Pline;
  CGeVertex*  pVer;
  CGe3dPoint  Norm, CenPt;
  double R = pOdCirc->radius();
  HANDLE hCirc, hPline, hVer;

  CenPt.x = pOdCirc->center().x;
  CenPt.y = pOdCirc->center().y;
  CenPt.z = pOdCirc->center().z;
  Norm.x = pOdCirc->normal().x;
  Norm.y = pOdCirc->normal().y;
  Norm.z = pOdCirc->normal().z;
  if (Norm.IsPlaneTop(0.001) || Norm.IsPlaneBottom(0.001)){
    hCirc = lcBlockAddCircle( hBlock, CenPt.x, CenPt.y, R, false );
    if (hCirc){
      lcPropPutFloat( hCirc, LC_PROP_ENT_Z, CenPt.z );
      lcPropPutFloat( hCirc, LC_PROP_ENT_THICKNESS, pOdCirc->thickness() );
      Read_EntBase( pOdCirc, hCirc, hLcDb );
    }
  }else{
    Cir.Set( 0, 0, R );
    Cir.GetPolyline( &Pline, 64 );
    Pline.Transform( CenPt, Norm );
//    Cir.GetPolyline( &CenPt, &Norm, &Pline, 64 );
    hPline = lcBlockAddPolyline( hBlock, 0, Pline.IsClosed(), false );
    if (hPline){
      Read_EntBase( pOdCirc, hPline, hLcDb );
      pVer = Pline.GetFirstVer();
      while( pVer ){
        hVer = lcPlineAddVer( hPline, NULL, pVer->X(), pVer->Y() );
        lcPropPutFloat( hVer, LC_PROP_VER_Z, pVer->Z() );
        pVer = Pline.GetNextVer( pVer );
      }
    }
  }
/*
  hCirc = lcBlockAddCircle( hBlock, X, Y, R, false );
  if (hCirc){
    lcPropPutFloat( hCirc, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdCirc, hCirc, hLcDb ); //, hBlock );
    if (X!=0. || Y!=0. || Z!= 1.){
//      lcEntPutNormal( hCirc, X, Y, Z );
    }
  }
*/
}

//#define ASD_ANG0 131.925*LC_DEG_TO_RAD
//#define ASD_ANG1 131.98*LC_DEG_TO_RAD
//-----------------------------------------------
void Read_Arc (const OdDbArcPtr pOdArc, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d StartPt, EndPt;
  CGe3dPoint  Norm, CenPt;
  CGeArc      Arc;
  CGePolyline Pline;
  CGeVertex*  pVer;
  double      Ang0, Ang1, AngArc, Thickness;
  double  R = pOdArc->radius();
  HANDLE  hArc, hLine, hCirc, hPline, hVer;
  bool    bAdd = false;

  CenPt.x = pOdArc->center().x;
  CenPt.y = pOdArc->center().y;
  CenPt.z = pOdArc->center().z;
  Norm.x  = pOdArc->normal().x;
  Norm.y  = pOdArc->normal().y;
  Norm.z  = pOdArc->normal().z;
  Thickness = pOdArc->thickness();

  Ang0 = geNormalizeAngle( pOdArc->startAngle() );
  Ang1 = geNormalizeAngle( pOdArc->endAngle() );
  if (Ang1 > Ang0){
    AngArc = Ang1 - Ang0;
  }else{
    AngArc = LC_DEG360 + Ang1 - Ang0;
  }

  if (Norm.IsPlaneTop(0.001) || Norm.IsPlaneBottom(0.001)){
    if (AngArc < 0.01){
      pOdArc->getStartPoint( StartPt );
      pOdArc->getEndPoint( EndPt );
      if (StartPt != EndPt){
        // as a line
        hLine = lcBlockAddLine( hBlock, StartPt.x, StartPt.y, EndPt.x, EndPt.y );
        if (hLine){
          Read_EntBase( pOdArc, hLine, hLcDb ); 
        }
      }else{
        // as a circle
        hCirc = lcBlockAddCircle( hBlock, CenPt.x, CenPt.y, R, false );
        if (hCirc){
          lcPropPutFloat( hCirc, LC_PROP_ENT_Z, CenPt.z );
          lcPropPutFloat( hCirc, LC_PROP_ENT_THICKNESS, Thickness );
          Read_EntBase( pOdArc, hCirc, hLcDb );
        }
      }
      return;
    }

    if (Norm.IsPlaneBottom(0.001)){
      AngArc = -AngArc;
      if (0<=Ang0 && Ang0<=GE_DEG90){
        Ang0 = GE_DEG90 + (GE_DEG90 - Ang0);
      }else{
        if (0<=GE_DEG90 && Ang0<=GE_DEG180){
          Ang0 = GE_DEG180 - Ang0;
        }else{
          if (0<=GE_DEG180 && Ang0<=GE_DEG270){
            Ang0 = GE_DEG270 + (GE_DEG270 - Ang0);
          }else{
            if (0<=GE_DEG270 && Ang0<=GE_DEG360){
              Ang0 = GE_DEG270 - (Ang0 - GE_DEG270);
            }else{
              Ang0 = 0.0;
            }
          }
        }
      }
    }
    hArc = lcBlockAddArc( hBlock, CenPt.x, CenPt.y, R, Ang0, AngArc );
    if (hArc){
      lcPropPutFloat( hArc, LC_PROP_ENT_Z, CenPt.z );
      lcPropPutFloat( hArc, LC_PROP_ENT_THICKNESS, Thickness );
      Read_EntBase( pOdArc, hArc, hLcDb );
    }

  }else{
    Arc.Set( CGePoint(0,0), R, Ang0, AngArc );
    Arc.GetPolylineA( &Pline, 128 ); //64 ); 
    Pline.Transform( CenPt, Norm );
    hPline = lcBlockAddPolyline( hBlock, 0, Pline.IsClosed(), false );
    if (hPline){
      Read_EntBase( pOdArc, hPline, hLcDb );
      pVer = Pline.GetFirstVer();
      while( pVer ){
        hVer = lcPlineAddVer( hPline, NULL, pVer->X(), pVer->Y() );
        lcPropPutFloat( hVer, LC_PROP_VER_Z, pVer->Z() );
        pVer = Pline.GetNextVer( pVer );
      }
    }
  }
}


/********************************************************************
* This function is used to convert AutoCAD ellipse start and end angles
* to VeCAD ellipse angles.
*   Ang0 - AutoCAD ellipse angle
*   Ratio - <radius horisontal>/<radius vertical>
*********************************************************************/
double GetAcadEllipseAngle (double Ang0, double Ratio)
{
  double Angle, dx, dy;
  int    sector;

  dx = cos( Ang0 );
  dy = sin( Ang0 );
  if (dx>0){
    sector = (dy>0)? 0 : 3;
  }else{
    sector = (dy>0)? 1 : 2;
    dx = -dx;
  }
  if (dy<0){
    dy = -dy;
  }

  if (dx!=0 && dy!=0){
    dx *= Ratio;
    Angle = atan( dy/dx );
    switch( sector ){
      case 1: Angle = LC_PI - Angle; break;
      case 2: Angle = LC_PI + Angle; break;
      case 3: Angle = LC_2PI - Angle; break;
    }
  }else{
    Angle = Ang0;
  }
  return Angle;
}

//-----------------------------------------------
void Read_Ellipse (const OdDbEllipsePtr pOdEllipse, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d  Center;
  OdGeVector3d unitNormal;
  OdGeVector3d majorAxis;
  OdGeVector3d minorAxis;
  double       Ratio, radiusRatio;
  double       R1, R2, RotAng, Ang0, Ang1, ArcAng;
  HANDLE       hEll;
  CGePoint     Pt0, Pt1;
  CGeEllipse   GeEll;

  pOdEllipse->get( Center, unitNormal, majorAxis, radiusRatio, Ang0, Ang1 );
  minorAxis = pOdEllipse->minorAxis();
  Pt0.Set( majorAxis.x, majorAxis.y );
  R1 = hypot( Pt0.x, Pt0.y );
  R2 = R1 * radiusRatio;
  Ratio = R1 / R2;
  RotAng = Pt0.Angle();
  if (Ang1 > Ang0){
    ArcAng = Ang1 - Ang0;
  }else{
    ArcAng = LC_2PI + Ang1 - Ang0;
  }
  if (ArcAng < 0.001){
    // add a line instead of ellipse
    Pt0.x = Center.x;
    Pt0.y = Center.y;
    GeEll.Set2( Pt0, R1, R2, RotAng, Ang0, ArcAng );
    Pt0 = GeEll.StartPoint();
    Pt1 = GeEll.EndPoint();
    hEll = lcBlockAddLine( hBlock, Pt0.x, Pt0.y, Pt1.x, Pt1.y );
  }else{
    hEll = lcBlockAddEllipse( hBlock, Center.x, Center.y, R1, R2, RotAng, Ang0, ArcAng );
  }
  if (hEll){
    lcPropPutFloat( hEll, LC_PROP_ENT_Z, Center.z );
    Read_EntBase( pOdEllipse, hEll, hLcDb );
    lcBlockRepEllipse( hBlock, hEll, NULL );
  }
}



