/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
* Dimensions objects
*********************************************************************/
#include "pch.h"

extern COptions g_Options;
extern bool g_bLeader;


//-----------------------------------------------
void Read_DimAlig (const OdDbAlignedDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3;
  OdChar      szIdRefBlock[32];
  HANDLE      hBlkRef, hStyle, hDimAli;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt1 = pOdDim->xLine1Point();
    Pt2 = pOdDim->xLine2Point();
//    Pt3 = pOdDim->dimLinePoint();
    Pt3 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimAli = lcBlockAddDimAli( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, szOverText );
    if (hDimAli){
      Read_EntBase( pOdDim, hDimAli, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimAli, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimRot (const OdDbRotatedDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3;
  OdChar      szIdRefBlock[32];
  double      Ang;
  HANDLE      hBlkRef, hStyle, hDimRot;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt1 = pOdDim->xLine1Point();
    Pt2 = pOdDim->xLine2Point();
//    Pt3 = pOdDim->dimLinePoint();
    Pt3 = pOdDim->textPosition();
    Ang = pOdDim->rotation();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimRot = lcBlockAddDimLin( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, Ang, szOverText );
    if (hDimRot){
      Read_EntBase( pOdDim, hDimRot, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimRot, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimAng3p (const OdDb3PointAngularDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3, PtA;
  OdChar      szIdRefBlock[32];
  double      TextPos=0.5;
  HANDLE      hBlkRef, hStyle, hDimAng;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt1 = pOdDim->centerPoint();
    PtA = pOdDim->arcPoint();
//    R = _hypot( Pt1.x-Pt2.x, Pt1.y-Pt2.y );
    Pt2 = pOdDim->xLine1Point();
    Pt3 = pOdDim->xLine2Point();
//    Pt3 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimAng = lcBlockAddDimAng( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, PtA.x, PtA.y, TextPos, szOverText );
    if (hDimAng){
      Read_EntBase( pOdDim, hDimAng, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimAng, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimAng2p (const OdDb2LineAngularDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3, Pt4, Pt5;
  OdChar      szIdRefBlock[32];
  double      TextPos=0.5;
  HANDLE      hBlkRef, hStyle, hDimAng;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt1 = pOdDim->xLine1Start();
    Pt2 = pOdDim->xLine1End();
    Pt3 = pOdDim->xLine2Start();
    Pt4 = pOdDim->xLine2End();
    Pt5 = pOdDim->arcPoint();
//    Pt3 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimAng = lcBlockAddDimAng2( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, Pt4.x, Pt4.y, 
                                 Pt5.x, Pt5.y, TextPos, szOverText );
    if (hDimAng){
      Read_EntBase( pOdDim, hDimAng, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimAng, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimOrd (const OdDbOrdinateDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2;
  OdChar      szIdRefBlock[32];
  bool        bX;
  HANDLE      hBlkRef, hStyle, hDimOrd;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, (LPCWSTR)szIdRefBlock, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb ); 
    }
  }else{
    bX = pOdDim->isUsingXAxis();
    Pt1 = pOdDim->definingPoint();   // leaderEndPoint()
    Pt2 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimOrd = lcBlockAddDimOrd( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, bX, szOverText );
    if (hDimOrd){
      Read_EntBase( pOdDim, hDimOrd, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimOrd, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimRad (const OdDbRadialDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3;
  OdChar      szIdRefBlock[32];
  HANDLE      hBlkRef, hStyle, hDimRad;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt1 = pOdDim->center();
    Pt2 = pOdDim->chordPoint();
    Pt3 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimRad = lcBlockAddDimRad( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, szOverText );
    if (hDimRad){
      Read_EntBase( pOdDim, hDimRad, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimRad, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimDiam (const OdDbDiametricDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d Pt1, Pt2, Pt3;
  OdChar      szIdRefBlock[32];
  HANDLE      hBlkRef, hStyle, hDimDia;
  WCHAR       szOverText[256];
  LPCWSTR     psz;

  if (g_Options.m_Imp_bDimAsBlock){
    Pt1 = pOdDim->dimBlockPosition();
    pOdDim->dimBlockId().getHandle().getIntoAsciiBuffer( szIdRefBlock );
    psz = (LPCWSTR)szIdRefBlock;
    hBlkRef = lcBlockAddBlockRefIDH( hBlock, psz, Pt1.x, Pt1.y, 1.0, 0.0 );
    if (hBlkRef){
      lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, Pt1.z );
      Read_EntBase( pOdDim, hBlkRef, hLcDb );
    }
  }else{
    Pt3 = pOdDim->farChordPoint();
    Pt2 = pOdDim->chordPoint();
    Pt1.x = (Pt2.x + Pt3.x) / 2,0;
    Pt1.y = (Pt2.y + Pt3.y) / 2,0;
    Pt3 = pOdDim->textPosition();
    memset( szOverText, 0, sizeof(szOverText) );
    wcsncpy( szOverText, (LPCWSTR)pOdDim->dimensionText().c_str(), 250 );
    hDimDia = lcBlockAddDimDia( hBlock, Pt1.x, Pt1.y, Pt2.x, Pt2.y, Pt3.x, Pt3.y, szOverText );
    if (hDimDia){
      Read_EntBase( pOdDim, hDimDia, hLcDb );
      hStyle = DrwGetObjectByHandle( pOdDim->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
      if (hStyle == NULL){
        hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
      }
      // dim style
      lcPropPutHandle( hDimDia, LC_PROP_DIM_STYLE, hStyle );
    }
  }
}

//-----------------------------------------------
void Read_DimAssoc (const OdDbDimAssocPtr pOdDim, HANDLE hLcDb, HANDLE hBlock)
{
  if (g_Options.m_Imp_bDimAsBlock){
  }else{
  }
}

//-----------------------------------------------
void Read_Leader (const OdDbLeaderPtr pOdLeader, HANDLE hLcDb, HANDLE hBlock)
{
  int    nVers = pOdLeader->numVertices();
  double Xt, Yt, Xa, Ya;
  double LandDist = 0.0;
  int    Attach = 0;
  int    Align = LC_LEADER_TCENTER;
  HANDLE hLeader, hStyle;
  WCHAR  szText[512];
  OdDbMTextPtr pOdMText;
  OdDbObjectId OdId;
  OdGePoint3d  OdPnt;

//  pOdLeader->getEndPoint( OdPnt );
//  Xt = OdPnt.x;
//  Yt = OdPnt.y;
  pOdLeader->getStartPoint( OdPnt );
  Xa = OdPnt.x;
  Ya = OdPnt.y;
  OdId = pOdLeader->annotationObjId();
  if (OdId.isNull()){
    szText[0] = 0;
    pOdLeader->getEndPoint( OdPnt );
    Xt = OdPnt.x;
    Yt = OdPnt.y;
  }else{
    pOdMText = OdId.safeOpenObject();
    wcsncpy( szText, (LPCWSTR)pOdMText->contents().c_str(), 510 );
    Xt = pOdMText->location().x + pOdMText->actualWidth()/2.0;
    Yt = pOdMText->location().y + pOdMText->actualHeight()/2.0;
    g_bLeader = true; // used to catch following MTEXT object
  }
  Attach = LC_LEADER_CORNER;
  Align = LC_LEADER_TLEFT;
  hLeader = lcBlockAddLeader( hBlock, szText, Xt, Yt, LandDist, Xa, Ya, Attach, Align );
  if (hLeader){
    Read_EntBase( pOdLeader, hLeader, hLcDb );
    hStyle = DrwGetObjectByHandle( pOdLeader->dimensionStyle(), hLcDb, LC_OBJ_DIMSTYLE );
    if (hStyle == NULL){
      hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );  // "Standard" dim style
    }
    // dim style
    lcPropPutHandle( hLeader, LC_PROP_LEADER_STYLE, hStyle );
    // landing distance
    LandDist = lcPropGetFloat( hStyle, LC_PROP_DIMST_ASZ );
    lcPropPutFloat( hLeader, LC_PROP_LEADER_LDIST, LandDist );
  }else{
    g_bLeader = false;
  }
}

//-----------------------------------------------
void Read_MLeader (const OdDbMLeaderPtr pOdMLeader, HANDLE hLcDb, HANDLE hBlock)
{
}




