/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
* Text objects
*********************************************************************/
#include "pch.h"

bool g_bLeader=false;

//-----------------------------------------------
void Read_Text (const OdDbTextPtr pOdText, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE     hStyle, hText;
  CGe3dPoint Pt0, Pta;
  double     X, Y, Z, ang;
  int        AlignHor, AlignVer;
  int        Align = LC_TA_LEFBOT;
  WCHAR      szText[512];
  OdGeVector3d Norm = pOdText->normal();

  memset( szText, 0, sizeof(szText) );
  wcsncpy( szText, (LPCWSTR)pOdText->textString().c_str(), 510 );
  if (szText[0] == 0){
    // don't add empty text entity
    return;
  }
  hStyle = DrwGetObjectByHandle( pOdText->textStyle(), hLcDb, LC_OBJ_TEXTSTYLE );
  if (hStyle == NULL){
    hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );  // "Standard" text style
  }
  hText = lcBlockAddText( hBlock, szText, 0, 0 );
  if (hText){
    Read_EntBase( pOdText, hText, hLcDb ); //, hBlock );
//    Zaxis.Set( Norm.x, Norm.y, Norm.z );
//    Zaxis.GetArbitraryAxes( &Xaxis, &Yaxis );
//    Pt3d.Set( pOdText->position().x, pOdText->position().y, pOdText->position().z );
//    Pt3d.Transform( Xaxis, Yaxis, Zaxis );
//    pt0 = Pt3d;
    Pt0.Set( pOdText->position().x, pOdText->position().y, pOdText->position().z );
//    Pt3d.Set( pOdText->alignmentPoint().x, pOdText->alignmentPoint().y, pOdText->alignmentPoint().z );
//    Pt3d.Transform( Xaxis, Yaxis, Zaxis );
//    pta = Pt3d;
    Pta.Set( pOdText->alignmentPoint().x, pOdText->alignmentPoint().y, pOdText->alignmentPoint().z );
    // style
    lcPropPutHandle( hText, LC_PROP_TEXT_STYLE, hStyle );
    // width factor
    lcPropPutFloat( hText, LC_PROP_TEXT_WSCALE, pOdText->widthFactor() );
    // height and other parameters
    lcPropPutFloat( hText, LC_PROP_TEXT_H, pOdText->height() );
    lcPropPutBool(  hText, LC_PROP_TEXT_BACKWARD, pOdText->isMirroredInX() );
    lcPropPutBool(  hText, LC_PROP_TEXT_UPDOWN, pOdText->isMirroredInY() ); 
    ang = pOdText->rotation();
    lcPropPutFloat( hText, LC_PROP_TEXT_ANGLE, ang );
    ang = pOdText->oblique();
    if (ang > GE_DEG270){
      ang -= GE_DEG360; 
    }
    lcPropPutFloat( hText, LC_PROP_TEXT_OBLIQUE, ang );
    AlignHor = pOdText->horizontalMode();
    AlignVer = pOdText->verticalMode();
    switch( AlignHor ){   // horizontal alignment, 0=Left;  1=Center; 2=Right; 3=Align; 4=Middle; 5=Fit
      case 0:
        switch( AlignVer ){ // vertical alignment, 0 = Baseline; 1 = Bottom; 2 = Middle; 3 = Top
          case 0: 
          case 1: Align = LC_TA_LEFBOT; break;
          case 2: Align = LC_TA_LEFCEN; break;
          case 3: Align = LC_TA_LEFTOP; break;
        }
        break;
      case 1:
        switch( AlignVer ){
          case 0: 
          case 1: Align = LC_TA_CENBOT; break;
          case 2: Align = LC_TA_CENTER; break;
          case 3: Align = LC_TA_CENTOP; break;
        }
        break;
      case 2:
        switch( AlignVer ){
          case 0: 
          case 1: Align = LC_TA_RIGBOT; break;
          case 2: Align = LC_TA_RIGCEN; break;
          case 3: Align = LC_TA_RIGTOP; break;
        }
        break;
      case 3:  Align = LC_TA_ALIGNED;  break;
      case 4:  Align = LC_TA_CENTER;  break;
      case 5:  Align = LC_TA_FIT;  break;
    }
    if (Align == LC_TA_LEFBOT){
      lcPropPutFloat( hText, LC_PROP_TEXT_X, Pt0.x );
      lcPropPutFloat( hText, LC_PROP_TEXT_Y, Pt0.y );
      lcPropPutFloat( hText, LC_PROP_ENT_Z, Pt0.z );
    }else{
      if (Align == LC_TA_ALIGNED || Align == LC_TA_FIT){
        lcPropPutFloat( hText, LC_PROP_TEXT_X, Pt0.x );
        lcPropPutFloat( hText, LC_PROP_TEXT_Y, Pt0.y );
        lcPropPutFloat( hText, LC_PROP_ENT_Z, Pt0.z );
        lcPropPutFloat( hText, LC_PROP_TEXT_XFIT, Pta.x );
        lcPropPutFloat( hText, LC_PROP_TEXT_YFIT, Pta.y );
      }else{
        lcPropPutFloat( hText, LC_PROP_TEXT_X, Pta.x );
        lcPropPutFloat( hText, LC_PROP_TEXT_Y, Pta.y );
        lcPropPutFloat( hText, LC_PROP_ENT_Z, Pta.z );
/*
        if (pta.IsEqual( pt0 ) == false){
          ang = pt0.Angle( pta, GE_PLANE_TOP );
          lcPropPutFloat( hText, LC_PROP_TEXT_ANGLE, ang );
        }
*/
      }
      lcPropPutInt( hText, LC_PROP_TEXT_ALIGN, Align );
    }
    X = pOdText->normal().x;
    Y = pOdText->normal().y;
    Z = pOdText->normal().z;
    if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//      lcEntPutNormal( hText, X, Y, Z );
    }
    lcPropPutFloat( hText, LC_PROP_ENT_THICKNESS, pOdText->thickness() );
  }
}

//-----------------------------------------------
void Read_MText (const OdDbMTextPtr pOdMText, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE  hStyle, hMText;
  double  Angle, X, Y, Z, H, WrapWidth;
  int     Align = LC_TA_LEFBOT;
  WCHAR   szText[32];
  OdDbMText::AttachmentPoint At;

  if (g_bLeader){
    // this MText object is a part of Leader object
    return;
  }

  memset( szText, 0, sizeof(szText) );
  wcsncpy( szText, (LPCWSTR)pOdMText->contents().c_str(), 30 );
  if (szText[0] == 0){
    // don't add empty text entity
    return;
  }
  hStyle = DrwGetObjectByHandle( pOdMText->textStyle(), hLcDb, LC_OBJ_TEXTSTYLE );
  if (hStyle == NULL){
    hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );  // "Standard" text style
  }
  X = pOdMText->location().x;
  Y = pOdMText->location().y;
  Z = pOdMText->location().z;
  H = pOdMText->textHeight();
  Angle = pOdMText->rotation();
  WrapWidth = pOdMText->width();
  hMText = lcBlockAddMText( hBlock, (LPCWSTR)pOdMText->contents().c_str(), X,Y, WrapWidth, Align, Angle, H, 1.0 );
  if (hMText){
    lcPropPutFloat( hMText, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdMText, hMText, hLcDb ); //, hBlock );
    lcPropPutHandle( hMText, LC_PROP_MTEXT_STYLE, hStyle );
    lcPropPutFloat( hMText, LC_PROP_MTEXT_LINESPACE, pOdMText->lineSpacingFactor() );
    At = pOdMText->attachment();
    switch( At ){
      case OdDbMText::kTopLeft:      Align = LC_TA_LEFTOP; break;
      case OdDbMText::kTopCenter:    Align = LC_TA_CENTOP; break;
      case OdDbMText::kTopRight:     Align = LC_TA_RIGTOP; break;
      case OdDbMText::kMiddleLeft:   Align = LC_TA_LEFCEN; break;
      case OdDbMText::kMiddleCenter: Align = LC_TA_CENTER; break;
      case OdDbMText::kMiddleRight:  Align = LC_TA_RIGCEN; break;
      case OdDbMText::kBottomLeft:   Align = LC_TA_LEFBOT; break;
      case OdDbMText::kBottomCenter: Align = LC_TA_CENBOT; break;
      case OdDbMText::kBottomRight:  Align = LC_TA_RIGBOT; break;
    }
    lcPropPutInt( hMText, LC_PROP_MTEXT_ALIGN, Align );
    X = pOdMText->normal().x;
    Y = pOdMText->normal().y;
    Z = pOdMText->normal().z;
    if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//      lcEntPutNormal( hMText, X, Y, Z );
    }
  }
}

//-----------------------------------------------
void Read_ArcText (const OdDbArcAlignedTextPtr pOdArcText, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d odpt;
  HANDLE     hStyle, hText;
  double     X, Y, Z, R, H, Ang0, WScale;
  int        Align = LC_ATA_LEFT;
  bool       bCW;
  WCHAR      szText[512];

  memset( szText, 0, sizeof(szText) );
  wcsncpy( szText, (LPCWSTR)pOdArcText->textString().c_str(), 510 );
  if (szText[0] == 0){
    // don't add empty text entity
    return;
  }
  hStyle = DrwGetObjectByHandle( pOdArcText->textStyle(), hLcDb, LC_OBJ_TEXTSTYLE );
  if (hStyle == NULL){
    hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );  // "Standard" text style
  }
  odpt = pOdArcText->center();
  X = odpt.x;
  Y = odpt.y;
  Z = odpt.z;
  R = pOdArcText->radius() + pOdArcText->offsetFromArc();
  if (R <= 0.0){
    return;
  }
  if (pOdArcText->textDirection() == kOutwardFromCenter){
    bCW = true;
    Ang0 = pOdArcText->endAngle();
  }else{
    bCW = false;
    Ang0 = pOdArcText->startAngle();
  }
  H = pOdArcText->textSize();
  WScale = pOdArcText->xScale();
  hText = lcBlockAddArcText( hBlock, szText, X, Y, R, Ang0, bCW, H, WScale, Align );
  if (hText){
    Read_EntBase( pOdArcText, hText, hLcDb ); //, hBlock );
    lcPropPutHandle( hText, LC_PROP_ATEXT_STYLE, hStyle );
  }
}

