/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
*********************************************************************/
#ifndef _OK_BASETYPES_
#define _OK_BASETYPES_

#if defined(_WIN64)
    typedef __int64 INT_PTR;
    typedef unsigned __int64 UINT_PTR;
    typedef __int64 LONG_PTR;
    typedef unsigned __int64 ULONG_PTR;
#else
    typedef int INT_PTR;
    typedef unsigned int UINT_PTR;
    typedef long LONG_PTR;
    typedef unsigned long ULONG_PTR;
#endif

typedef unsigned char    UCHAR;
typedef unsigned char    BYTE;
typedef wchar_t          WCHAR;
#ifndef _WINDOWS_
  typedef const char*      LPCSTR;
  typedef const wchar_t*   LPCWSTR;
#endif

typedef unsigned short   WORD;
typedef unsigned short   USHORT;

typedef int              BOOL;
typedef int              INT32;
typedef int              INT;
typedef long             LONG;
typedef unsigned int     UINT32;
typedef unsigned int     UINT;
typedef unsigned long    DWORD;
typedef unsigned long    COLORREF;

typedef __int64          INT64;
typedef unsigned __int64 UINT64;

typedef UINT_PTR         WPARAM;
typedef LONG_PTR         LPARAM;
typedef LONG_PTR         LRESULT;
typedef ULONG_PTR        DWORD_PTR;
typedef void*            HANDLE;

struct WPOINT
{
  LONG  x;
  LONG  y;
};


#define LOWORD(l)      ((WORD)(((DWORD_PTR)(l)) & 0xffff))
#define HIWORD(l)      ((WORD)((((DWORD_PTR)(l)) >> 16) & 0xffff))
#define LOBYTE(w)      ((BYTE)(((DWORD_PTR)(w)) & 0xff))
#define HIBYTE(w)      ((BYTE)((((DWORD_PTR)(w)) >> 8) & 0xff))

#define RGB(r,g,b)     ((COLORREF)(((BYTE)(r)|((WORD)((BYTE)(g))<<8))|(((DWORD)(BYTE)(b))<<16)))
#define GetRValue(rgb) (LOBYTE(rgb))
#define GetGValue(rgb) (LOBYTE(((WORD)(rgb)) >> 8))
#define GetBValue(rgb) (LOBYTE((rgb)>>16))


#define NULL  0
#define CALLBACK __stdcall

// DWG Entity Types
#define DET_LINE       1
#define DET_XLINE      2   
#define DET_RAY        3  
#define DET_LWPLINE    4 
#define DET_3DPLINE    5
#define DET_2DPLINE    6   
#define DET_SPLINE     7   
#define DET_CIRCLE     8
#define DET_ARC        9
#define DET_ELLIPSE    10
#define DET_BLOCKREF   11
#define DET_TEXT       12   
#define DET_MTEXT      13  
#define DET_ARCTEXT    14   // ToDo
#define DET_ATTDEF     15   
#define DET_POINT      16   // ToDo
#define DET_DIMALIG    17   // ToDo
#define DET_DIMROT     18   // ToDo
#define DET_DIMANG3P   19   // ToDo
#define DET_DIMANG2P   20   // ToDo
#define DET_DIMRAD     21   // ToDo
#define DET_DIMDIAM    22   // ToDo
#define DET_DIMORD     23   // ToDo
#define DET_LEADER     24   // ToDo
#define DET_HATCH      25   // ToDo
#define DET_IMAGE      26   // ToDo
#define DET_WIPEOUT    27   // ToDo
#define DET_SHAPE      28   // ToDo
#define DET_MLINE      29   // ToDo
#define DET_SOLID      30   // ToDo
#define DET_TRACE      31   // ToDo
#define DET_VIEWPORT   32   // ToDo
#define DET_REGION     33   // ToDo
#define DET_3DFACE     34   // ToDo
#define DET_3DSOLID    35   // ToDo

#define DWG_BLKATT_YES  30591
#define DWG_BLKATT_NO   30592

#endif // _OK_BASETYPES_
