/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2023, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
*********************************************************************/
#include "pch.h"

extern COptions g_Options;

void Write_Header     (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_Linetypes  (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_Layers     (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_TextStyles (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_ImageDefs  (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_DimStyles  (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_MlineStyles(HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_BlockDefs  (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void Write_Entities   (HANDLE hLcDb, OdDbDatabasePtr pOdDb);
void SetCurrentLayout (HANDLE hLcDb, OdDbDatabasePtr pOdDb);

void Write_Point    (HANDLE hPoint, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Line     (HANDLE hLine, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Xline    (HANDLE hXline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Pline    (HANDLE hPline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Rect     (HANDLE hRect, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Arc      (HANDLE hArc, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Circle   (HANDLE hCirc, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Ellipse  (HANDLE hEll, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Text     (HANDLE hText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_TextWin  (HANDLE hText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_MText    (HANDLE hMText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_ArcText  (HANDLE hAText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_BlockRef (HANDLE hIns, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_ImageRef (HANDLE hImageRef, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Viewport (HANDLE hVport, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Hatch    (HANDLE hHatch, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_MLine    (HANDLE hMLine, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimLin   (HANDLE hDimRot, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimAli   (HANDLE hDimAli, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimOrd   (HANDLE hDimOrd, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimRad   (HANDLE hDimRad, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimDia   (HANDLE hDimDia, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_DimAng   (HANDLE hDimAng, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Leader   (HANDLE hLeader, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_RPlan    (HANDLE hRPlan, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Attdef   (HANDLE hAttdef, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Face3D   (HANDLE hFace, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Shape    (HANDLE hShape, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
void Write_Exploded (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);

//void ExportEntities (HANDLE  hLcDb, OdDbDatabasePtr pOdDb);

void CorrectTableName (WCHAR* szName);


//=============================================================================
bool WriteFile (LPCWSTR szFileName, const BYTE* ImgBuffer, int ImgSize, HANDLE hLcDb, WCHAR* szError)
{
  WCHAR szBuf[256], szExt[16];
  bool  bRet = false;
//  int   AngUnits = lcPropGetInt( 0, LC_PROP_G_ANGLE );
  OdDb::SaveType   FileType;
  OdDb::DwgVersion OutVer = (OdDb::DwgVersion)g_Options.m_Exp_FileVer;
  int DxfPrecision = g_Options.m_Exp_DxfPrec;

  memset( szBuf, 0, sizeof(szBuf) );
  memset( szExt, 0, sizeof(szExt) );
  okGetFileExt( szFileName, szExt );
  if (wcsicmp( szExt, L"dxf" ) == 0){
    FileType = OdDb::kDxf;
  }else
  if (wcsicmp( szExt, L"dwg" ) == 0){
    FileType = OdDb::kDwg;
  }else
  if (wcsicmp( szExt, L"dxb" ) == 0){
    FileType = OdDb::kDxb;
  }else
    return false;  // unknown file type

  OdStaticRxObject<MyServices> svcs;

  svcs.disableOutput( false );
  svcs.setRecomputeDimBlocksRequired( false );
  try 
  {
    odInitialize( &svcs );
    OdDbDatabasePtr  pDb = svcs.createDatabase( true, OdDb::kMetric );

    OdDbDatabaseSummaryInfoPtr pSumInfo = oddbGetSummaryInfo( pDb );
    wcscpy( szBuf, L"File was created by LiteCAD plugin " );
    wcscat( szBuf, lcPropGetStr( 0, LC_PROP_G_VERSION ) );
    pSumInfo->setComments(OdString(szBuf));
    oddbPutSummaryInfo(pSumInfo);
    memset( szBuf, 0, sizeof(szBuf) );

//    lcPropPutInt( 0, LC_PROP_G_ANGLE, LC_ANGLE_RADIAN );
    Write_Header     (hLcDb, pDb);
    Write_TextStyles (hLcDb, pDb);
    Write_Linetypes  (hLcDb, pDb);
    Write_Layers     (hLcDb, pDb);
    Write_ImageDefs  (hLcDb, pDb);
    Write_DimStyles  (hLcDb, pDb);
    Write_MlineStyles(hLcDb, pDb);
    Write_BlockDefs  (hLcDb, pDb);
    Write_Entities   (hLcDb, pDb);
    SetCurrentLayout (hLcDb, pDb);

    if (ImgSize > 0 && ImgBuffer){
      // Write the database with preview image
      BITMAPINFO* bi = (BITMAPINFO*)(ImgBuffer+14);
      pDb->setThumbnailBitmap( bi, ImgSize-14 );
      pDb->writeFile( szFileName, FileType, OutVer, true, DxfPrecision );
    }else{
      // Write the database without preview image
      pDb->writeFile( szFileName, FileType, OutVer, false, DxfPrecision );
    }
    bRet = true;
  }
  // Display the error
  catch (OdError& e)
  {
    wcscpy( szBuf, (WCHAR*)svcs.getErrorDescription(e.code()).c_str() );
  }
  catch (...)
  {
    wcscpy( szBuf, L"Unknown Error." );
  }
  if (szError && szBuf[0]){
    wcscat( szError, szBuf );
  }

  // Uninitialize Windows NT Encription/Decription
//  ::odrxUninitWinNTCrypt();

  odUninitialize();
  return bRet;
}


//-----------------------------------------------
void Write_Header (HANDLE hLcDb, OdDbDatabasePtr pDb)
{
  BOOL b;
//  int  ival;
//  double dval;
  // Set Creation and Last Update times 
  OdDbDate date;
  date.getUniversalTime();
  odDbSetTDUCREATE( *pDb, date );
  odDbSetTDUUPDATE( *pDb, date );
  b = false; //lcPropGetBool( hLcDb, LC_PROP_DRW_LWDISPLAY );
  pDb->setLWDISPLAY( b?true:false );
/*
  ival = lcPropGetInt( hLcDb, LC_PROP_DRW_PDMODE );
  pDb->setPDMODE( ival );
  dval = lcPropGetFloat( hLcDb, LC_PROP_DRW_PDSIZE );
  pDb->setPDSIZE( dval );
  dval = lcPropGetFloat( hLcDb, LC_PROP_DRW_LTSCALE );
  pDb->setLTSCALE( dval );
*/
//  pDb->setATTMODE( pDwgHdr->attmode );
//  pDb->setCMLJUST( pDwgHdr->cmljust );
//  pDb->setCMLSCALE( pDwgHdr->cmlscale );
//  pLcDb->SetUnits( pDb->getINSUNITS() );
}


//-----------------------------------------------
void Write_TextStyles (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szBuf[256];
  double  Height, WScale, Oblique;
  BOOL    bStd, bShapes, bDeleted, bUpdown, bBack, bTTF;
  HANDLE  hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );
  OdChar                      szOdBuf[32];
  OdDbTextStyleTableRecordPtr pOdStyle;
  OdDbTextStyleTablePtr       pOdStyles = pOdDb->getTextStyleTableId().safeOpenObject( OdDb::kForWrite );
  while( hStyle ){
    bDeleted = lcPropGetBool( hStyle, LC_PROP_TSTYLE_DELETED );
    if (bDeleted == false){
      bStd    = lcPropGetBool( hStyle, LC_PROP_TSTYLE_STANDARD );
      bShapes = lcPropGetBool( hStyle, LC_PROP_TSTYLE_SHAPES );
      if (bStd){
        pOdStyle = pOdDb->getTextStyleStandardId().safeOpenObject( OdDb::kForWrite );
      }else{
        pOdStyle = OdDbTextStyleTableRecord::createObject();
        if (bShapes){
          pOdStyle->setIsShapeFile( true );
        }else{
          pOdStyle->setIsShapeFile( false );
          wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_TSTYLE_NAME ) );
          CorrectTableName( szBuf );
          pOdStyle->setName( szBuf );
        }
        pOdStyles->add( pOdStyle );
      }
      pOdStyle->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
      lcPropPutStr( hStyle, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );

      Height  = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_HEIGHT );
      WScale  = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_WSCALE );
      Oblique = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_OBLIQUE );
      bUpdown = lcPropGetBool( hStyle, LC_PROP_TSTYLE_UPDOWN );
      bBack   = lcPropGetBool( hStyle, LC_PROP_TSTYLE_BACKWARD );
      wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_TSTYLE_FONT ) );
      bTTF    = lcPropGetBool( hStyle, LC_PROP_TSTYLE_WINFONT );

      pOdStyle->setTextSize( Height );
//      pOdStyle->setPriorSize( Height ); 
      pOdStyle->setXScale( WScale );
      pOdStyle->setObliquingAngle( Oblique );
      pOdStyle->setIsBackwards( bBack? true : false );
      pOdStyle->setIsUpsideDown( bUpdown? true : false );
      if (bTTF){
        pOdStyle->setFont( szBuf, false, false, 0, 0 );
      }else{
        wcscat( szBuf, L".shx" );
        pOdStyle->setFileName( szBuf );
/*
        hFont = lcPropGetHandle( hStyle, LC_PROP_TSTYLE_HFONT );
        if (hFont){
          n = lcPropGetInt( hFont, LC_PROP_FONT_NCHARS );
          if (n > 1000){
            pOdStyle->setBigFontFileName( szBuf );
          }
        }
*/
      }
    }
    hStyle = lcDrwGetNextObject( hLcDb, hStyle );
  }
}


//-----------------------------------------------
void Write_Linetypes (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szBuf[256], szText[128], szFontName[256], szOdDbHandle[32];
  BOOL    bStd, bDeleted, bContinuous, bByLayer, bByBlock, bShape, bAbsAng;
  double  LtScale, PatLen, DashLen, Angle, Scale;
  int     i, nElems;
  HANDLE  hTStyle;
  HANDLE  hLtype = lcDrwGetFirstObject( hLcDb, LC_OBJ_LINETYPE );
  OdGeVector2d                Offset;
  OdChar                      szOdBuf[32];
  OdDbHandle                  OdHandle;
  OdDbTextStyleTableRecordPtr pOdTStyle;
  OdDbTextStyleTablePtr       pOdTStyles = pOdDb->getTextStyleTableId().safeOpenObject( OdDb::kForWrite );
  OdDbSymbolTableIteratorPtr  pIter;
  OdDbLinetypeTableRecordPtr  pOdLtype;
  OdDbLinetypeTablePtr        pOdLtypes = pOdDb->getLinetypeTableId().safeOpenObject( OdDb::kForWrite );
  while( hLtype ){
    bDeleted = lcPropGetBool( hLtype, LC_PROP_LINETYPE_DELETED );
    if (bDeleted == false){
      bContinuous = lcPropGetBool( hLtype, LC_PROP_LINETYPE_CONTINUOUS );
      bByLayer    = lcPropGetBool( hLtype, LC_PROP_LINETYPE_BYLAYER );
      bByBlock    = lcPropGetBool( hLtype, LC_PROP_LINETYPE_BYBLOCK );
      bStd = false;
      if (bContinuous){
        pOdLtype = pOdDb->getLinetypeContinuousId().safeOpenObject( OdDb::kForWrite );
        bStd = true;
      }else{
        if (bByLayer){
          pOdLtype = pOdDb->getLinetypeByLayerId().safeOpenObject( OdDb::kForWrite );
          bStd = true;
        }else{
          if (bByBlock){
            pOdLtype = pOdDb->getLinetypeByBlockId().safeOpenObject( OdDb::kForWrite );
            bStd = true;
          }else{
            pOdLtype = OdDbLinetypeTableRecord::createObject();
          }
        }
      }
      if (!bStd){
        wcscpy( szBuf, lcPropGetStr( hLtype, LC_PROP_LINETYPE_NAME ) );
        CorrectTableName( szBuf );
        pOdLtype->setName( szBuf );
        pOdLtypes->add( pOdLtype );
      }
      pOdLtype->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
      lcPropPutStr( hLtype, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );

      wcscpy( szBuf, lcPropGetStr( hLtype, LC_PROP_LINETYPE_DESC ) );
      pOdLtype->setComments( szBuf );
      LtScale = lcPropGetFloat( hLtype, LC_PROP_LINETYPE_SCALE );
      PatLen  = lcPropGetFloat( hLtype, LC_PROP_LINETYPE_PATLEN );
      nElems = lcPropGetInt( hLtype, LC_PROP_LINETYPE_NELEM );
      if (nElems >= 2){
        pOdLtype->setNumDashes( nElems );
        pOdLtype->setPatternLength( PatLen * LtScale );
        for (i=0; i<nElems; i++){
          lcPropPutInt( hLtype, LC_PROP_LINETYPE_IELEM, i ); 

          DashLen  = lcPropGetFloat( hLtype, LC_PROP_LTELEM_LEN );
          bShape   = lcPropGetBool( hLtype, LC_PROP_LTELEM_SHAPE );
          wcscpy( szText, lcPropGetStr( hLtype, LC_PROP_LTELEM_TEXT ));
          Scale    = lcPropGetFloat( hLtype, LC_PROP_LTELEM_SCALE );
          Angle    = lcPropGetFloat( hLtype, LC_PROP_LTELEM_ANGLE );
          bAbsAng  = lcPropGetBool( hLtype, LC_PROP_LTELEM_ABSANGLE );
          Offset.x = lcPropGetFloat( hLtype, LC_PROP_LTELEM_X );
          Offset.y = lcPropGetFloat( hLtype, LC_PROP_LTELEM_Y );

          pOdLtype->setDashLengthAt( i, DashLen * LtScale );
          if (szText[0]){
            if (bShape){
              // single shape and font filename
              wcscpy( szFontName, lcPropGetStr( hLtype, LC_PROP_LTELEM_FONTNAME ));
              pOdLtype->setShapeNumberAt ( i, szText[0] );
              // look the shape filename in exist styles
              pIter = pOdTStyles->newIterator();
              for (pIter->start(); !pIter->done(); pIter->step()){
                pOdTStyle = pIter->getRecordId().safeOpenObject();
                if (pOdTStyle->isShapeFile()){
                  wcscpy( szBuf, pOdTStyle->fileName() );
                  if (wcsicmp( szBuf, szFontName ) == 0){
                    // the style already exist
                    pOdLtype->setShapeStyleAt( i, pOdTStyle->objectId() );
                    goto m1;
                  }
                }
              }
              // the style don't exist, create it!
              pOdTStyle = OdDbTextStyleTableRecord::createObject();
              pOdTStyle->setName( "" );
              pOdTStyle->setTextSize( 0.0 );
              pOdTStyle->setIsShapeFile( true );
              pOdTStyles->add( pOdTStyle );
              pOdTStyle->setFileName( szFontName );
              pOdLtype->setShapeStyleAt( i, pOdTStyle->objectId() ); 
            }else{
              // text and text style
              pOdLtype->setTextAt( i, szText );
              hTStyle = lcPropGetHandle( hLtype, LC_PROP_LTELEM_STYLE );
              wcscpy( szOdDbHandle, lcPropGetStr( hTStyle, LC_PROP_TABLE_ODA_HANDLE ) );
              OdHandle = OdDbHandle( szOdDbHandle );
              pOdLtype->setShapeStyleAt( i, pOdDb->getOdDbObjectId( OdHandle ) );
            }
m1:
            pOdLtype->setShapeScaleAt( i, Scale );
            pOdLtype->setShapeRotationAt( i, Angle );
            pOdLtype->setShapeIsUcsOrientedAt( i, bAbsAng?true:false );
            pOdLtype->setShapeOffsetAt( i, Offset );
          }
        }
      }
    }
    hLtype = lcDrwGetNextObject( hLcDb, hLtype );
  }
}


//-----------------------------------------------
void Write_Layers (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR  szBuf[256], szOdDbHandle[32];
  BOOL   bStd, bDeleted, bLocked, bVisible, bNoPrint, bNoExport;
  HANDLE hLtype;
  HANDLE hLayer = lcDrwGetFirstObject( hLcDb, LC_OBJ_LAYER );
  int    LwLcd;
  OdDb::LineWeight        OdLW;
  OdCmColor               OdColor;
  OdChar                  szOdBuf[32];
  OdDbHandle              OdHandle;
  OdDbLayerTableRecordPtr pOdLayer;
  OdDbLayerTablePtr       pOdLayers = pOdDb->getLayerTableId().safeOpenObject( OdDb::kForWrite );
  while( hLayer ){
    bDeleted = lcPropGetBool( hLayer, LC_PROP_LAYER_DELETED );
    if (bDeleted == false){
      bStd = lcPropGetBool( hLayer, LC_PROP_LAYER_0 );
      if (bStd){
        pOdLayer = pOdDb->getLayerZeroId().safeOpenObject( OdDb::kForWrite );
      }else{
        bNoExport = lcPropGetBool( hLayer, LC_PROP_LAYER_NOEXPORT );
        if (bNoExport){
          goto mNextLayer;
        }
        pOdLayer = OdDbLayerTableRecord::createObject();
        wcscpy( szBuf, lcPropGetStr( hLayer, LC_PROP_LAYER_NAME ) );
        okDelLastBlanks( szBuf );
        CorrectTableName( szBuf );
        pOdLayer->setName( szBuf );
        pOdLayers->add( pOdLayer );
      }
      pOdLayer->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
      lcPropPutStr( hLayer, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );

      bLocked  = lcPropGetBool( hLayer, LC_PROP_LAYER_LOCKED );
      bVisible = lcPropGetBool( hLayer, LC_PROP_LAYER_VISIBLE );
      bNoPrint = lcPropGetBool( hLayer, LC_PROP_LAYER_NOPRINT );
      LwLcd  = lcPropGetInt( hLayer, LC_PROP_LAYER_LWEIGHT );
      OdLW = LWeight_LcdToDwg( LwLcd );
      wcscpy( szBuf, lcPropGetStr( hLayer, LC_PROP_LAYER_COLOR ));
      MakeOdColor( szBuf, &OdColor );
/*
      if (lcColorIsRGB( szBuf )){
        // Color specified by RGB value
        OdColor.setColorMethod( OdCmEntityColor::kByColor );  
        OdColor.setRed(   lcColorGetRed(szBuf) );
        OdColor.setGreen( lcColorGetGreen(szBuf) );
        OdColor.setBlue(  lcColorGetBlue(szBuf) );
      }else{
        // Color specified by an index (ACI) into a color table.
        OdColor.setColorMethod( OdCmEntityColor::kByACI );    
        OdColor.setColorIndex( lcColorGetIndex(szBuf) );
      }
*/

      pOdLayer->setIsLocked( bLocked? true : false );
      pOdLayer->setIsOff( bVisible? false : true );
      pOdLayer->setIsPlottable( bNoPrint? false : true );
      pOdLayer->setLineWeight( OdLW );
      pOdLayer->setColor( OdColor );
      hLtype = lcPropGetHandle( hLayer, LC_PROP_LAYER_LINETYPE );
      wcscpy( szOdDbHandle, lcPropGetStr( hLtype, LC_PROP_TABLE_ODA_HANDLE ));
      OdHandle = OdDbHandle( szOdDbHandle );
      pOdLayer->setLinetypeObjectId( pOdDb->getOdDbObjectId( OdHandle ) );
    }
mNextLayer:
    hLayer = lcDrwGetNextObject( hLcDb, hLayer );
  }
//    ::MessageBox( ::GetActiveWindow(), L"Layers OK", L"DWG", MB_OK );
}


//-----------------------------------------------
void Write_ImageDefs (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR  szBuf[256], szFileName[256]; //, szOdDbHandle[32];
  BOOL   bDeleted;
  BOOL   bFirst = true;
  HANDLE hImage = lcDrwGetFirstObject( hLcDb, LC_OBJ_IMAGE );
  int    W, H;
  OdChar                szOdBuf[32];
  OdDbObjectId          imageDictId;
  OdDbDictionaryPtr     pImageDict;
  OdDbRasterImageDefPtr pOdImageDef;
  OdDbObjectId          imageDefId;
  OdGiRasterImagePtr    pImageDesc;

  while( hImage ){
    bDeleted = lcPropGetBool( hImage, LC_PROP_IMAGE_DELETED );
    if (bDeleted == false){
      if (bFirst){
        bFirst = false;
        // create image dictionary
        imageDictId = OdDbRasterImageDef::createImageDictionary( pOdDb );
        // Open the image dictionary.
        pImageDict = imageDictId.safeOpenObject( OdDb::kForWrite );
      }
      // Add imagedef object, which is similar to a block definition in that
      // it defines the associated raster file and attributes for this image.
      pOdImageDef = OdDbRasterImageDef::createObject();
      // Add imagedef to image dictionary.
      wcscpy( szFileName, lcPropGetStr( hImage, LC_PROP_IMAGE_FILENAME ) );
      wcscpy( szBuf, okGetFileName( szFileName, false ));
      imageDefId = pImageDict->setAt( szBuf, pOdImageDef );
      // Set remaining imagedef attributes.
      pOdImageDef->setSourceFileName( szFileName );
      W = lcPropGetInt( hImage, LC_PROP_IMAGE_WPIX );  // Image width (pixels) 
      H = lcPropGetInt( hImage, LC_PROP_IMAGE_HPIX );  // Image height (pixels) 
      pImageDesc = OdGiRasterImageDesc::createObject( W, H, OdGiRasterImage::kNone );
      pOdImageDef->setImage( pImageDesc );
      pOdImageDef->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
//      wcscpy( szOdDbHandle, (LPCWSTR)szOdBuf );
      lcPropPutStr( hImage, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );
    }
    hImage = lcDrwGetNextObject( hLcDb, hImage );
  }
}


//-----------------------------------------------
void Write_DimStyles (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR  szBuf[256], szBuf2[32], szOdDbHandle[32], szName[256];
  BOOL   bDeleted; //, bStd;
  HANDLE hTxtStyle;
  HANDLE hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_DIMSTYLE );
  OdDbDimStyleTablePtr pOdDimStyles = pOdDb->getDimStyleTableId().safeOpenObject( OdDb::kForWrite );
  OdDbDimStyleTableRecordPtr pOdDimStyle;
  OdChar szOdBuf[32];
  OdCmColor OdColor;
  OdDb::LineWeight OdLWeight;
  OdDbHandle OdHandle;
  OdDbSymbolTableIteratorPtr pIter = pOdDimStyles->newIterator();

  while( hStyle ){
    bDeleted = lcPropGetBool( hStyle, LC_PROP_DIMST_DELETED );
    if (bDeleted == false){
      wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_DIMST_NAME ) );
      // check if the name already exist
      for (pIter->start(); !pIter->done(); pIter->step()){
        pOdDimStyle = pIter->getRecordId().safeOpenObject( OdDb::kForWrite );
        wcscpy( szName, (LPCWSTR)pOdDimStyle->getName().c_str() );
        if (wcsicmp( szName, szBuf ) == 0){
          // exist, do not add new style
          goto m1;
        }
      }
      // add new style
      pOdDimStyle = OdDbDimStyleTableRecord::createObject();
      CorrectTableName( szBuf );
      pOdDimStyle->setName( szBuf );
      pOdDimStyles->add( pOdDimStyle );
/*
      bStd = lcPropGetBool( hStyle, LC_PROP_DIMST_STANDARD );
      if (bStd){
        pOdDimStyle = pOdDb->getDimStyleStandardId().safeOpenObject( OdDb::kForWrite );
      }else{
        wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_DIMST_NAME ) );
        // check if the name already exist
        for (pIter->start(); !pIter->done(); pIter->step()){
          pOdDimStyle = pIter->getRecordId().safeOpenObject( OdDb::kForWrite );
          wcscpy( szName, (LPCWSTR)pOdDimStyle->getName().c_str() );
          if (wcsicmp( szName, szBuf ) == 0){
            goto m1;
          }
        }
        pOdDimStyle = OdDbDimStyleTableRecord::createObject();
        pOdDimStyle->setName( szBuf );
        pOdDimStyles->add( pOdDimStyle );
      }
*/
m1:
      pOdDimStyle->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
      lcPropPutStr( hStyle, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );

      pOdDimStyle->setDimadec( lcPropGetInt( hStyle, LC_PROP_DIMST_ADEC )); 
      pOdDimStyle->setDimasz( lcPropGetFloat( hStyle, LC_PROP_DIMST_ASZ ));
      pOdDimStyle->setDimaunit( lcPropGetInt( hStyle, LC_PROP_DIMST_AUNIT ));
      pOdDimStyle->setDimazin( lcPropGetInt( hStyle, LC_PROP_DIMST_AZIN ));
      pOdDimStyle->setDimcen( lcPropGetFloat( hStyle, LC_PROP_DIMST_CEN ));
      MakeOdColor( lcPropGetInt( hStyle, LC_PROP_DIMST_CLRD ), &OdColor );
      pOdDimStyle->setDimclrd( OdColor );
      MakeOdColor( lcPropGetInt( hStyle, LC_PROP_DIMST_CLRE ), &OdColor );
      pOdDimStyle->setDimclre( OdColor );
      MakeOdColor( lcPropGetInt( hStyle, LC_PROP_DIMST_CLRT ), &OdColor );
      pOdDimStyle->setDimclrt( OdColor );
      pOdDimStyle->setDimdec( lcPropGetInt( hStyle, LC_PROP_DIMST_DEC ));
      pOdDimStyle->setDimdsep( lcPropGetInt( hStyle, LC_PROP_DIMST_DSEP ));
      pOdDimStyle->setDimexe( lcPropGetFloat( hStyle, LC_PROP_DIMST_EXE ));
      pOdDimStyle->setDimexo( lcPropGetFloat( hStyle, LC_PROP_DIMST_EXO ));
      pOdDimStyle->setDimgap( lcPropGetFloat( hStyle, LC_PROP_DIMST_GAP ));
      pOdDimStyle->setDimlfac( lcPropGetFloat( hStyle, LC_PROP_DIMST_LFAC ));
      OdLWeight = (OdDb::LineWeight)lcPropGetInt( hStyle, LC_PROP_DIMST_LWD );
      pOdDimStyle->setDimlwd( OdLWeight );
      OdLWeight = (OdDb::LineWeight)lcPropGetInt( hStyle, LC_PROP_DIMST_LWE );
      pOdDimStyle->setDimlwe( OdLWeight );
      pOdDimStyle->setDimpost( lcPropGetStr( hStyle, LC_PROP_DIMST_POST ));
      pOdDimStyle->setDimrnd( lcPropGetFloat( hStyle, LC_PROP_DIMST_RND )); 
      pOdDimStyle->setDimscale( lcPropGetFloat( hStyle, LC_PROP_DIMST_SCALE ));
      pOdDimStyle->setDimtad( lcPropGetInt( hStyle, LC_PROP_DIMST_TAD ));
      pOdDimStyle->setDimtih( lcPropGetInt( hStyle, LC_PROP_DIMST_TIH )?true:false);
      pOdDimStyle->setDimtoh( lcPropGetInt( hStyle, LC_PROP_DIMST_TIH )?true:false);  // same as TIH
      pOdDimStyle->setDimtxt( lcPropGetFloat( hStyle, LC_PROP_DIMST_TXT ));
      pOdDimStyle->setDimlunit( lcPropGetInt( hStyle, LC_PROP_DIMST_LUNIT ));
      pOdDimStyle->setDimzin( lcPropGetInt( hStyle, LC_PROP_DIMST_ZIN ));
      // arrowheads
      wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_DIMST_LDRBLK ));
      pOdDimStyle->setDimldrblk( szBuf );
      wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_DIMST_BLK1 ));
      wcscpy( szBuf2, lcPropGetStr( hStyle, LC_PROP_DIMST_BLK2 ));
      if (wcscmp( szBuf, szBuf2 ) == 0){
        pOdDimStyle->setDimsah( false );  // same arrowheads
        pOdDimStyle->setDimblk( szBuf );
      }else{
        pOdDimStyle->setDimsah( true );  // separate arrowheads
        pOdDimStyle->setDimblk1( szBuf );
        pOdDimStyle->setDimblk2( szBuf2 );
      }
      // text style
      hTxtStyle = lcPropGetHandle( hStyle, LC_PROP_DIMST_TXSTY );
      wcscpy( szOdDbHandle, lcPropGetStr( hTxtStyle, LC_PROP_TABLE_ODA_HANDLE ));
      OdHandle = OdDbHandle( szOdDbHandle );
      pOdDimStyle->setDimtxsty( pOdDb->getOdDbObjectId( OdHandle ) );
    }
    hStyle = lcDrwGetNextObject( hLcDb, hStyle );
  }
}


//-----------------------------------------------
void Write_MlineStyles (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR  szBuf[1024], szOdDbHandle[32];
  BOOL   bDeleted, bStd, b;
  HANDLE hLtype;
  HANDLE hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_MLSTYLE );
  int    i, nLines;
  double Off;
  OdChar            szOdBuf[32];
  OdCmColor         OdColor;
  OdDbHandle        OdHandle;
  OdDbObjectId      idMls;
  OdDbMlineStylePtr pOdMLStyle;
  OdDbDictionaryPtr pMLDic = pOdDb->getMLStyleDictionaryId().safeOpenObject( OdDb::kForWrite );
  while( hStyle ){
    bDeleted = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_DELETED );
    if (bDeleted == false){
      bStd = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_STANDARD );
      if (bStd){
        goto m1;
        idMls = pMLDic->getAt( L"Standard" );
        pOdMLStyle = idMls.safeOpenObject( OdDb::kForWrite );
      }else{
        pOdMLStyle = OdDbMlineStyle::createObject();
        wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_MLSTYLE_NAME ) );
        CorrectTableName( szBuf );
        pOdMLStyle->setName( szBuf );
        idMls = pMLDic->setAt( szBuf, pOdMLStyle );
      }
      pOdMLStyle->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
      lcPropPutStr( hStyle, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );

      wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_MLSTYLE_DESCR ));
      pOdMLStyle->setDescription( szBuf ); 
      b = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_JOINTS );
      pOdMLStyle->setShowMiters( b?true:false );
      b = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_STARTLINE );
      pOdMLStyle->setStartSquareCap( b?true:false );
      b = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_STARTARC );
      pOdMLStyle->setStartRoundCap( b?true:false );
      b = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_ENDLINE );
      pOdMLStyle->setEndSquareCap( b?true:false );
      b = lcPropGetBool( hStyle, LC_PROP_MLSTYLE_ENDARC );
      pOdMLStyle->setEndRoundCap( b?true:false );
//      pOdMLStyle->setStartAngle( pMLStyle->get_StartAngle() );
//      pOdMLStyle->setEndAngle( pMLStyle->get_EndAngle() );
      nLines = lcPropGetInt( hStyle, LC_PROP_MLSTYLE_NLINES );
      for (i=0; i<nLines; i++){
        lcPropPutInt( hStyle, LC_PROP_MLSTYLE_ILINE, i );
        Off = lcPropGetFloat( hStyle, LC_PROP_MLSTYLE_OFFSET ); 
        hLtype = lcPropGetHandle( hStyle, LC_PROP_MLSTYLE_LTYPE ); 
        wcscpy( szBuf, lcPropGetStr( hStyle, LC_PROP_MLSTYLE_COLOR ));
        MakeOdColor( szBuf, &OdColor );
/*
        off = pMLStyle->get_LineOffset( i );
        color.setColorIndex( pMLStyle->get_LineColor( i ) );
        id = pMLStyle->get_LineLtype( i );
        pLinetype = GetLinetype( id );
        if (pLinetype==NULL){
          pLinetype = GetLinetypeContinuous();
          if (pLinetype==NULL){
            pLinetype = GetFirstLinetype();
          }
        }
*/
        wcscpy( szOdDbHandle, lcPropGetStr( hLtype, LC_PROP_TABLE_ODA_HANDLE ));
        OdHandle = OdDbHandle( szOdDbHandle );
        if (bStd && (i < 2)){
          pOdMLStyle->setElement( i, Off, OdColor, pOdDb->getOdDbObjectId( OdHandle ) );
        }else{
          pOdMLStyle->addElement( Off, OdColor, pOdDb->getOdDbObjectId( OdHandle ) );
        }
      }
    }
m1:
    hStyle = lcDrwGetNextObject( hLcDb, hStyle );
  }
}


//-----------------------------------------------
void Write_Entity (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  int EntType = lcPropGetInt( hEnt, LC_PROP_ENT_TYPE );
  switch( EntType ){
    case LC_ENT_POINT:    Write_Point( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_LINE:     Write_Line( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_XLINE:    Write_Xline( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_POLYLINE: Write_Pline( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_RECT:     Write_Rect( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_ARC:      Write_Arc( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_CIRCLE:   Write_Circle( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_ELLIPSE:  Write_Ellipse( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_TEXT:     Write_Text( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_TEXTWIN:  Write_TextWin( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_MTEXT:    Write_MText( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_ARCTEXT:  Write_ArcText( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_BLOCKREF: Write_BlockRef( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_IMAGEREF: Write_ImageRef( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_HATCH:    Write_Hatch( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_VIEWPORT: Write_Viewport( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_MLINE:    Write_MLine( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMLIN:   Write_DimLin( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMALI:   Write_DimAli( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMORD:   Write_DimOrd( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMRAD:   Write_DimRad( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMDIA:   Write_DimDia( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_DIMANG:   Write_DimAng( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_LEADER:   Write_Leader( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_ATTDEF:   Write_Attdef( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_RPLAN:    Write_RPlan( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_BARCODE:  Write_Exploded( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_SHAPE:    Write_Shape( hEnt, pOdDb, pOdBlock );  break;
    case LC_ENT_FACE:     Write_Face3D( hEnt, pOdDb, pOdBlock );  break;
  }
}


//-----------------------------------------------
void LoopDefaultLayouts (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szOdDbHandle[32], szBlockName[128], szName[128];
  HANDLE  hBlock;
  BOOL    bDeleted, bPaper;
  OdChar                     szOdBuf[32];
  OdDbBlockTableRecordPtr    pOdBlock;
  OdDbObjectId               OdId;
  OdDbLayoutPtr              pOdLayout;
  OdDbBlockTableRecordPtr    pOdModelBlock;
  OdDbBlockTablePtr          pTable;
  OdDbSymbolTableIteratorPtr pBlkIter;
  OdInt16 nLayouts;

  // clear ODHANDLE in Litecad blocks
  memset( szOdDbHandle, 0, sizeof(szOdDbHandle) );
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
  while( hBlock ){
    lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, szOdDbHandle );
    lcPropPutStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE, szOdDbHandle );
    hBlock = lcDrwGetNextObject( hLcDb, hBlock );
  }

  nLayouts = pOdDb->countLayouts();
  pOdModelBlock = pOdDb->getModelSpaceId().safeOpenObject( OdDb::kForWrite );
  pTable = pOdDb->getBlockTableId().safeOpenObject();
  // connect default DWG blocks and layouts with exist LiteCAD blocks
  pBlkIter = pTable->newIterator();
  for (pBlkIter->start(); ! pBlkIter->done(); pBlkIter->step()){
    pOdBlock = pBlkIter->getRecordId().safeOpenObject();
    wcscpy( szBlockName, (LPCWSTR)pOdBlock->getName().c_str() );          // Block Name
    if (pOdBlock->isLayout()){
      if (pOdBlock == pOdModelBlock){
        // this is DWG ModelSpace block
        hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
        if (hBlock){
          // save DWG handle of "*Model_Space" block in LiteCAD "*Model_Space" block
          pOdModelBlock->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
          lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );   // save OdBlock handle in LiteCAD block object
        }
      }else{
        // this is DWG PaperSpace block
        OdId = pOdBlock->getLayoutId();
        if (OdId.isValid()){
          pOdLayout = OdId.safeOpenObject( OdDb::kForWrite );
          wcscpy( szName, (LPCWSTR)pOdLayout->getLayoutName().c_str() );
          hBlock = lcDrwGetObjectByName( hLcDb, LC_OBJ_BLOCK, szBlockName );
          if (hBlock){
            bDeleted = lcPropGetBool( hBlock, LC_PROP_BLOCK_DELETED );
            if (bDeleted == false){
              bPaper = lcPropGetBool( hBlock, LC_PROP_BLOCK_PAPER );
              if (bPaper){
                pOdBlock->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
                lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );  // save OdBlock handle in LiteCAD block object
              }
            }
          }
/*
          pOdDb->setCurrentLayout( OdId );
          OdId = pOdLayout->overallVportId();
          if (OdId.isValid()){
            pOverallViewport = OdId.safeOpenObject();
            if (pOverallViewport.isNull() == false){
              overallCenter = pOverallViewport->centerPoint();
              overallWidth = pOverallViewport->width(); //  / 1.058 - 2 * margin;
              overallHeight = pOverallViewport->height(); // / 1.058 - 2 * margin;
            }
          }
*/
        }
      }
    }
  }
}


//-----------------------------------------------
void Write_BlockDefs (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szOdDbHandle[32];
  WCHAR   szLayoutName[160], szBlockName[160], szOdBlockName[160];
  LPCWSTR psz;
  HANDLE  hBlock, hVisBlock;
  BOOL    bStandard, bLayout, bDeleted, bModel, bHidden;
  double  PaperL, PagerB, PaperW, PaperH;
  double  Lef, Bot, Rig, Top, W, H, X, Y;
  CGeRect Rect;
  int     PaperSize, PaperOrient;
  OdDbViewportTablePtr         pVportTable;
  OdDbViewportTableRecordPtr   pVportRec;
  OdDbPlotSettingsValidatorPtr pValidator;
  OdDbBlockTablePtr            pOdBlocks;
  OdDbBlockTableRecordPtr      pOdBlock;
  OdDbObjectId    OdId, VportID;
  OdDbHandle      OdHandle;
  OdDbLayoutPtr   pOdLayout;
  OdDbViewportPtr pOverallViewport;
  OdGePoint2d     viewCenter, limMin, limMax;
  OdResult        OdRes;
  OdChar          szOdBuf[32];
  OdGePoint3d     OdPt3;
  int i = 1;
  const int nNames = 14;
  WCHAR* szFooName[nNames] = {
    L"_ArchTick", L"_Box"     , L"_Closed"  , L"_Datum", L"_Dot"    , 
    L"_DotBlank", L"_DotSmall", L"_Integral", L"_None" , L"_Oblique", 
    L"_Open30"  , L"_Open90"  , L"_Origin"  , L"_Small"
  };

  LoopDefaultLayouts( hLcDb, pOdDb );

  hVisBlock = lcPropGetHandle( hLcDb, LC_PROP_DRW_VISBLOCK );
  // add Model and Paper blocks
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
  while( hBlock ){
    bDeleted = lcPropGetBool( hBlock, LC_PROP_BLOCK_DELETED );
    if (bDeleted == false){
      bLayout = lcPropGetBool( hBlock, LC_PROP_BLOCK_LAYOUT );
      if (bLayout){
        bModel = false;
        wcscpy( szBlockName , lcPropGetStr( hBlock, LC_PROP_BLOCK_NAME ) );
        wcscpy( szLayoutName, lcPropGetStr( hBlock, LC_PROP_BLOCK_LAYOUTNAME ) );
        wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_TABLE_ODA_HANDLE ) );
        OdHandle = OdDbHandle( szOdDbHandle );
        if (OdHandle.isNull()){
          // add new layout to OdDbDatabase
          OdId = pOdDb->createLayout( szLayoutName );
          pOdLayout = OdId.safeOpenObject();
          // get block which is linked with the layout
          pOdBlock = pOdLayout->getBlockTableRecordId().safeOpenObject(OdDb::kForWrite);
          wcscpy( szOdBlockName, (LPCWSTR)pOdBlock->getName().c_str() );
          // save pOdBlock handle in LiteCAD block object
          pOdBlock->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
          lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );
        }else{
          // default block, aready exist in OdDbDatabase
          pOdBlock = pOdDb->getOdDbObjectId( OdHandle ).safeOpenObject( OdDb::kForWrite );
          wcscpy( szOdBlockName, (LPCWSTR)pOdBlock->getName().c_str() );
          OdId = pOdBlock->getLayoutId();
          pOdLayout = OdId.safeOpenObject( OdDb::kForWrite );
          bModel = lcPropGetBool( hBlock, LC_PROP_BLOCK_MODEL );
          if (bModel == false){
            // set Layout name as in LiteCAD (Model Space can't be renamed)
            pOdLayout->setLayoutName( szLayoutName );
          }
        }

        pOdDb->setCurrentLayout( OdId );
        Lef = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISLEF );
        Bot = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISBOT );
        Rig = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISRIG );
        Top = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISTOP );
        Rect.Set( Lef, Bot, Rig, Top );
        viewCenter.x = Rect.Center().x;
        viewCenter.y = Rect.Center().y;
        if (bModel){
          // Model space layout
          pVportTable = pOdDb->getViewportTableId().safeOpenObject( OdDb::kForWrite );
          VportID = pVportTable->getActiveViewportId();
          pVportRec = VportID.safeOpenObject( OdDb::kForWrite );
          pVportRec->setCenterPoint( viewCenter );
          pVportRec->setHeight( Rect.Height() ); // * 1.05 );
          pVportRec->setWidth( Rect.Width() ); //* 1.05 );
          pVportRec->setCircleSides( 20000 );
        }else{
          // Paper space layout
          pOdDb->setCurrentLayout( pOdLayout->objectId() );
          pOdDb->setTILEMODE(1);
          W = H = 0.0;
          OdId = pOdLayout->overallVportId();
          if (OdId.isValid()){
            pOverallViewport = OdId.safeOpenObject( OdDb::kForWrite );
            if (pOverallViewport.isNull() == false){
              pOverallViewport->setViewCenter( viewCenter );
              pOverallViewport->setViewHeight( Rect.Height() ); // * 1.05 );
              pOverallViewport->setCircleSides( 20000 );
            }else{
              i = 0; 
            }
          }else{
            i = 0;
          }

          // write data of paper sheet
          PaperL = lcPropGetFloat( hBlock, LC_PROP_PAPER_X0 );
          PagerB = lcPropGetFloat( hBlock, LC_PROP_PAPER_Y0 );
          PaperW = lcPropGetFloat( hBlock, LC_PROP_PAPER_W );
          PaperH = lcPropGetFloat( hBlock, LC_PROP_PAPER_H );
          PaperSize = lcPropGetInt( hBlock, LC_PROP_PAPER_SIZE );
          PaperOrient = lcPropGetInt( hBlock, LC_PROP_PAPER_ORIENT );
          pValidator = pOdLayout->database()->appServices()->plotSettingsValidator();
          if (pValidator.isNull() == false){
            X = Y = 0.0;
            OdRes = pValidator->setClosestMediaName( pOdLayout, PaperW, PaperH, OdDbPlotSettings::kMillimeters, false );
            if (OdRes == OdResult::eOk){
              psz = (LPCWSTR)pOdLayout->getCanonicalMediaName();
              if (PaperOrient == 0){
                // book
                pValidator->setPlotRotation( pOdLayout, PlotRotation::k0degrees );
              }else{
                // album
                pValidator->setPlotRotation( pOdLayout, PlotRotation::k90degrees );
              }
            }
            limMin.x = PaperL;
            limMin.y = PagerB;
            limMax.x = limMin.x + (PaperOrient)?PaperH:PaperW;
            limMax.y = limMin.y + (PaperOrient)?PaperW:PaperH;
            pOdLayout->setLIMMIN( limMin );
            pOdLayout->setLIMMAX( limMax );
            X = pOdLayout->getLeftMargin();
            Y = pOdLayout->getBottomMargin();
            pValidator->setPlotOrigin( pOdLayout, -X, -Y );
            pValidator->setPlotType( pOdLayout, OdDbPlotSettings::kLayout );
            pValidator->setUseStandardScale( pOdLayout, true );
            pValidator->setStdScaleType( pOdLayout, OdDbPlotSettings::k1_1 );
          }
        }
        // save OdLayout handle in LiteCAD block object
        pOdLayout->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
        lcPropPutStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE, (LPCWSTR)szOdBuf );  
      }
    }
//m1:
    hBlock = lcDrwGetNextObject( hLcDb, hBlock );
  }
  
  // add standard blocks (non-layouts)
  pOdBlocks = pOdDb->getBlockTableId().safeOpenObject( OdDb::kForWrite );
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
  while( hBlock ){
    bDeleted = lcPropGetBool( hBlock, LC_PROP_BLOCK_DELETED );
    if (bDeleted == false){
      bStandard = lcPropGetBool( hBlock, LC_PROP_BLOCK_STANDARD );
      if (bStandard){
        wcscpy( szBlockName, lcPropGetStr( hBlock, LC_PROP_BLOCK_NAME ) );
        for (i=0; i<nNames; ++i){
          if (_wcsicmp( szBlockName, szFooName[i] ) == 0){
            goto m2;
          }
        }
        bHidden = lcPropGetBool( hBlock, LC_PROP_BLOCK_HIDDEN );
/*
        if (bHidden){
          szBlockName[2] = 0;
        }
*/
        pOdBlock = OdDbBlockTableRecord::createObject();
//        pOdBlock->setAttributes()
/*
        if (szBlockName[0] == '*'){
          CorrectTableName( szBlockName+1 );
        }else{
          CorrectTableName( szBlockName );
        }
*/
        CorrectTableName( szBlockName );
        pOdBlock->setName( szBlockName );
        pOdBlocks->add( pOdBlock );
        pOdBlock->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
        lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szOdBuf );
        OdPt3.x = lcPropGetFloat( hBlock, LC_PROP_BLOCK_X );
        OdPt3.y = lcPropGetFloat( hBlock, LC_PROP_BLOCK_Y );
        OdPt3.z = 0.0; //lcPropGetFloat( hBlock, LC_PROP_BLOCK_Z );
        pOdBlock->setOrigin( OdPt3 );
      }
    }
m2:
    hBlock = lcDrwGetNextObject( hLcDb, hBlock );
  }
}


//#define _IENT_  1

//-----------------------------------------------
void Write_Entities (HANDLE  hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szID[20];
  WCHAR   szOdDbHandle[32]; //, szBuf[32];
  HANDLE  hBlock, hEnt, hLayer;
  BOOL    bDeleted, bNoExport;
  OdDbHandle              OdHandle;
  OdDbBlockTableRecordPtr pOdBlock;
#ifdef _IENT_
  int iBlock=0, iEnt, nEnts, j=0;
  int nBlocks = lcDrwCountObjects( hLcDb, LC_OBJ_BLOCK );
#endif
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
  while( hBlock ){
    bDeleted = lcPropGetBool( hBlock, LC_PROP_BLOCK_DELETED );
    if (bDeleted == false){
      wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_TABLE_ODA_HANDLE ) );
      OdHandle = OdDbHandle( szOdDbHandle );
      if (OdHandle.isNull() == false){
        pOdBlock = pOdDb->getOdDbObjectId( OdHandle ).safeOpenObject( OdDb::kForWrite );
        // write entities
#ifdef _IENT_
        iEnt=0;
        nEnts = lcPropGetInt( hBlock, LC_PROP_BLOCK_NOBJ );
//_itow( nEnts, szBuf, 10 );
//    ::MessageBox( ::GetActiveWindow(), szBuf, L"Num of Entities", MB_OK );
#endif
        hEnt = lcBlockGetFirstEnt( hBlock );
        while( hEnt ){
          bDeleted = lcPropGetBool( hEnt, LC_PROP_ENT_DELETED );
          if (bDeleted == false){
            hLayer = lcPropGetHandle( hEnt, LC_PROP_ENT_LAYER );
            bNoExport = lcPropGetBool( hLayer, LC_PROP_LAYER_NOEXPORT );
            if (bNoExport == false){
              wcscpy( szID, lcPropGetStr( hEnt, LC_PROP_ENT_ID ));
/*
              if (wcscmp(szID,L"4E6")==0){
                szID[10] = 0;
              }
*/
              Write_Entity( hEnt, pOdDb, pOdBlock );
/*
++j;
if (iEnt == 7435){
  _bDebugEnt = true;
//  j = 0;
_itow( iEnt, szBuf, 10 );
    ::MessageBox( ::GetActiveWindow(), szBuf, L"Entity", MB_OK );
}
*/
            }
          }
#ifdef _IENT_
          ++iEnt;
          if (iEnt==117 || iEnt==129 || iEnt==149){
            iEnt = iEnt;
          }
#endif
          hEnt = lcBlockGetNextEnt( hBlock, hEnt );
        }
      }
    }
#ifdef _IENT_
    if (iBlock == 32){
      iBlock = iBlock;
    }
    ++iBlock;
#endif
    hBlock = lcDrwGetNextObject( hLcDb, hBlock );
  }
}


//-----------------------------------------------
void SetCurrentLayout (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  HANDLE  hBlock;
  WCHAR   szOdDbHandle[32];
  OdDbHandle    OdHandle;
  OdDbLayoutPtr pOdLayout;

  // set current layout
  hBlock = lcPropGetHandle( hLcDb, LC_PROP_DRW_VISBLOCK );  // currently visible block
  if (hBlock){
    wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE ) );
    OdHandle = OdDbHandle( szOdDbHandle );
    if (OdHandle.isNull() == false){
      pOdLayout = pOdDb->getOdDbObjectId( OdHandle ).safeOpenObject( OdDb::kForWrite );
      if (pOdLayout.isNull() == false){
        pOdDb->setCurrentLayout( pOdLayout->objectId() );
      }
    }
  }
}

#ifdef _UUUEUEUEUDHDHDH
//-----------------------------------------------
void Write_Viewports (HANDLE hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR   szOdDbHandle[32], szBuf[256];
  HANDLE  hBlock;
  BOOL    bDeleted, bPaper;
  CGeRect Rect;
  double  W, H, Lef, Bot, Rig, Top;
  int     Orient;
  OdDbHandle                 OdHandle;
  OdDbLayoutPtr              pOdLayout;
  OdDbViewportPtr            pOverallViewport;
  OdGePoint2d                viewCenter;
  OdGePoint2d                limMin, limMax;
  OdDbViewportTablePtr       pVpTable = pOdDb->getViewportTableId().safeOpenObject( OdDb::kForWrite );
  OdDbObjectId               vpID = pVpTable->getActiveViewportId();
  OdDbViewportTableRecordPtr vPortRec = vpID.safeOpenObject(OdDb::kForWrite);
  OdDbPlotSettingsValidator* pValidator;

  // set zoom for Paper Space layouts
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );
  while( hBlock ){
    wcscpy( szBuf, lcPropGetStr( hBlock, LC_PROP_BLOCK_NAME ) );
    bDeleted = lcPropGetBool( hBlock, LC_PROP_BLOCK_DELETED );
    if (bDeleted == false){
      bPaper = lcPropGetBool( hBlock, LC_PROP_BLOCK_PAPER );
      if (bPaper){
        Lef = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISLEF );
        Bot = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISBOT );
        Rig = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISRIG );
        Top = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISTOP );
        Rect.Set( Lef, Bot, Rig, Top );
        wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE ) );
        if (szOdDbHandle[0]){
          OdHandle = OdDbHandle( szOdDbHandle );
          pOdLayout = pOdDb->getOdDbObjectId( OdHandle ).safeOpenObject( OdDb::kForWrite );
          pOdDb->setCurrentLayout( pOdLayout->objectId() );
          pOverallViewport = pOdLayout->overallVportId().safeOpenObject(OdDb::kForWrite);
          // set view rect for entire layout
          viewCenter.x = Rect.Center().x;
          viewCenter.y = Rect.Center().y;
          pOverallViewport->setViewCenter( viewCenter );
          pOverallViewport->setViewHeight( Rect.Height() );

          W = lcPropGetFloat( hBlock, LC_PROP_PAPER_W );
          H = lcPropGetFloat( hBlock, LC_PROP_PAPER_H );
          Orient = lcPropGetInt( hBlock, LC_PROP_PAPER_ORIENT );
          limMin.x = lcPropGetFloat( hBlock, LC_PROP_PAPER_X0 );
          limMin.y = lcPropGetFloat( hBlock, LC_PROP_PAPER_Y0 );
          if (Orient == LC_PAPER_BOOK){
            limMax.x = limMin.x + W;
            limMax.y = limMin.y + H;
          }else{
            limMax.x = limMin.x + H;
            limMax.y = limMin.y + W;
          }
//                pLayout->pplotPaperUnits();  // 0 - inches, 1 - millimeters
//          pLayout->setPaperImageOrigin( limMin );
          pOdLayout->setLIMMIN( limMin );
          pOdLayout->setLIMMAX( limMax );
/*
pValidator->setClosestMediaName ( pLayout, ys, xs, OdDbPlotSettings::kMillimeters, false );
pValidator->setPlotOrigin ( pLayout, xo, yo );
pValidator->setPlotViewName ( pLayout, plotviewname );
pValidator->setPlotCfgName ( pLayout, "none_device" );
pValidator->setStdScale ( pLayout, m_docCover.m_systemUnit );
*/

          pValidator = pOdLayout->database()->appServices()->plotSettingsValidator();

// set the scale to 1:1 in mm, and the plot target to be kLayout 
// (i.e. the sheet of paper and not the current view or events, etc)

          pValidator->setClosestMediaName( pOdLayout, W, H, OdDbPlotSettings::kMillimeters, false );
//            setMediaSize( pOdLayout, L"A4", 210, 297, OdDbPlotSettings::kMillimeters, 5,5,5,5 );
          pValidator->setPlotPaperUnits( pOdLayout, OdDbPlotSettings::kMillimeters );
          if (Orient == LC_PAPER_BOOK){
            pValidator->setPlotRotation( pOdLayout, OdDbPlotSettings::k90degrees );
          }
          pValidator->setPlotType( pOdLayout, OdDbPlotSettings::kLayout );
          pValidator->setUseStandardScale( pOdLayout, true );
          pValidator->setStdScale( pOdLayout, OdDbPlotSettings::k1_1 );
//          pValidator->setStdScaleType( pOdLayout, 1.0 );
        }
      }
    }
    hBlock = lcDrwGetNextObject( hLcDb, hBlock );
  }

  // set zoom for Model Space layout
  hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );   // first view is always for "Model Space"
  wcscpy( szBuf, lcPropGetStr( hBlock, LC_PROP_BLOCK_LAYOUTNAME ) );
  pOdDb->setCurrentLayout( szBuf );
  Lef = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISLEF );
  Bot = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISBOT );
  Rig = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISRIG );
  Top = lcPropGetFloat( hBlock, LC_PROP_BLOCK_VISTOP );
  Rect.Set( Lef, Bot, Rig, Top );
  viewCenter.x = Rect.Center().x;
  viewCenter.y = Rect.Center().y;
  vPortRec->setCenterPoint( viewCenter );
  vPortRec->setHeight( Rect.Height() * 1.05 );
  vPortRec->setWidth( Rect.Width() * 1.05 );
  vPortRec->setCircleSides( 20000 );

  // set current layout
  hBlock = lcPropGetHandle( hLcDb, LC_PROP_DRW_VISBLOCK );  // currently visible block
  if (hBlock){
    wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE ) );
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdLayout = pOdDb->getOdDbObjectId( OdHandle ).safeOpenObject( OdDb::kForWrite );
    pOdDb->setCurrentLayout( pOdLayout->objectId() );
  }
/*
  wcscpy( szBuf, lcPropGetStr( hBlock, LC_PROP_BLOCK_LAYOUTNAME ) );
  if (lcPropGetBool( hBlock, LC_PROP_BLOCK_MODEL )){
    pOdDb->setCurrentLayout( L"Model" );
  }else{
    pOdDb->setCurrentLayout( szBuf );
  }
*/
}
#endif  // _UUUEUEUEUDHDHDH

// Layer names cannot include the following characters: < > / \ " : ; ? * | = ' ,
//-----------------------------------------------
void CorrectTableName (WCHAR* szName)
{
  const UINT nBads = 13;
  WCHAR BadChar[nBads] = {L'<', L'>', L'/', L'\\', L'"', L':', L';', L'?', L'*', L'|', L'=', L'\'', L','};
  WCHAR ch;
  UINT  i, j;

  i = 0;
  for (;;){
    ch = szName[i];
    if (ch == 0){
      break;
    }
    for (j=0; j<nBads; ++j){
      if (ch == BadChar[j]){
        szName[i] = L'_';
        break;
      }
    }
    ++i;
  }
}
/*
was used only in Write_MlineStyles, before ver.169
//-----------------------------------------------
void CorrectTableName (WCHAR* szName)
{
  UINT_PTR i;
  UINT_PTR len = wcslen( szName );
  for (i=0; i<len; i++){
    if (szName[i] == L' '  || szName[i] == L'.' || szName[i] == L'/' || 
        szName[i] == L'\\' || szName[i] == 39   || szName[i] == L'(' ||
        szName[i] == L')')
    {
      szName[i] = '_';
    }
  }
}
*/

/*
//-----------------------------------------------
void ExportLayouts (HANDLE  hLcDb, OdDbDatabasePtr pOdDb)
{
  WCHAR  szName[256], szOdDbHandle[32];
  BOOL   bDeleted;
//  BOOL   bFirst = true;
  BOOL   bPaper;
  HANDLE hView = lcDrwGetFirstObject( hLcDb, LC_OBJ_VIEW );
//  int    W, H;
  CGeRect Rect;
  OdChar                  szOdBuf[32];
  OdDbObjectId            LayoutId; // = pDb->createLayout(DD_T("ODA Layout"));
  OdDbLayoutPtr           pOdLayout; // = layoutId.safeOpenObject();
  OdDbViewportPtr         pOverallViewport;
  OdGePoint2d             viewCenter;
  OdDbBlockTableRecordPtr pOdBlock;
  OdDbViewportPtr         pViewport;

  while( hView ){
    bDeleted = lcPropGetBool( hView, LC_PROP_VIEW_DELETED );
    if (bDeleted == false){
      Rect.SetLeft(   lcPropGetFloat( hView, LC_PROP_VIEW_LEF ) );
      Rect.SetBottom( lcPropGetFloat( hView, LC_PROP_VIEW_BOT ) );
      Rect.SetRight(  lcPropGetFloat( hView, LC_PROP_VIEW_RIG ) );
      Rect.SetTop(    lcPropGetFloat( hView, LC_PROP_VIEW_TOP ) );
      bPaper = lcPropGetBool( hView, LC_PROP_VIEW_PAPER );
      if (bPaper){
        wcscpy( szName, lcPropGetStr( hView, LC_PROP_VIEW_NAME ) );
        LayoutId = pOdDb->createLayout( szName );
        pOdLayout = LayoutId.safeOpenObject();
        pOdBlock = pOdLayout->getBlockTableRecordId().safeOpenObject(OdDb::kForWrite);
        pOdBlock->objectId().getHandle().getIntoAsciiBuffer( szOdBuf );
        wcscpy( szOdDbHandle, (LPCWSTR)szOdBuf );
        lcPropPutStr( hView, LC_PROP_TABLE_STR, szOdDbHandle ); 
        pOdDb->setCurrentLayout( LayoutId );
      }else{
        // model space layout
        LayoutId = pOdDb->findLayoutNamed( L"Model" );
        pOdLayout = LayoutId.safeOpenObject( OdDb::kForWrite );
      }
        // set view rect for entire layout
        pOverallViewport = pOdLayout->overallVportId().safeOpenObject(OdDb::kForWrite);
        viewCenter.x = Rect.Center().x;
        viewCenter.y = Rect.Center().y;
        pOverallViewport->setViewCenter( viewCenter ); //overallCenter );
        pOverallViewport->setViewHeight( Rect.Height() );
    }
    hView = lcDrwGetNextObject( hLcDb, hView );
  }
}
*/

#ifdef JHNBGFRDLO
    /********************************************************************/
    /* Create a new Layout                                              */
    /********************************************************************/  
    OdDbObjectId layoutId = pDb->createLayout(DD_T("ODA Layout"));
    OdDbLayoutPtr pLayout = layoutId.safeOpenObject();

    /********************************************************************/
    /* Make it current, creating the overall PaperSpace viewport        */
    /********************************************************************/  
    pDb->setCurrentLayout(layoutId);
     
    /********************************************************************/
    /* Open the overall viewport for this layout                        */
    /********************************************************************/  
    OdDbViewportPtr pOverallViewport = pLayout->overallVportId().safeOpenObject();

    /********************************************************************/
    /* Get some useful parameters                                       */
    /********************************************************************/  
    OdGePoint3d overallCenter   = pOverallViewport->centerPoint();
    
    /********************************************************************/
    /* Note:                                                            */
    /* If a viewport is an overall viewport,                            */
    /* the values returned by width() and height() must be divided by a */
    /* factor of 1.058, and the parameters of setWidth and setHeight()  */
    /* must be multiplied a like factor.                                */
    /********************************************************************/
    const double margin = 0.25;
    double overallWidth         = pOverallViewport->width()  / 1.058 - 2 * margin;
    double overallHeight        = pOverallViewport->height() / 1.058 - 2 * margin;
    OdGePoint3d overallLLCorner = overallCenter -                         
      OdGeVector3d( 0.5 * overallWidth, 0.5 * overallHeight, 0.0 );
    
    /********************************************************************/
    /* Open the PaperSpace BlockTableRecord for this layout             */ 
    /********************************************************************/  
    OdDbBlockTableRecordPtr pPS = pLayout->getBlockTableRecordId().safeOpenObject(OdDb::kForWrite);
    
    /********************************************************************/
    /* Create a new viewport, and append it to PaperSpace               */ 
    /********************************************************************/  
    OdDbViewportPtr pViewport = OdDbViewport::createObject();
    pViewport->setDatabaseDefaults(pDb);
    pPS->appendOdDbEntity(pViewport);

    /********************************************************************/
    /* Set some parameters                                              */ 
    /*                                                                  */
    /* This viewport occupies the upper half of the overall viewport,   */
    /* and displays all objects in model space                          */
    /********************************************************************/  
    
    pViewport->setWidth         (overallWidth);
    pViewport->setHeight        (overallHeight*0.5);
    pViewport->setCenterPoint   (overallCenter + OdGeVector3d(0.0, 0.5*pViewport->height(), 0.0));
    pViewport->setViewCenter    (pOverallViewport->viewCenter());
    pViewport->zoomExtents();

#endif 

