/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
*********************************************************************/
#include "pch.h"

void Write_Exploded (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);

//-----------------------------------------------
void MakeOdColor (LPCWSTR szColor, OdCmColor* pOdColor)
{
  int bRGB, iC, R, G, B;
  if (lcColorToVal( szColor, &bRGB, &iC, &R, &G, &B )){
    if (bRGB){
      // Color specified by RGB value
      pOdColor->setColorMethod( OdCmEntityColor::kByColor );  
      pOdColor->setRed( R );
      pOdColor->setGreen( G );
      pOdColor->setBlue( B );
    }else{
      // Color specified by an index (ACI) into a color table.
      pOdColor->setColorMethod( OdCmEntityColor::kByACI );    
      pOdColor->setColorIndex( iC );
    }
  }else{
    pOdColor->setColorMethod( OdCmEntityColor::kForeground );
  }
}

// Color specified by an index (ACI) into a color table.
//-----------------------------------------------
void MakeOdColor (int iColor, OdCmColor* pOdColor)
{
  pOdColor->setColorMethod( OdCmEntityColor::kByACI );    
  pOdColor->setColorIndex( iColor );
}

//-----------------------------------------------
void WriteEntityBase (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbEntityPtr pOdEnt, bool bSolid)
{
  WCHAR  szColor[32], szOdDbHandle[32];
  double LtScale;
  HANDLE hLayer, hLtype;
  int    Lweight;
  OdDbHandle       OdHandle;
  OdCmColor        OdColor;
  OdDb::LineWeight OdLWeight = (OdDb::LineWeight)25;

  // color
  if (bSolid){
    wcscpy( szColor, lcPropGetStr( hEnt, LC_PROP_ENT_FCOLOR ));
  }else{
    wcscpy( szColor, lcPropGetStr( hEnt, LC_PROP_ENT_COLOR ));
  }
  MakeOdColor( szColor, &OdColor );
  pOdEnt->setColor( OdColor, false );
  
  // layer
  hLayer = lcPropGetHandle( hEnt, LC_PROP_ENT_LAYER );
  wcscpy( szOdDbHandle, lcPropGetStr( hLayer, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdEnt->setLayer( pOdDb->getOdDbObjectId( OdHandle ) );
  }

  // linetype
  hLtype = lcPropGetHandle( hEnt, LC_PROP_ENT_LINETYPE );
  wcscpy( szOdDbHandle, lcPropGetStr( hLtype, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdEnt->setLinetype( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  // linetype stale
  LtScale = lcPropGetFloat( hEnt, LC_PROP_ENT_LTSCALE );
  pOdEnt->setLinetypeScale( LtScale );

  // lineweight
  Lweight = lcPropGetInt( hEnt, LC_PROP_ENT_LWEIGHT );
  if (Lweight == LC_LWIDTH_DEFAULT){
    OdLWeight = OdDb::kLnWtByLwDefault;
  }else
  if (Lweight == LC_LWIDTH_BYLAYER){
    OdLWeight = OdDb::kLnWtByLayer;
  }else
  if (Lweight == LC_LWIDTH_BYBLOCK){
    OdLWeight = OdDb::kLnWtByBlock;
  }else
  if (0<=Lweight && Lweight<25){
    OdLWeight = OdDb::kLnWt000;
  }else
  if (25<=Lweight && Lweight<70){
    OdLWeight = OdDb::kLnWt005;
  }else
  if (70<=Lweight && Lweight<110){
    OdLWeight = OdDb::kLnWt009;
  }else
  if (110<=Lweight && Lweight<140){
    OdLWeight = OdDb::kLnWt013;
  }else
  if (110<=Lweight && Lweight<140){
    OdLWeight = OdDb::kLnWt013;
  }else
  if (140<=Lweight && Lweight<165){
    OdLWeight = OdDb::kLnWt015;
  }else
  if (165<=Lweight && Lweight<190){
    OdLWeight = OdDb::kLnWt018;
  }else
  if (190<=Lweight && Lweight<225){
    OdLWeight = OdDb::kLnWt020;
  }else
  if (225<=Lweight && Lweight<275){
    OdLWeight = OdDb::kLnWt025;
  }else
  if (275<=Lweight && Lweight<325){
    OdLWeight = OdDb::kLnWt030;
  }else
  if (325<=Lweight && Lweight<375){
    OdLWeight = OdDb::kLnWt035;
  }else
  if (375<=Lweight && Lweight<450){
    OdLWeight = OdDb::kLnWt040;
  }else
  if (450<=Lweight && Lweight<515){
    OdLWeight = OdDb::kLnWt050;
  }else
  if (515<=Lweight && Lweight<565){
    OdLWeight = OdDb::kLnWt053;
  }else
  if (565<=Lweight && Lweight<650){
    OdLWeight = OdDb::kLnWt060;
  }else
  if (650<=Lweight && Lweight<750){
    OdLWeight = OdDb::kLnWt070;
  }else
  if (750<=Lweight && Lweight<850){
    OdLWeight = OdDb::kLnWt080;
  }else
  if (850<=Lweight && Lweight<950){
    OdLWeight = OdDb::kLnWt090;
  }else
  if (950<=Lweight && Lweight<1030){
    OdLWeight = OdDb::kLnWt100;
  }else
  if (1030<=Lweight && Lweight<1130){
    OdLWeight = OdDb::kLnWt106;
  }else
  if (1130<=Lweight && Lweight<1300){
    OdLWeight = OdDb::kLnWt120;
  }else
  if (1300<=Lweight && Lweight<1490){
    OdLWeight = OdDb::kLnWt140;
  }else
  if (1490<=Lweight && Lweight<1790){
    OdLWeight = OdDb::kLnWt158;
  }else
  if (1790<=Lweight && Lweight<2100){
    OdLWeight = OdDb::kLnWt200;
  }else
  if (2100<=Lweight){
    OdLWeight = OdDb::kLnWt211;
  }
  pOdEnt->setLineWeight( OdLWeight );
/*
  if (pEnt->IsInvisible()){
    pOdEnt->setVisibility( OdDb::kInvisible );
  }else{
    pOdEnt->setVisibility( OdDb::kVisible );
  }
*/
}

//-----------------------------------------------
void Write_Point (HANDLE hPoint, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdGePoint3d OdPnt;
  OdDbPointPtr pOdPoint = OdDbPoint::createObject();

  OdPnt.x = lcPropGetFloat( hPoint, LC_PROP_POINT_X );
  OdPnt.y = lcPropGetFloat( hPoint, LC_PROP_POINT_Y );
  OdPnt.z = lcPropGetFloat( hPoint, LC_PROP_ENT_Z );
  pOdPoint->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdPoint );
  WriteEntityBase( hPoint, pOdDb, pOdPoint ); 
  pOdPoint->setPosition( OdPnt );

/*
  WCHAR   szText[256], szOdDbHandle[32];
  HANDLE  hStyle = lcPropGetHandle( hPoint, LC_PROP_POINT_STYLE );
  HANDLE  hBlock = lcPropGetHandle( hStyle, LC_PROP_PSTYLE_BLOCK );
  HANDLE  hTextStyle = lcPropGetHandle( hStyle, LC_PROP_PSTYLE_TSTYLE );
  double  BScale = lcPropGetFloat( hStyle, LC_PROP_PSTYLE_BSCALE );
  double  BAngle = lcPropGetFloat( hPoint, LC_PROP_POINT_BANGLE );
  int     TextLen = lcPropGetInt( hPoint, LC_PROP_POINT_TEXTLEN );
  double  Tdx, Tdy, THeight, TWScale, TAngle;
  OdGePoint3d OdPnt;
  OdDbHandle  OdHandle;
  OdGeScale3d OdScale;

  OdPnt.x = lcPropGetFloat( hPoint, LC_PROP_POINT_X );
  OdPnt.y = lcPropGetFloat( hPoint, LC_PROP_POINT_Y );
  OdPnt.z = lcPropGetFloat( hPoint, LC_PROP_POINT_Z );
  if (hBlock){
    OdDbBlockReferencePtr pOdInsert = OdDbBlockReference::createObject();
    pOdInsert->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdInsert );
    WriteEntityBase( hPoint, pOdDb, pOdInsert );
    wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_TABLE_STR ));
    if (szOdDbHandle[0]){
      OdHandle = OdDbHandle( szOdDbHandle );
      pOdInsert->setBlockTableRecord( pOdDb->getOdDbObjectId( OdHandle ) );
    } 
    OdScale.set( BScale, BScale, BScale );
    pOdInsert->setScaleFactors( OdScale );
    pOdInsert->setPosition( OdPnt );
    pOdInsert->setRotation( BAngle );
  }else{
    // simple point
    OdDbPointPtr pOdPoint = OdDbPoint::createObject();
    pOdPoint->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPoint );
    WriteEntityBase( hPoint, pOdDb, pOdPoint ); 
    pOdPoint->setPosition( OdPnt );
  }
  if (hTextStyle && (0 < TextLen && TextLen < 254)){
    OdDbTextPtr pOdText = OdDbText::createObject();
    pOdText->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdText );
    WriteEntityBase( hPoint, pOdDb, pOdText );
    wcscpy( szOdDbHandle, lcPropGetStr( hTextStyle, LC_PROP_TABLE_STR ));
    if (szOdDbHandle[0]){
      OdHandle = OdDbHandle( szOdDbHandle );
      pOdText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
    } 
    Tdx = lcPropGetFloat( hPoint, LC_PROP_POINT_TDX );
    Tdy = lcPropGetFloat( hPoint, LC_PROP_POINT_TDY );
    TAngle  = lcPropGetFloat( hPoint, LC_PROP_POINT_TANGLE );
    THeight = lcPropGetFloat( hStyle, LC_PROP_PSTYLE_TH );
    TWScale = lcPropGetFloat( hStyle, LC_PROP_PSTYLE_TWS );
    wcscpy( szText, lcPropGetStr( hPoint, LC_PROP_POINT_TEXT ));
    OdPnt.set( OdPnt.x+Tdx, OdPnt.y+Tdy, OdPnt.z );
    pOdText->setPosition( OdPnt );
    pOdText->setHeight( THeight );
    pOdText->setWidthFactor( TWScale );
    pOdText->setTextString( szText );
    pOdText->setOblique( 0. );
    pOdText->setRotation( TAngle );
    pOdText->setHorizontalMode( OdDb::kTextLeft );
    pOdText->setVerticalMode( OdDb::kTextBase );
  }
*/
}




//-----------------------------------------------
void Write_ImageRef (HANDLE hImageRef, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szOdDbHandle[32];
  HANDLE hImage = lcPropGetHandle( hImageRef, LC_PROP_IMGREF_IMAGE );
  double scx = lcPropGetFloat( hImageRef, LC_PROP_IMGREF_SCALE );
  double scy = lcPropGetFloat( hImageRef, LC_PROP_IMGREF_SCALE );
  OdGePoint3d        odpt;
  OdGePoint3d        origin;
  OdGeVector3d       uCorner;
  OdGeVector3d       vOnPlane;
  OdDbHandle         OdHandle;
  OdDbRasterImagePtr pOdImage;

  if (hImage == NULL){
    return;
  }
  pOdImage = OdDbRasterImage::createObject();
  pOdImage->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdImage );
  WriteEntityBase( hImageRef, pOdDb, pOdImage );
//  origin.x = lcPropGetFloat( hImageRef, LC_PROP_IMGREF_XMIN );
//  origin.y = lcPropGetFloat( hImageRef, LC_PROP_IMGREF_YMIN );
  origin.x = lcPropGetFloat( hImageRef, LC_PROP_ENT_XMIN );
  origin.y = lcPropGetFloat( hImageRef, LC_PROP_ENT_YMIN );
  origin.z = 0.;
  uCorner.set( scx, 0., 0. );
  vOnPlane.set( 0., scy, 0. );
  pOdImage->setOrientation( origin, uCorner, vOnPlane );
  pOdImage->setBrightness( OdUInt8(50) );
  pOdImage->setContrast( OdUInt8(50) );
  pOdImage->setDisplayOpt( OdDbRasterImage::kShow, true );
  pOdImage->setDisplayOpt(OdDbRasterImage::kShowUnAligned, true );
  pOdImage->setClipBoundaryToWholeImage();
  wcscpy( szOdDbHandle, lcPropGetStr( hImage, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdImage->setImageDefId( pOdDb->getOdDbObjectId( OdHandle ) );
  }
//    pOdImage->setClipBoundary(const OdGePoint2dArray& clipPoints)
//    pOdImage->setClipped( false );
}


//-----------------------------------------------
void Write_Viewport (HANDLE hVport, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdDbViewportPtr pOdVport;
  CGePoint        ViewPt;
  OdGePoint3d     TargetPt;
  double  Lef, Bot, W, H, Vxc, Vyc, Vscale, Vangle, Vh;

  Lef = lcPropGetFloat( hVport, LC_PROP_VPORT_LEF );
  Bot = lcPropGetFloat( hVport, LC_PROP_VPORT_BOT );
  W   = lcPropGetFloat( hVport, LC_PROP_VPORT_W );
  H   = lcPropGetFloat( hVport, LC_PROP_VPORT_H );
  Vxc = lcPropGetFloat( hVport, LC_PROP_VPORT_VX );
  Vyc = lcPropGetFloat( hVport, LC_PROP_VPORT_VY );
  Vscale = lcPropGetFloat( hVport, LC_PROP_VPORT_VSCALE );
  Vangle = lcPropGetFloat( hVport, LC_PROP_VPORT_VANGLE );

  pOdVport = OdDbViewport::createObject();
  pOdVport->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdVport );
  WriteEntityBase( hVport, pOdDb, pOdVport );
  pOdVport->setWidth( W );
  pOdVport->setHeight( H );
  pOdVport->setCenterPoint( OdGePoint3d( Lef+W/2.0, Bot+H/2.0, 0.0 ));
  TargetPt.x = Vxc;
  TargetPt.y = Vyc;
  TargetPt.z = 0.0;
  pOdVport->setViewTarget( TargetPt );
  Vh = Vscale * H;
  pOdVport->setViewHeight( Vh );
  pOdVport->setTwistAngle( Vangle );
/*
  hView = lcPropGetHandle( hVport, LC_PROP_VPORT_VIEW );
  ViewPt.x = lcPropGetFloat( hVport, LC_PROP_VPORT_VX );
  ViewPt.y = lcPropGetFloat( hVport, LC_PROP_VPORT_VY );
  ViewH = lcPropGetFloat( hVport, LC_PROP_VPORT_VH );
//  lcCoordViewToModel( hView, ViewPt.x, ViewPt.y, &TargetPt.x, &TargetPt.y, &TargetPt.z );
  TargetPt.x = ViewPt.x;
  TargetPt.y = ViewPt.y;
  TargetPt.z = 0.0;
  ViewVec.x = lcPropGetFloat( hView, LC_PROP_VIEW_PROJX );
  ViewVec.y = lcPropGetFloat( hView, LC_PROP_VIEW_PROJY );
  ViewVec.z = lcPropGetFloat( hView, LC_PROP_VIEW_PROJZ );

  pOdVport = OdDbViewport::createObject();
  pOdVport->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdVport );
  WriteEntityBase( hVport, pOdDb, pOdVport );
  pOdVport->setWidth( W );
  pOdVport->setHeight( H );
  pOdVport->setCenterPoint( OdGePoint3d( Lef+W/2., Bot+H/2., 0. ) );
//  TargetPt.x = TargetPt.y = TargetPt.z = 0.;
  pOdVport->setViewTarget( TargetPt );
  pOdVport->setViewHeight( ViewH );
  pOdVport->setViewDirection( ViewVec );
*/
//  pOdVport->setViewCenter( OdGePoint2d(0,0) );
//  pOdVport->zoomExtents();
}


//-----------------------------------------------
void Write_Hatch (HANDLE hHatch, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szName[256];
  int    i, n;
  int    j, m;
  double len;
  double Scale  = lcPropGetFloat( hHatch, LC_PROP_HATCH_SCALE );
  double Angle  = lcPropGetFloat( hHatch, LC_PROP_HATCH_ANGLE );
  BOOL   bSolid = lcPropGetBool( hHatch, LC_PROP_HATCH_SOLID );
  int    nPaths = lcPropGetInt( hHatch, LC_PROP_HATCH_NLOOP );
//  LcGePolyline  Pline;
//  LcGePoint     pnt;
  OdGePoint2d   pt2d;
  OdDbHatchPtr  pOdHatch;

  memset( szName, 0, sizeof(szName) );
  if (bSolid){
//    strcpy( szName, "SOLID" );
  }else{
    wcscpy( szName, lcPropGetStr( hHatch, LC_PROP_HATCH_NAME ));
    n = (int)wcslen( szName );
    for (i=0; i<n; i++){
      if (szName[i] == L','){
        szName[i] = 0;
        break;
      }
    }
    if (wcsicmp(szName, L"SOLID")==0){
      wcscpy( szName, L"SOLID2" );
    }

    // adding our pattern (szName)
    OdHatchPattern     pattern;
    OdHatchPatternLine line;

    n = lcPropGetInt( hHatch, LC_PROP_HATCH_NHPL );
    for (i=0; i<n; i++){
      lcPropPutInt( hHatch, LC_PROP_HATCH_IHPL, i );
      line.m_dLineAngle      = lcPropGetFloat( hHatch, LC_PROP_HPL_ANGLE );
      line.m_basePoint.x     = lcPropGetFloat( hHatch, LC_PROP_HPL_X0 );
      line.m_basePoint.y     = lcPropGetFloat( hHatch, LC_PROP_HPL_Y0 );
      line.m_patternOffset.x = lcPropGetFloat( hHatch, LC_PROP_HPL_DX );
      line.m_patternOffset.y = lcPropGetFloat( hHatch, LC_PROP_HPL_DY );
      line.m_dashes.clear();
      m = lcPropGetInt( hHatch, LC_PROP_HPL_NDASH );
      for (j=0; j<m; j++) {
        len = lcPropGetFloat( hHatch, LC_PROP_HPL_DASH1+j );
        line.m_dashes.push_back( len );
      }
      pattern.push_back( line );
    }
    pOdDb->appServices()->patternManager()->appendPattern( OdDbHatch::kCustomDefined, szName, pattern ); 
  }
  pOdHatch = OdDbHatch::createObject();
  pOdHatch->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdHatch );
  WriteEntityBase( hHatch, pOdDb, pOdHatch ); 

  pOdHatch->setAssociative( false );
  pOdHatch->setElevation( 0. );
  pOdHatch->setHatchStyle( OdDbHatch::kNormal );
  if (bSolid){
    pOdHatch->setPattern( OdDbHatch::kPreDefined, L"SOLID" );
  }else{
    pOdHatch->setPattern( OdDbHatch::kCustomDefined, szName );
    pOdHatch->setPatternAngle( Angle );
    pOdHatch->setPatternScale( Scale );
  }

  int    k;
  double x, y;
  OdGePoint2dArray vertexPts;
  OdGeDoubleArray vertexBulges;
  nPaths = lcPropGetInt( hHatch, LC_PROP_HATCH_NLOOP );
  k = 0;
  for (i=0; i<nPaths; ++i){
    n = lcHatchGetLoopSize( hHatch, i );
    vertexPts.resize( n );
    for (j=0; j<n; ++j) {
      lcHatchGetPoint( hHatch, k, &x, &y );
      ++k;
      vertexPts[j].set( x, y );
    }
    pOdHatch->appendLoop( OdDbHatch::kExternal | OdDbHatch::kPolyline, vertexPts, vertexBulges );
  }
}


//-----------------------------------------------
void Write_RPlan (HANDLE hRPlan, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
}



/*
//-----------------------------------------------
void Write_Hatch (const LcDbHatch* pHatch, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  LcChar   szName[256];
  LcInt    i, j, n;
  LcDouble len;
  double Scale = pHatch->PatScale();
  double Angle = pHatch->PatAngle();
  BOOL  bSolid = pHatch->IsSolid();
  int           nPaths   = pHatch->NumPaths();
  LcGePolyline Pline;
  LcGePoint    pnt;
  OdGePoint2d   pt2d;
  OdDbHatchPtr  pOdHatch;

  memset( szName, 0, sizeof(szName) );
  if (bSolid){
//    strcpy( szName, "SOLID" );
  }else{
    wcscpy( szName, pHatch->PatName() );
    n = wcslen( szName );
    for (i=0; i<n; i++){
      if (szName[i] == L','){
        szName[i] = 0;
        break;
      }
    }
    if (wcsicmp(szName, L"SOLID")==0){
      wcscpy( szName, L"SOLID2" );
    }

    // adding our pattern (szName)
    OdHatchPattern     pattern;
    OdHatchPatternLine line;
    LC_HPLINE PatLine;

    n = pHatch->NumPatLines();
    for (i=0; i<n; i++){
      pHatch->GetPatLine( i, &PatLine );
      line.m_dLineAngle = PatLine.Angle;
      line.m_basePoint = OdGePoint2d( PatLine.x0, PatLine.y0 );
      line.m_patternOffset = OdGeVector2d( PatLine.dx, PatLine.dy );
      line.m_dashes.clear();
      for (j=0; j<PatLine.nDash; j++) {
        len = PatLine.Dash[j]; 
        line.m_dashes.push_back( len );
      }
      pattern.push_back( line );
    }
    pOdDb->appServices()->patternManager()->appendPattern( OdDbHatch::kCustomDefined, szName, pattern ); 
  }

  pOdHatch = OdDbHatch::createObject();
  pOdHatch->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdHatch );
  WriteEntityBase( pHatch, pOdDb, pOdHatch ); 

  pOdHatch->setAssociative( false );
  pOdHatch->setElevation( 0. );
  pOdHatch->setHatchStyle( OdDbHatch::kNormal );
  if (bSolid){
    pOdHatch->setPattern( OdDbHatch::kPreDefined, L"SOLID" );
  }else{
    pOdHatch->setPattern( OdDbHatch::kCustomDefined, szName );
    pOdHatch->setPatternAngle( Angle );
    pOdHatch->setPatternScale( Scale );
  }

  OdGePoint2dArray vertexPts;
  OdGeDoubleArray vertexBulges;

  for (i=0; i<nPaths; i++){
    pHatch->GetPath( i, &Pline );
    n = Pline.NumPoints();  //pHatch->PathSize(i) - 1;
    vertexPts.resize( n );
    for (j=0; j<n; j++) {
      Pline.FastGetPointAt( j, &pnt );
      vertexPts[j].set( pnt.x, pnt.y );
    }
    pOdHatch->appendLoop( OdDbHatch::kExternal | OdDbHatch::kPolyline, vertexPts, vertexBulges );
  }
}
*/

//-----------------------------------------------
void Write_Face3D (HANDLE hFace, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdGePoint3d OdPnt[4];
  BOOL bEdge[4];
  int  i;
  OdDbFacePtr pOdFace = OdDbFace::createObject();

  OdPnt[0].x = lcPropGetFloat( hFace, LC_PROP_FACE_X0 );
  OdPnt[0].y = lcPropGetFloat( hFace, LC_PROP_FACE_Y0 );
  OdPnt[0].z = lcPropGetFloat( hFace, LC_PROP_FACE_Z0 );
  OdPnt[1].x = lcPropGetFloat( hFace, LC_PROP_FACE_X1 );
  OdPnt[1].y = lcPropGetFloat( hFace, LC_PROP_FACE_Y1 );
  OdPnt[1].z = lcPropGetFloat( hFace, LC_PROP_FACE_Z1 );
  OdPnt[2].x = lcPropGetFloat( hFace, LC_PROP_FACE_X2 );
  OdPnt[2].y = lcPropGetFloat( hFace, LC_PROP_FACE_Y2 );
  OdPnt[2].z = lcPropGetFloat( hFace, LC_PROP_FACE_Z2 );
  OdPnt[3].x = lcPropGetFloat( hFace, LC_PROP_FACE_X3 );
  OdPnt[3].y = lcPropGetFloat( hFace, LC_PROP_FACE_Y3 );
  OdPnt[3].z = lcPropGetFloat( hFace, LC_PROP_FACE_Z3 );
  bEdge[0] = lcPropGetBool( hFace, LC_PROP_FACE_EDGE1 );
  bEdge[1] = lcPropGetBool( hFace, LC_PROP_FACE_EDGE2 );
  bEdge[2] = lcPropGetBool( hFace, LC_PROP_FACE_EDGE3 );
  bEdge[3] = lcPropGetBool( hFace, LC_PROP_FACE_EDGE4 );
  pOdFace->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdFace );
  WriteEntityBase( hFace, pOdDb, pOdFace ); 
  for (i=0; i<4; ++i){
    pOdFace->setVertexAt( i, OdPnt[i] );
  }
  for (i=0; i<4; ++i){
    if (bEdge[i]){
      pOdFace->makeEdgeVisibleAt( i );
    }else{
      pOdFace->makeEdgeInvisibleAt( i );
    }
  }
}

//-----------------------------------------------
void Write_Shape (HANDLE hShape, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  Write_Exploded( hShape, pOdDb, pOdBlock );
}

//-----------------------------------------------
void Write_Exploded (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  int      iVer, i, nVers;
  int      nPlines;
  double   X,Y; 
  OdGePoint2d     OdPt2d; 
  OdDbPolylinePtr pOdPolyline;

  //--------------------------------------------- LC_PLFIT_BULGE
  nPlines = lcExpEntity( hEnt, 0, LC_EXP_OUTLINE | LC_EXP_HATCH, false );
  if (nPlines > 0){
    for (i=0; i<nPlines; ++i){
      pOdPolyline = OdDbPolyline::createObject();
      pOdPolyline->setDatabaseDefaults( pOdDb );
      pOdBlock->appendOdDbEntity( pOdPolyline );
      WriteEntityBase( hEnt, pOdDb, pOdPolyline ); 
      nVers = lcExpGetPline( 0.0 );
      for (iVer=0; iVer<nVers; ++iVer){
        lcExpGetVertex( &X, &Y );
        OdPt2d.x = X;
        OdPt2d.y = Y;
        pOdPolyline->addVertexAt( iVer, OdPt2d );
      }
//      pOdPolyline->setClosed( bClosed?true:false );
      pOdPolyline->setPlinegen( true );
    }
  }
}





