/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
* Arc, Circle, Ellipse
*********************************************************************/
#include "pch.h"

//-----------------------------------------------
void Write_Arc (HANDLE hArc, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
//  double      x, y, z;
  WCHAR  szColor[32];
  BOOL   bFilled = false; //lcPropGetBool( hArc, LC_PROP_ARC_FILLED );
  BOOL   bBorder = false; //lcPropGetBool( hArc, LC_PROP_ARC_BORDER );
  BOOL   bSector = lcPropGetBool( hArc, LC_PROP_ARC_SECTOR );
  double AngStart, AngEnd, AngArc, Z;
  OdGePoint3d     Center;
  OdGePoint2d     OdPnt; 
  OdCmColor       OdColor;
  OdDbPolylinePtr pOdPline;
  OdDbArcPtr      pOdArc = OdDbArc::createObject();

  pOdArc->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdArc );
  WriteEntityBase( hArc, pOdDb, pOdArc ); 
  Center.x = lcPropGetFloat( hArc, LC_PROP_ARC_X );
  Center.y = lcPropGetFloat( hArc, LC_PROP_ARC_Y );
  Center.z = lcPropGetFloat( hArc, LC_PROP_ENT_Z );
  pOdArc->setCenter( Center );
  pOdArc->setRadius( lcPropGetFloat( hArc, LC_PROP_ARC_RAD ) );
  AngStart = lcPropGetFloat( hArc, LC_PROP_ARC_ANG0 );
  AngEnd   = lcPropGetFloat( hArc, LC_PROP_ARC_ANGEND );
  AngArc   = lcPropGetFloat( hArc, LC_PROP_ARC_ANGARC );
  if (AngArc > 0.){
    pOdArc->setStartAngle( AngStart );
    pOdArc->setEndAngle( AngEnd );
  }else{
    pOdArc->setStartAngle( AngEnd );
    pOdArc->setEndAngle( AngStart );
  }
  if (bSector){
    pOdPline = OdDbPolyline::createObject();
    pOdBlock->appendOdDbEntity( pOdPline );
    WriteEntityBase( hArc, pOdDb, pOdPline ); 
    OdPnt.x = lcPropGetFloat( hArc, LC_PROP_ARC_X0 );
    OdPnt.y = lcPropGetFloat( hArc, LC_PROP_ARC_Y0 );
    pOdPline->addVertexAt( 0, OdPnt );
    OdPnt.x = lcPropGetFloat( hArc, LC_PROP_ARC_X );
    OdPnt.y = lcPropGetFloat( hArc, LC_PROP_ARC_Y );
    pOdPline->addVertexAt( 0, OdPnt );
    OdPnt.x = lcPropGetFloat( hArc, LC_PROP_ARC_XEND );
    OdPnt.y = lcPropGetFloat( hArc, LC_PROP_ARC_YEND );
    pOdPline->addVertexAt( 0, OdPnt );
  }
  if (bFilled){
    if (bBorder){
//      wcscpy( szColor, lcPropGetStr( hArc, LC_PROP_ENT_BCOLOR ) );
      wcscpy( szColor, lcPropGetStr( hArc, LC_PROP_ENT_FCOLOR ) );
      MakeOdColor( szColor, &OdColor );
      pOdArc->setColor( OdColor, false );
      pOdPline->setColor( OdColor, false );
    }
    // inside filling (as solid hatch)
    OdDbObjectIdArray loopIds;
    OdDbHatchPtr pOdHatch = OdDbHatch::createObject();
    pOdBlock->appendOdDbEntity( pOdHatch );
    WriteEntityBase( hArc, pOdDb, pOdHatch ); 
    loopIds.push_back( pOdArc->objectId() );
    loopIds.push_back( pOdPline->objectId() );
    pOdHatch->setAssociative( true );
    Z = lcPropGetFloat( hArc, LC_PROP_ENT_Z );
    pOdHatch->setElevation( Z );
    pOdHatch->setHatchStyle( OdDbHatch::kNormal );
    pOdHatch->setPattern( OdDbHatch::kPreDefined, "SOLID" );
    pOdHatch->appendLoop( OdDbHatch::kDefault, loopIds );
    // swap objects
    OdDbSortentsTablePtr pOdSort = pOdBlock->getSortentsTable();
    pOdSort->swapOrder( pOdArc->objectId(), pOdHatch->objectId() );
  }
}


//-----------------------------------------------
void Write_Circle (HANDLE hCirc, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szColor[32];
  double Z;
  BOOL   bFilled = false; //lcPropGetBool( hCirc, LC_PROP_CIRCLE_FILLED );
  BOOL   bBorder = false; //lcPropGetBool( hCirc, LC_PROP_CIRCLE_BORDER );
  OdGePoint3d   Center;
  OdCmColor     OdColor;
  OdDbCirclePtr pOdCircle = OdDbCircle::createObject();

  pOdCircle->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdCircle );
  WriteEntityBase( hCirc, pOdDb, pOdCircle ); 
  Center.x = lcPropGetFloat( hCirc, LC_PROP_CIRCLE_X );
  Center.y = lcPropGetFloat( hCirc, LC_PROP_CIRCLE_Y );
  Center.z = lcPropGetFloat( hCirc, LC_PROP_ENT_Z );
  pOdCircle->setCenter( Center );
  pOdCircle->setRadius( lcPropGetFloat( hCirc, LC_PROP_CIRCLE_RAD ) ); 

  if (bFilled){
    if (bBorder){
//      wcscpy( szColor, lcPropGetStr( hCirc, LC_PROP_ENT_BCOLOR ) );
      wcscpy( szColor, lcPropGetStr( hCirc, LC_PROP_ENT_COLOR ) );
      MakeOdColor( szColor, &OdColor );
      pOdCircle->setColor( OdColor, false );
    }
    // inside filling (as solid hatch)
    OdDbObjectIdArray loopIds;
    OdDbHatchPtr pOdHatch = OdDbHatch::createObject();
    pOdBlock->appendOdDbEntity( pOdHatch );
    WriteEntityBase( hCirc, pOdDb, pOdHatch ); 
    loopIds.push_back( pOdCircle->objectId() );
    pOdHatch->setAssociative( true );
    Z = lcPropGetFloat( hCirc, LC_PROP_ENT_Z );
    pOdHatch->setElevation( Z );
    pOdHatch->setHatchStyle( OdDbHatch::kNormal );
    pOdHatch->setPattern( OdDbHatch::kPreDefined, "SOLID" );
    pOdHatch->appendLoop( OdDbHatch::kDefault, loopIds );
    // swap objects
    OdDbSortentsTablePtr pOdSort = pOdBlock->getSortentsTable();
    pOdSort->swapOrder( pOdCircle->objectId(), pOdHatch->objectId() );
  }
}


//-----------------------------------------------
void Write_Ellipse (HANDLE hEll, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szColor[32];
  BOOL   bFilled = false; //lcPropGetBool( hEll, LC_PROP_ELL_FILLED );
  BOOL   bBorder = false; //lcPropGetBool( hEll, LC_PROP_ELL_BORDER );
  BOOL   bSector = lcPropGetBool( hEll, LC_PROP_ELL_SECTOR );
  double          Rv, Rh, Ang0, AngArc, AngRot, startAngle, endAngle, radiusRatio, Z;
  bool            bCW = false;
  CGePoint        Pt1;
  OdGePoint2d     OdPnt; 
  OdCmColor       OdColor;
  OdDbPolylinePtr pOdPline;
  OdGePoint3d     Center;
  OdGeVector3d    Normal;
  OdGeVector3d    majorAxis;
  OdDbEllipsePtr  pOdEllipse = OdDbEllipse::createObject();

  pOdEllipse->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdEllipse );
  WriteEntityBase( hEll, pOdDb, pOdEllipse );
  Center.x = lcPropGetFloat( hEll, LC_PROP_ELL_X );
  Center.y = lcPropGetFloat( hEll, LC_PROP_ELL_Y );
  Center.z = lcPropGetFloat( hEll, LC_PROP_ENT_Z );
  Rh       = lcPropGetFloat( hEll, LC_PROP_ELL_R1 );
  Rv       = lcPropGetFloat( hEll, LC_PROP_ELL_R2 );
  AngRot   = lcPropGetFloat( hEll, LC_PROP_ELL_ANGLE );
  Ang0     = lcPropGetFloat( hEll, LC_PROP_ELL_ANG0 );
  AngArc   = lcPropGetFloat( hEll, LC_PROP_ELL_ANGARC );
  Pt1.Set( 0.0, 0.0 );
  if (Rh > Rv){
    Pt1.Polar( AngRot, Rh );
  }else{
    Pt1.Polar( AngRot+GE_DEG90, Rv );
  }
  majorAxis.x = Pt1.x;
  majorAxis.y = Pt1.y;
  majorAxis.z = 0.0;
  if (Rh > Rv){
    radiusRatio = Rv / Rh;
  }else{
    radiusRatio = Rh / Rv;
    Ang0 -= GE_DEG90;
  }
  if (AngArc < 0.0){
    bCW = true;
    AngArc = -AngArc;
  }
  if (0.01 < AngArc && AngArc < (GE_DEG360-0.01)){
    if (bCW){
      startAngle = Ang0 + AngArc;
      endAngle = Ang0;
    }else{
      startAngle = Ang0;
      endAngle = Ang0 + AngArc;
    }
  }else{
    startAngle = 0.0;
    endAngle = GE_DEG360;
  }
  Normal.set( 0, 0, 1 );
  pOdEllipse->set( Center, Normal, majorAxis, radiusRatio, startAngle, endAngle );
  if (bSector){
    pOdPline = OdDbPolyline::createObject();
    pOdBlock->appendOdDbEntity( pOdPline );
    WriteEntityBase( hEll, pOdDb, pOdPline ); 
    OdPnt.x = lcPropGetFloat( hEll, LC_PROP_ELL_X0 );
    OdPnt.y = lcPropGetFloat( hEll, LC_PROP_ELL_Y0 );
    pOdPline->addVertexAt( 0, OdPnt );
    OdPnt.x = lcPropGetFloat( hEll, LC_PROP_ELL_X );
    OdPnt.y = lcPropGetFloat( hEll, LC_PROP_ELL_Y );
    pOdPline->addVertexAt( 0, OdPnt );
    OdPnt.x = lcPropGetFloat( hEll, LC_PROP_ELL_XEND );
    OdPnt.y = lcPropGetFloat( hEll, LC_PROP_ELL_YEND );
    pOdPline->addVertexAt( 0, OdPnt );
  }
  if (bFilled){
    if (bBorder){
//      wcscpy( szColor, lcPropGetStr( hEll, LC_PROP_ENT_BCOLOR ) );
      wcscpy( szColor, lcPropGetStr( hEll, LC_PROP_ENT_FCOLOR ) );
      MakeOdColor( szColor, &OdColor );
      pOdEllipse->setColor( OdColor, false );
      if (bSector){
        pOdPline->setColor( OdColor, false );
      }
    }
    // inside filling (as solid hatch)
    OdDbObjectIdArray loopIds;
    OdDbHatchPtr pOdHatch = OdDbHatch::createObject();
    pOdBlock->appendOdDbEntity( pOdHatch );
    WriteEntityBase( hEll, pOdDb, pOdHatch ); 
    loopIds.push_back( pOdEllipse->objectId() );
    if (bSector){
      loopIds.push_back( pOdPline->objectId() );
    }
    pOdHatch->setAssociative( true );
    Z = lcPropGetFloat( hEll, LC_PROP_ENT_Z );
    pOdHatch->setElevation( Z );
    pOdHatch->setHatchStyle( OdDbHatch::kNormal );
    pOdHatch->setPattern( OdDbHatch::kPreDefined, "SOLID" );
    pOdHatch->appendLoop( OdDbHatch::kDefault, loopIds );
    // swap objects
    OdDbSortentsTablePtr pOdSort = pOdBlock->getSortentsTable();
    pOdSort->swapOrder( pOdEllipse->objectId(), pOdHatch->objectId() );
  }
}
