/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
* Block refs and attributes
*********************************************************************/
#include "pch.h"

//-----------------------------------------------
void Write_Attdef (HANDLE hAttdef, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szBuf[256], szOdDbHandle[32];
  HANDLE hTStyle;
  double Height, WScale, Angle, Oblique; 
  int    Align, Mode;
  bool   b;
  OdGePoint3d  Pt0, Pta;
  OdDbAttributeDefinitionPtr pOdAttDef;
  OdDbHandle   OdHandle;

  pOdAttDef = OdDbAttributeDefinition::createObject();
  pOdAttDef->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdAttDef );
  WriteEntityBase( hAttdef, pOdDb, pOdAttDef );

  // text style
  hTStyle = lcPropGetHandle( hAttdef, LC_PROP_ATT_TSTYLE );
  wcscpy( szOdDbHandle, lcPropGetStr( hTStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdAttDef->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  // prompt
  wcscpy( szBuf, lcPropGetStr( hAttdef, LC_PROP_ATT_PROMPT ));
  pOdAttDef->setPrompt( szBuf );
  // tag
  wcscpy( szBuf, lcPropGetStr( hAttdef, LC_PROP_ATT_TAG ));
  pOdAttDef->setTag( szBuf );
  // default value
  wcscpy( szBuf, lcPropGetStr( hAttdef, LC_PROP_ATT_VALUE ));
  pOdAttDef->setTextString( szBuf );
  // Flags
  Mode = lcPropGetInt( hAttdef, LC_PROP_ATT_MODE );
  b = (Mode & LC_ATTRIB_HIDDEN)? true : false;
  pOdAttDef->setInvisible( b );
  b = (Mode & LC_ATTRIB_CONST)? true : false;
  pOdAttDef->setConstant( b );
  b = (Mode & LC_ATTRIB_VERIFY)? true : false;
  pOdAttDef->setVerifiable( b );
  b = (Mode & LC_ATTRIB_PRESET)? true : false;
  pOdAttDef->setPreset( b );
  b = (Mode & LC_ATTRIB_LOCK)? true : false;
  pOdAttDef->setLockPositionInBlock( b );
  b = (Mode & LC_ATTRIB_MTEXT)? true : false;
//  pOdAttDef->setMTextAttributeDefinition( b );
  // insertion point
  Pt0.x = lcPropGetFloat( hAttdef, LC_PROP_ATT_X0 ); 
  Pt0.y = lcPropGetFloat( hAttdef, LC_PROP_ATT_Y0 ); 
  Pt0.z = lcPropGetFloat( hAttdef, LC_PROP_ENT_Z );
  pOdAttDef->setPosition( Pt0 );
  // alignment point
  Pta.x = lcPropGetFloat( hAttdef, LC_PROP_ATT_X ); 
  Pta.y = lcPropGetFloat( hAttdef, LC_PROP_ATT_Y ); 
  Pta.z = lcPropGetFloat( hAttdef, LC_PROP_ENT_Z );  
  pOdAttDef->setAlignmentPoint( Pta );
  // text height
  Height = lcPropGetFloat( hAttdef, LC_PROP_ATT_H );
  pOdAttDef->setHeight( Height );
  // width scale
  WScale = lcPropGetFloat( hAttdef, LC_PROP_ATT_WSCALE ); 
  pOdAttDef->setWidthFactor( WScale );
  // oblique angle
  Oblique = lcPropGetFloat( hAttdef, LC_PROP_ATT_OBLIQUE ); 
  pOdAttDef->setOblique( Oblique );
  // rotation angle
  Angle = lcPropGetFloat( hAttdef, LC_PROP_ATT_ANGLE ); 
  pOdAttDef->setRotation( Angle );
  // alignment
  Align = lcPropGetInt( hAttdef, LC_PROP_ATT_ALIGN ); 
  switch( Align ){
    case  LC_TA_LEFBOT:
      pOdAttDef->setHorizontalMode( OdDb::kTextLeft );
      pOdAttDef->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_LEFCEN:
      pOdAttDef->setHorizontalMode( OdDb::kTextLeft );
      pOdAttDef->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_LEFTOP:
      pOdAttDef->setHorizontalMode( OdDb::kTextLeft );
      pOdAttDef->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_CENBOT:
      pOdAttDef->setHorizontalMode( OdDb::kTextCenter );
      pOdAttDef->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_CENTER:
      pOdAttDef->setHorizontalMode( OdDb::kTextCenter );
      pOdAttDef->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_CENTOP:
      pOdAttDef->setHorizontalMode( OdDb::kTextCenter );
      pOdAttDef->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_RIGBOT:
      pOdAttDef->setHorizontalMode( OdDb::kTextRight );
      pOdAttDef->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_RIGCEN:
      pOdAttDef->setHorizontalMode( OdDb::kTextRight );
      pOdAttDef->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_RIGTOP:
      pOdAttDef->setHorizontalMode( OdDb::kTextRight );
      pOdAttDef->setVerticalMode( OdDb::kTextTop );
      break;
  }
}

//-----------------------------------------------
void Write_Attrib (HANDLE hAtt, OdDbAttributePtr pOdAtt, OdGePoint3d OdInsPos, 
                   OdGeScale3d BlkScale, double BlkAngle, OdDbDatabasePtr pOdDb)
{
  OdGePoint3d  OdAttPos;
//  OdGeMatrix3d OdBlkXfm;
  OdDbHandle   OdHandle;
  OdCmColor    OdColor;
  OdDb::LineWeight OdLWeight = (OdDb::LineWeight)25;
  WCHAR        szOdDbHandle[32], szColor[32];
  HANDLE hLayer, hLtype, hStyle;
  WCHAR  szBuf[256];
  double Height, WScale, Oblique, Angle, Ratio, LtScale;
  int    AttMode, Align, Lweight;
  bool   b;

  // color
  wcscpy( szColor, lcPropGetStr( hAtt, LC_PROP_ATT_COLOR ));
  MakeOdColor( szColor, &OdColor );
  pOdAtt->setColor( OdColor, false );

  // Layer
  hLayer = lcPropGetHandle( hAtt, LC_PROP_ATT_LAYER ); 
  wcscpy( szOdDbHandle, lcPropGetStr( hLayer, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdAtt->setLayer( pOdDb->getOdDbObjectId( OdHandle ) );
  }

  // linetype
  hLtype = lcPropGetHandle( hAtt, LC_PROP_ATT_LINETYPE );
  wcscpy( szOdDbHandle, lcPropGetStr( hLtype, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdAtt->setLinetype( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  // linetype stale
  LtScale = lcPropGetFloat( hAtt, LC_PROP_ATT_LTSCALE );
  pOdAtt->setLinetypeScale( LtScale );

  // lineweight
  Lweight = lcPropGetInt( hAtt, LC_PROP_ATT_LWIDTH );
  if (Lweight == LC_LWIDTH_DEFAULT){
    OdLWeight = OdDb::kLnWtByLwDefault;
  }else
  if (Lweight == LC_LWIDTH_BYLAYER){
    OdLWeight = OdDb::kLnWtByLayer;
  }else
  if (Lweight == LC_LWIDTH_BYBLOCK){
    OdLWeight = OdDb::kLnWtByBlock;
  }else
  if (0<=Lweight && Lweight<25){
    OdLWeight = OdDb::kLnWt000;
  }else
  if (25<=Lweight && Lweight<70){
    OdLWeight = OdDb::kLnWt005;
  }else
  if (70<=Lweight && Lweight<110){
    OdLWeight = OdDb::kLnWt009;
  }else
  if (110<=Lweight && Lweight<140){
    OdLWeight = OdDb::kLnWt013;
  }else
  if (110<=Lweight && Lweight<140){
    OdLWeight = OdDb::kLnWt013;
  }else
  if (140<=Lweight && Lweight<165){
    OdLWeight = OdDb::kLnWt015;
  }else
  if (165<=Lweight && Lweight<190){
    OdLWeight = OdDb::kLnWt018;
  }else
  if (190<=Lweight && Lweight<225){
    OdLWeight = OdDb::kLnWt020;
  }else
  if (225<=Lweight && Lweight<275){
    OdLWeight = OdDb::kLnWt025;
  }else
  if (275<=Lweight && Lweight<325){
    OdLWeight = OdDb::kLnWt030;
  }else
  if (325<=Lweight && Lweight<375){
    OdLWeight = OdDb::kLnWt035;
  }else
  if (375<=Lweight && Lweight<450){
    OdLWeight = OdDb::kLnWt040;
  }else
  if (450<=Lweight && Lweight<515){
    OdLWeight = OdDb::kLnWt050;
  }else
  if (515<=Lweight && Lweight<565){
    OdLWeight = OdDb::kLnWt053;
  }else
  if (565<=Lweight && Lweight<650){
    OdLWeight = OdDb::kLnWt060;
  }else
  if (650<=Lweight && Lweight<750){
    OdLWeight = OdDb::kLnWt070;
  }else
  if (750<=Lweight && Lweight<850){
    OdLWeight = OdDb::kLnWt080;
  }else
  if (850<=Lweight && Lweight<950){
    OdLWeight = OdDb::kLnWt090;
  }else
  if (950<=Lweight && Lweight<1030){
    OdLWeight = OdDb::kLnWt100;
  }else
  if (1030<=Lweight && Lweight<1130){
    OdLWeight = OdDb::kLnWt106;
  }else
  if (1130<=Lweight && Lweight<1300){
    OdLWeight = OdDb::kLnWt120;
  }else
  if (1300<=Lweight && Lweight<1490){
    OdLWeight = OdDb::kLnWt140;
  }else
  if (1490<=Lweight && Lweight<1790){
    OdLWeight = OdDb::kLnWt158;
  }else
  if (1790<=Lweight && Lweight<2100){
    OdLWeight = OdDb::kLnWt200;
  }else
  if (2100<=Lweight){
    OdLWeight = OdDb::kLnWt211;
  }
  pOdAtt->setLineWeight( OdLWeight );

  // text style
  hStyle = lcPropGetHandle( hAtt, LC_PROP_ATT_TSTYLE );
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdAtt->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }

  // Flags
  AttMode = lcPropGetInt( hAtt, LC_PROP_ATT_MODE );
  b = (AttMode & LC_ATTRIB_HIDDEN)? true : false;
  pOdAtt->setInvisible( b );

  // tag
  wcscpy( szBuf, lcPropGetStr( hAtt, LC_PROP_ATT_TAG ) );
  pOdAtt->setTag( szBuf );

  // value
  wcscpy( szBuf, lcPropGetStr( hAtt, LC_PROP_ATT_VALUE ) );
  pOdAtt->setTextString( szBuf );

  // insertion point
  OdAttPos.x = lcPropGetFloat( hAtt, LC_PROP_ATT_X0 );
  OdAttPos.y = lcPropGetFloat( hAtt, LC_PROP_ATT_Y0 );
  OdAttPos.z = 0.0; //lcPropGetFloat( hAttr, LC_PROP_ATT_Z );
  OdAttPos.x += OdInsPos.x;
  OdAttPos.y += OdInsPos.y;
  pOdAtt->setPosition( OdAttPos );
  // alignment point
  OdAttPos.x = lcPropGetFloat( hAtt, LC_PROP_ATT_X );
  OdAttPos.y = lcPropGetFloat( hAtt, LC_PROP_ATT_Y );
  OdAttPos.x += OdInsPos.x;
  OdAttPos.y += OdInsPos.y;
//    OdPos.z = 0.; //lcPropGetFloat( hAttr, LC_PROP_ATT_Z );
  pOdAtt->setAlignmentPoint( OdAttPos );

  Height  = lcPropGetFloat( hAtt, LC_PROP_ATT_H );
  Height *= BlkScale.sy;
  if (Height < 0.0){
    Height = -Height;
  }
  pOdAtt->setHeight( Height );

  WScale  = lcPropGetFloat( hAtt, LC_PROP_ATT_WSCALE ); 
  Ratio = BlkScale.sy / BlkScale.sx;
  WScale /= Ratio;
  pOdAtt->setWidthFactor( WScale );

  Angle   = lcPropGetFloat( hAtt, LC_PROP_ATT_ANGLE ); 
  Angle += BlkAngle;
  pOdAtt->setRotation( Angle );

  Oblique = lcPropGetFloat( hAtt, LC_PROP_ATT_OBLIQUE ); 
  pOdAtt->setOblique( Oblique );

  Align   = lcPropGetInt( hAtt, LC_PROP_ATT_ALIGN ); 
  switch( Align ){
    case LC_TA_LEFBOT:
      pOdAtt->setHorizontalMode( OdDb::kTextLeft );
      pOdAtt->setVerticalMode( OdDb::kTextBase );
      break;
    case LC_TA_LEFCEN:
      pOdAtt->setHorizontalMode( OdDb::kTextLeft );
      pOdAtt->setVerticalMode( OdDb::kTextVertMid );
      break;
    case LC_TA_LEFTOP:
      pOdAtt->setHorizontalMode( OdDb::kTextLeft );
      pOdAtt->setVerticalMode( OdDb::kTextTop );
      break;
    case LC_TA_CENBOT:
      pOdAtt->setHorizontalMode( OdDb::kTextCenter );
      pOdAtt->setVerticalMode( OdDb::kTextBase );
      break;
    case LC_TA_CENTER:
      pOdAtt->setHorizontalMode( OdDb::kTextCenter );
      pOdAtt->setVerticalMode( OdDb::kTextVertMid );
      break;
    case LC_TA_CENTOP:
      pOdAtt->setHorizontalMode( OdDb::kTextCenter );
      pOdAtt->setVerticalMode( OdDb::kTextTop );
      break;
    case LC_TA_RIGBOT:
      pOdAtt->setHorizontalMode( OdDb::kTextRight );
      pOdAtt->setVerticalMode( OdDb::kTextBase );
      break;
    case LC_TA_RIGCEN:
      pOdAtt->setHorizontalMode( OdDb::kTextRight );
      pOdAtt->setVerticalMode( OdDb::kTextVertMid );
      break;
    case LC_TA_RIGTOP:
      pOdAtt->setHorizontalMode( OdDb::kTextRight );
      pOdAtt->setVerticalMode( OdDb::kTextTop );
      break;
  }
/*
  OdBlkXfm.setTranslation( OdInsPos.asVector() );
//    pOdInsert->transformBy( BlkXfm );
  pOdAtt->transformBy( OdBlkXfm );
*/
}


//-----------------------------------------------
void Write_BlockRef (HANDLE hIns, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szOdDbHandle[32];
  HANDLE hAtt;
  HANDLE hBlock   = lcPropGetHandle( hIns, LC_PROP_BLKREF_BLOCK );
  double BlkAngle = lcPropGetFloat( hIns, LC_PROP_BLKREF_ANGLE );
//  bool   bAttrib = lcPropGetBool( hIns, LC_PROP_BLKREF_ATTRIB );
  int    NumRows = 1;
  int    NumCols = 1;
  double ColDist = 0.0;
  double RowDist = 0.0;
  double Ratio;
  OdDbHandle            OdHandle;
  OdGeScale3d           BlkScale;
  OdGePoint3d           OdInsPos, OdAttPos;
  OdDbBlockReferencePtr pOdInsert;
  OdDbMInsertBlockPtr   pOdMIns;
  OdDbAttributePtr      pOdAttrib;

  if (hBlock == NULL){
    return;
  }
  wcscpy( szOdDbHandle, lcPropGetStr( hBlock, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
  }
  OdInsPos.x = lcPropGetFloat( hIns, LC_PROP_BLKREF_X );
  OdInsPos.y = lcPropGetFloat( hIns, LC_PROP_BLKREF_Y );
  OdInsPos.z = lcPropGetFloat( hIns, LC_PROP_ENT_Z );
  BlkScale.sx = lcPropGetFloat( hIns, LC_PROP_BLKREF_SCX );
  BlkScale.sy = lcPropGetFloat( hIns, LC_PROP_BLKREF_SCY );
  BlkScale.sz = 1.0; //lcPropGetFloat( hIns, LC_PROP_BLKREF_SCZ );
  if (BlkScale.sx == 0.0){
    BlkScale.sx = 1.0;
  }
  if (BlkScale.sy == 0.0){
    BlkScale.sy = 1.0;
  }
  if (BlkScale.sz == 0.0){
    BlkScale.sz = 1.0;
  }
  Ratio = BlkScale.sy / BlkScale.sx;
  if (NumRows > 1 || NumCols > 1){
    // array of blocks
    pOdMIns = OdDbMInsertBlock::createObject();
    pOdMIns->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdMIns );
    WriteEntityBase( hIns, pOdDb, pOdMIns );
    pOdMIns->setBlockTableRecord( pOdDb->getOdDbObjectId( OdHandle ) );
    pOdMIns->setScaleFactors( BlkScale );
    pOdMIns->setPosition( OdInsPos );
    pOdMIns->setRotation( BlkAngle );
    pOdMIns->setColumns( NumCols );
    pOdMIns->setColumnSpacing( ColDist );
    pOdMIns->setRows( NumRows );
    pOdMIns->setRowSpacing( RowDist );
    pOdInsert = pOdMIns;
  }else{
    // single block
    pOdInsert = OdDbBlockReference::createObject();
    pOdInsert->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdInsert );
    WriteEntityBase( hIns, pOdDb, pOdInsert );
    pOdInsert->setBlockTableRecord( pOdDb->getOdDbObjectId( OdHandle ) );
    pOdInsert->setScaleFactors( BlkScale );
    pOdInsert->setPosition( OdInsPos );
    pOdInsert->setRotation( BlkAngle );
  }
//  if (lcEntGetNormal( hIns, &xx, &yy, &zz )){
//    pOdInsert->setNormal( OdGeVector3d(xx,yy,zz) );
//  }

  // write attributes, if exist
  hAtt = lcBlkRefGetFirstAtt( hIns );
  while( hAtt ){
    pOdAttrib = OdDbAttribute::createObject();
    pOdInsert->appendAttribute( pOdAttrib );
    Write_Attrib( hAtt, pOdAttrib, OdInsPos, BlkScale, BlkAngle, pOdDb );
    hAtt = lcBlkRefGetNextAtt( hIns, hAtt );
  }
}

