/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
* Dimensions
*********************************************************************/
#include "pch.h"


//-----------------------------------------------
void Write_DimLin (HANDLE hDimRot, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  double Ang;
  HANDLE hStyle = lcPropGetHandle( hDimRot, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDbRotatedDimensionPtr pOdDimRot = OdDbRotatedDimension::createObject();

  pOdDimRot->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimRot );
  WriteEntityBase( hDimRot, pOdDb, pOdDimRot ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimRot->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // first ext. point
  odpt.x = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_DP1X );
  odpt.y = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_DP1Y );
  pOdDimRot->setXLine1Point( odpt );
  // second ext. point
  odpt.x = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_DP2X );
  odpt.y = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_DP2Y );
  pOdDimRot->setXLine2Point( odpt );
  // dim. line point
  odpt.x = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_TPX );
  odpt.y = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_TPY );
  pOdDimRot->setDimLinePoint( odpt );
  // text position
  pOdDimRot->setTextPosition( odpt );
  pOdDimRot->useSetTextPosition();
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimRot, LC_PROP_DIMLIN_TEXT ));
  pOdDimRot->setDimensionText( szText );
  // rotation
  Ang = lcPropGetFloat( hDimRot, LC_PROP_DIMLIN_ANGLE );
  pOdDimRot->setRotation( Ang );
//  pOdDimRot->createExtensionDictionary();
  pOdDimRot->recomputeDimBlock();
}

//-----------------------------------------------
void Write_DimAli (HANDLE hDimAli, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hDimAli, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDbAlignedDimensionPtr pOdDimAli = OdDbAlignedDimension::createObject();

  pOdDimAli->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimAli );
  WriteEntityBase( hDimAli, pOdDb, pOdDimAli ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimAli->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // first ext. point
  odpt.x = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_DP1X );
  odpt.y = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_DP1Y );
  pOdDimAli->setXLine1Point( odpt );
  // second ext. point
  odpt.x = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_DP2X );
  odpt.y = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_DP2Y );
  pOdDimAli->setXLine2Point( odpt );
  // dim. line point
  odpt.x = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_TPX );
  odpt.y = lcPropGetFloat( hDimAli, LC_PROP_DIMALI_TPY );
  pOdDimAli->setDimLinePoint( odpt );
  // text position
  pOdDimAli->setTextPosition( odpt );
  pOdDimAli->useSetTextPosition();
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimAli, LC_PROP_DIMALI_TEXT ));
  pOdDimAli->setDimensionText( szText );

//  pOdDimAli->createExtensionDictionary();
}

//-----------------------------------------------
void Write_DimOrd (HANDLE hDimOrd, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  BOOL   bX;
  HANDLE hStyle = lcPropGetHandle( hDimOrd, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDbOrdinateDimensionPtr pOdDimOrd = OdDbOrdinateDimension::createObject();

  pOdDimOrd->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimOrd );
  WriteEntityBase( hDimOrd, pOdDb, pOdDimOrd ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimOrd->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  bX = lcPropGetBool( hDimOrd, LC_PROP_DIMORD_XORD );
  if (bX){
    pOdDimOrd->useXAxis();
  }else{
    pOdDimOrd->useYAxis();
  }
  // ordinate origin
  odpt.x = 0.0;
  odpt.y = 0.0;
  odpt.z = 0.0;
  pOdDimOrd->setOrigin( odpt );
  // def. point
  odpt.x = lcPropGetFloat( hDimOrd, LC_PROP_DIMORD_DPX );
  odpt.y = lcPropGetFloat( hDimOrd, LC_PROP_DIMORD_DPY );
  pOdDimOrd->setDefiningPoint( odpt );
  // end point
  odpt.x = lcPropGetFloat( hDimOrd, LC_PROP_DIMORD_TPX );
  odpt.y = lcPropGetFloat( hDimOrd, LC_PROP_DIMORD_TPY );
  pOdDimOrd->setLeaderEndPoint( odpt );
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimOrd, LC_PROP_DIMORD_TEXT ));
  pOdDimOrd->setDimensionText( szText );

//  pOdDimOrd->createExtensionDictionary();
}

//-----------------------------------------------
void Write_DimRad (HANDLE hDimRad, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hDimRad, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDbRadialDimensionPtr pOdDimRad = OdDbRadialDimension::createObject();

  pOdDimRad->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimRad );
  WriteEntityBase( hDimRad, pOdDb, pOdDimRad ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimRad->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // center point
  odpt.x = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_CPX );
  odpt.y = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_CPY );
  pOdDimRad->setCenter( odpt );
  // radial point
  odpt.x = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_RPX );
  odpt.y = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_RPY );
  pOdDimRad->setChordPoint( odpt );
  // text point
  odpt.x = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_TPX );
  odpt.y = lcPropGetFloat( hDimRad, LC_PROP_DIMRAD_TPY );
  pOdDimRad->setTextPosition( odpt );
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimRad, LC_PROP_DIMRAD_TEXT ));
  pOdDimRad->setDimensionText( szText );

//  pOdDimRad->createExtensionDictionary();
}

//-----------------------------------------------
void Write_DimDia (HANDLE hDimDia, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hDimDia, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDbDiametricDimensionPtr pOdDimDia = OdDbDiametricDimension::createObject();

  pOdDimDia->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimDia );
  WriteEntityBase( hDimDia, pOdDb, pOdDimDia ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimDia->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // radial point
  odpt.x = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_RPX );
  odpt.y = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_RPY );
  pOdDimDia->setChordPoint( odpt );
  // far radial point
  odpt.x = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_FPX );
  odpt.y = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_FPY );
  pOdDimDia->setFarChordPoint( odpt );
  // text point
  odpt.x = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_TPX );
  odpt.y = lcPropGetFloat( hDimDia, LC_PROP_DIMDIA_TPY );
  pOdDimDia->setTextPosition( odpt );
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimDia, LC_PROP_DIMDIA_TEXT ));
  pOdDimDia->setDimensionText( szText );

//  pOdDimDia->createExtensionDictionary();
}

// 2 lines angular dimansion
//-----------------------------------------------
void Write_DimAng2Line (HANDLE hDimAng, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hDimAng, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDb2LineAngularDimensionPtr pOdDimAng = OdDb2LineAngularDimension::createObject();

  pOdDimAng->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimAng );
  WriteEntityBase( hDimAng, pOdDb, pOdDimAng );
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimAng->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }

  odpt.z = 0.0;
  // line 1 start
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L1P1X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L1P1Y );
  pOdDimAng->setXLine1Start( odpt );
  // line 1 end
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L1P2X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L1P2Y );
  pOdDimAng->setXLine1End( odpt );
  // line 2 start
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L2P1X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L2P1Y );
  pOdDimAng->setXLine2Start( odpt );
  // line 2 end
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L2P2X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_L2P2Y );
  pOdDimAng->setXLine2End( odpt );
  // arc point
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_APX );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_APY );
  pOdDimAng->setArcPoint( odpt );
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimAng, LC_PROP_DIMANG_TEXT ));
  pOdDimAng->setDimensionText( szText );

//  pOdDimAng->createExtensionDictionary();
}


// 3 points angular dimansion
//-----------------------------------------------
void Write_DimAng (HANDLE hDimAng, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  if (lcPropGetBool( hDimAng, LC_PROP_DIMANG_2LINE )){
    Write_DimAng2Line( hDimAng, pOdDb, pOdBlock );
    return;
  }

  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hDimAng, LC_PROP_DIM_STYLE );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt;
  OdDb3PointAngularDimensionPtr pOdDimAng = OdDb3PointAngularDimension::createObject();
   
  pOdDimAng->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdDimAng );
  WriteEntityBase( hDimAng, pOdDb, pOdDimAng ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdDimAng->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // line 1 start
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_CPX );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_CPY );
  pOdDimAng->setCenterPoint( odpt );
  // def point 1
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_DP1X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_DP1Y );
  pOdDimAng->setXLine1Point( odpt );
  // def point 2
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_DP2X );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_DP2Y );
  pOdDimAng->setXLine2Point( odpt );
  // arc point
  odpt.x = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_APX );
  odpt.y = lcPropGetFloat( hDimAng, LC_PROP_DIMANG_APY );
  pOdDimAng->setArcPoint( odpt );
  // dim. text
  wcscpy( szText, lcPropGetStr( hDimAng, LC_PROP_DIMANG_TEXT ));
  pOdDimAng->setDimensionText( szText );

//  pOdDimAng->createExtensionDictionary();
}


//-----------------------------------------------
void Write_Leader (HANDLE hLeader, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
#ifdef JUHHHGNBBVNNJKOOIU
  double Htxt, Gap;
  double Wbox = lcPropGetFloat( hLeader, LC_PROP_LEADER_TBW );
  double Hbox = lcPropGetFloat( hLeader, LC_PROP_LEADER_TBH );
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hLeader, LC_PROP_LEADER_STYLE );
  HANDLE hTextStyle = lcPropGetHandle( hStyle, LC_PROP_DIMST_TXSTY );
  OdDbHandle  OdHandle;
  OdGePoint3d odpt, odpt2;
  OdGeVector3d off;
  OdDbLeaderPtr pOdLeader = OdDbLeader::createObject();
   
  Htxt = lcPropGetFloat( hStyle, LC_PROP_DIMST_TXT );
  Gap = lcPropGetFloat( hStyle, LC_PROP_DIMST_GAP );
  pOdLeader->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdLeader );
  WriteEntityBase( hLeader, pOdDb, pOdLeader ); 
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdLeader->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt.z = 0.0;
  // arrow point 
  odpt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_APX );
  odpt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_APY );
  pOdLeader->appendVertex( odpt );
  // land point 
  odpt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1X );
  odpt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1Y );
  pOdLeader->appendVertex( odpt );
  // start point 
  odpt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0X );
  odpt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0Y );
  pOdLeader->appendVertex( odpt );
  // text position
  odpt2.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_TPX );
  odpt2.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_TPY );
  odpt2.z = 0.0;

  // create MTEXT object for the leader annotation
  OdDbMTextPtr pMText = OdDbMText::createObject();
  pMText->setDatabaseDefaults( pOdDb );
  OdDbObjectId IdMText = pOdBlock->appendOdDbEntity( pMText );
  pMText->setLocation( odpt2 );
  pMText->setRotation( 0.0 );
  pMText->setTextHeight( Htxt );
  pMText->setWidth( 0.0 ); //Wbox );
  pMText->setAttachment( OdDbMText::kMiddleCenter );
  // text contents
  wcscpy( szText, lcPropGetStr( hLeader, LC_PROP_LEADER_TEXT ));
  pMText->setContents( szText );
  // test style
  wcscpy( szOdDbHandle, lcPropGetStr( hTextStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pMText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  // attach text to leader
  pOdLeader->attachAnnotation( 0 ); //IdMText );
/*
  // annotation offset
  off.x = (Wbox/2.0); // + Gap); //odpt.x - odpt2.x;
  off.y = (Hbox/2.0); // + Gap); //odpt.y - odpt2.y;
  off.z = 0.0;
  pOdLeader->setAnnotationOffset( off );
*/
//  pOdLeader->createExtensionDictionary();
#endif 

//  int    i;
  double Htxt, Gap, dx, dy;
  double Wbox = lcPropGetFloat( hLeader, LC_PROP_LEADER_TBW );
  double Hbox = lcPropGetFloat( hLeader, LC_PROP_LEADER_TBH );
  WCHAR  szText[256], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hLeader, LC_PROP_LEADER_STYLE );
  HANDLE hTextStyle = lcPropGetHandle( hStyle, LC_PROP_DIMST_TXSTY );
  OdDbHandle  OdHandle;
  OdGePoint2d odpt, odpt2;
  OdGePoint3d OdTextPt, odpt3;
//  OdGeVector3d off;
//  OdDbLeaderPtr pOdLeader = OdDbLeader::createObject();
  OdDbPolylinePtr pOdPline;
  OdDbLeaderPtr   pOdLeader;
  OdDbMTextPtr    pMText;
  OdDbObjectId    IdMText;

  Htxt = lcPropGetFloat( hStyle, LC_PROP_DIMST_TXT );
  Gap = lcPropGetFloat( hStyle, LC_PROP_DIMST_GAP );
  // create MTEXT object for the leader annotation
  pMText = OdDbMText::createObject();
  pMText->setDatabaseDefaults( pOdDb );
  IdMText = pOdBlock->appendOdDbEntity( pMText );
  OdTextPt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_TPX );
  OdTextPt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_TPY );
  OdTextPt.z = 0.0;
  pMText->setLocation( OdTextPt );
  pMText->setRotation( 0.0 );
  pMText->setTextHeight( Htxt );
  pMText->setWidth( Wbox );
  pMText->setAttachment( OdDbMText::kMiddleCenter );
  // text contents
  wcscpy( szText, lcPropGetStr( hLeader, LC_PROP_LEADER_TEXT ));
  pMText->setContents( szText );
  // test style
  wcscpy( szOdDbHandle, lcPropGetStr( hTextStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pMText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }

  if (Gap < 0.0){
    pOdPline = OdDbPolyline::createObject();
    pOdPline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPline );
    WriteEntityBase( hLeader, pOdDb, pOdPline ); 
    Gap = -Gap;
    dx = Wbox / 2.0 + Gap;
    dy = Hbox / 2.0 + Gap;
    odpt.x = OdTextPt.x - dx;
    odpt.y = OdTextPt.y - dy;
    pOdPline->addVertexAt( 0, odpt );
    odpt.x = OdTextPt.x - dx;
    odpt.y = OdTextPt.y + dy;
    pOdPline->addVertexAt( 1, odpt );
    odpt.x = OdTextPt.x + dx;
    odpt.y = OdTextPt.y + dy;
    pOdPline->addVertexAt( 2, odpt );
    odpt.x = OdTextPt.x + dx;
    odpt.y = OdTextPt.y - dy;
    pOdPline->addVertexAt( 3, odpt );
    pOdPline->setClosed( true );
  }
/*
  pOdPline = OdDbPolyline::createObject();
  pOdBlock->appendOdDbEntity( pOdPline );
  WriteEntityBase( hLeader, pOdDb, pOdPline ); 
  i = 0;
  // start point 
  odpt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0X );
  odpt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0Y );
  pOdPline->addVertexAt( i++, odpt );
  // land point 
  odpt2.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1X );
  odpt2.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1Y );
  if (odpt.x != odpt2.x || odpt.y != odpt2.y){
    pOdPline->addVertexAt( i++, odpt2 );
  }
  // arrow point 
  odpt.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_APX );
  odpt.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_APY );
  pOdPline->addVertexAt( i++, odpt );
*/
  pOdLeader = OdDbLeader::createObject();
  pOdLeader->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdLeader );
  WriteEntityBase( hLeader, pOdDb, pOdLeader ); 
  pOdLeader->attachAnnotation( 0 );
  pOdLeader->setDimgap( 0.0 );
  // dimension style
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdLeader->setDimensionStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  odpt3.z = 0.0;
  // arrow point 
  odpt3.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_APX );
  odpt3.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_APY );
  pOdLeader->appendVertex( odpt3 );
  // land point 
  odpt3.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1X );
  odpt3.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P1Y );
  pOdLeader->appendVertex( odpt3 );
  // start point 
  odpt3.x = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0X );
  odpt3.y = lcPropGetFloat( hLeader, LC_PROP_LEADER_P0Y );
  pOdLeader->appendVertex( odpt3 );
}
