/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
* Lines and polylines
*********************************************************************/
#include "pch.h"

void Write_Exploded (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);
//void Write_3dPline (HANDLE hPline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);

//-----------------------------------------------
void Write_Line (HANDLE hLine, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdGePoint3d odpt;
  OdDbLinePtr pOdLine = OdDbLine::createObject();
  pOdLine->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdLine );
  WriteEntityBase( hLine, pOdDb, pOdLine ); 
  odpt.x = lcPropGetFloat( hLine, LC_PROP_LINE_X0 );
  odpt.y = lcPropGetFloat( hLine, LC_PROP_LINE_Y0 );
  odpt.z = lcPropGetFloat( hLine, LC_PROP_LINE_Z0 );
  pOdLine->setStartPoint( odpt );
  odpt.x = lcPropGetFloat( hLine, LC_PROP_LINE_X1 );
  odpt.y = lcPropGetFloat( hLine, LC_PROP_LINE_Y1 );
  odpt.z = lcPropGetFloat( hLine, LC_PROP_LINE_Z1 );
  pOdLine->setEndPoint( odpt );

/*
old
  OdGePoint3d point(0,0,0);
  OdGePoint3d line1Pt, line2Pt;
  double h  = 10.;
  double w  = 20.;

  line1Pt.x  = point.x + w * 1.0 / 8.0;
  line1Pt.y  = point.y + h * 2.0 / 8.0;
  line2Pt.x  = line1Pt.x + 3.75;
  line2Pt.y  = point.y + h * 7.0 / 8.0;

  OdDbRotatedDimensionPtr pDimension = OdDbRotatedDimension::createObject();
  pDimension->setDatabaseDefaults(pOdDb);
  OdDbObjectId dimensionId = pOdBlock->appendOdDbEntity(pDimension);

  OdGePoint3d dimLinePt;
  dimLinePt.x = point.x + w / 2.0;
  dimLinePt.y = point.y + h * 1.0 / 8.0;
  pDimension->setDatabaseDefaults(pOdDb);
  pDimension->setXLine1Point( line1Pt );
  pDimension->setXLine2Point( line2Pt );
  pDimension->setDimLinePoint( dimLinePt );
  pDimension->useDefaultTextPosition();
  pDimension->createExtensionDictionary();
*/
}


//-----------------------------------------------
void Write_Xline (HANDLE hXline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdGePoint3d  basePoint;
  OdGeVector3d unitDir;
  BOOL  bRay  = lcPropGetBool( hXline, LC_PROP_XLINE_RAY );
  basePoint.x = lcPropGetFloat( hXline, LC_PROP_XLINE_X0 );
  basePoint.y = lcPropGetFloat( hXline, LC_PROP_XLINE_Y0 );
  basePoint.z = lcPropGetFloat( hXline, LC_PROP_ENT_Z );
  unitDir.x   = lcPropGetFloat( hXline, LC_PROP_XLINE_DIRX );
  unitDir.y   = lcPropGetFloat( hXline, LC_PROP_XLINE_DIRY );
  unitDir.z   = 0.0; //lcPropGetFloat( hXline, LC_PROP_XLINE_DIRZ );
  if (bRay){
    OdDbRayPtr pOdRay = OdDbRay::createObject();
    pOdRay->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdRay );
    WriteEntityBase( hXline, pOdDb, pOdRay ); 
    pOdRay->setBasePoint( basePoint );
    pOdRay->setUnitDir( unitDir );  
  }else{
    OdDbXlinePtr pOdXline = OdDbXline::createObject();
    pOdXline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdXline );
    WriteEntityBase( hXline, pOdDb, pOdXline ); 
    pOdXline->setBasePoint( basePoint );
    pOdXline->setUnitDir( unitDir );  
  }
}

#define MAX_PLINES 100

//-----------------------------------------------
void Write_Pline (HANDLE hPline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  int nVers = lcPropGetInt( hPline, LC_PROP_PLINE_NVERS );
  if (nVers < 2){
    return;
  }
  int      CurveQuality = 5; //lcPropGetInt( hPline, LC_PROP_PLINE_RESOL );
  int      iVer, i;
  int      nPlines = 1;
  HANDLE   hVer;
  int      FitType = lcPropGetInt( hPline, LC_PROP_PLINE_FIT );
  BOOL     bClosed = lcPropGetBool( hPline, LC_PROP_PLINE_CLOSED );
  BOOL     bFilled = lcPropGetBool( hPline, LC_PROP_ENT_SOLIDFILL );
  BOOL     bBorder = true; //lcPropGetBool( hPline, LC_PROP_ENT_BORDER );
  BOOL     bConstZ = lcPropGetBool( hPline, LC_PROP_PLINE_CONSTZ );
  BOOL     bR0, bLastVer;  // bEndPath, 
  double   Zglob   = lcPropGetFloat( hPline, LC_PROP_ENT_Z );
  double   Bulge, Angle, R, x00, y00, x0, y0, x1, y1, X, Y;
  CGePoint Pnt;
  WCHAR    szColor[32];
  OdDbCurvePtr      pOdCurve[MAX_PLINES];
  OdDbHatchPtr      pOdHatch;
  OdDbObjectId      id1, id2;
  OdDbObjectIdArray loopIds;
  OdGePoint2d       OdPt2d; 
  OdGePoint3d       OdPt3d;
  OdCmColor         OdColor;
  OdDb3dPolylineVertexPtr pOdVer;

#ifdef HJJJHHBBVGF

/************************************************************************/
/* Add a Wipeout to to the specified BlockTableRecord                   */
/************************************************************************/
void DbFiller::addWipeout(const OdDbObjectId& btrId, 
                          const int boxRow,
                          const int boxCol,
                          const OdDbObjectId& layerId,
                          const OdDbObjectId& styleId)
{
  /**********************************************************************/
  /* Open the BlockTableRecord                                          */
  /**********************************************************************/
  OdDbBlockTableRecordPtr bBTR = btrId.safeOpenObject(OdDb::kForWrite);

  /**********************************************************************/
  /* Get the lower-left corner and center of the box                    */
  /**********************************************************************/
  OdGePoint3d point = m_EntityBoxes.getBox(boxRow, boxCol);
  OdGePoint3d center = m_EntityBoxes.getBoxCenter(boxRow, boxCol);
  double h    = m_EntityBoxes.getHeight();
  double w    = m_EntityBoxes.getWidth(boxRow, boxCol);
  
  /**********************************************************************/
  /* Add the label                                                     */
  /**********************************************************************/
  addTextEnt(bBTR,
    point + m_textOffset, point + m_textOffset, 
    OD_T("WIPEOUT"), m_textSize, OdDb::kTextLeft, OdDb::kTextTop, layerId, styleId );

  /**********************************************************************/
  /* Get the lower-left corner of the box                               */
  /**********************************************************************/
  point.y -= h;

  /**********************************************************************/
  /* Create a hatch object to be wiped out                              */
  /**********************************************************************/
  OdDbHatchPtr pHatch = OdDbHatch::createObject();
  pHatch->setDatabaseDefaults(bBTR->database());
  OdDbObjectId hatchId = bBTR->appendOdDbEntity(pHatch);

  /**********************************************************************/
  /* Create a hatch object to be wiped out                              */
  /**********************************************************************/
  pHatch->setAssociative(true);
  pHatch->setPattern(OdDbHatch::kUserDefined, OD_T("_USER"));
  pHatch->setPatternSpace(0.125);
  pHatch->setPatternAngle(0.5); // near 30 degrees
  pHatch->setPatternDouble(true); // Cross hatch
  pHatch->setHatchStyle(OdDbHatch::kNormal);

  /**********************************************************************/
  /* Create an outer loop for the hatch                                 */
  /**********************************************************************/
  OdDbCirclePtr pCircle = OdDbCircle::createObject();
  pCircle->setDatabaseDefaults(bBTR->database());
  OdDbObjectIdArray loopIds;
  loopIds.push_back(bBTR->appendOdDbEntity(pCircle));
  pCircle->setCenter(center);
  pCircle->setRadius( odmin(w, h) * 0.4);
  pHatch->appendLoop(OdDbHatch::kDefault, loopIds);

  /**********************************************************************/
  /* Create the wipeout                                                  */
  /**********************************************************************/
  OdDbWipeoutPtr pWipeout = OdDbWipeout::createObject();
  pWipeout->setDatabaseDefaults(bBTR->database());
  bBTR->appendOdDbEntity(pWipeout);

  OdGePoint3dArray boundary;
  boundary.append(center + OdGeVector3d(-w * 0.4, -h *0.4, 0.0));
  boundary.append(center + OdGeVector3d( w * 0.4, -h *0.4, 0.0));
  boundary.append(center + OdGeVector3d( 0.0,      h *0.4, 0.0));
  boundary.append(center + OdGeVector3d(-w * 0.4, -h *0.4, 0.0));

  pWipeout->setBoundary(boundary);

  pWipeout->setDisplayOpt(OdDbRasterImage::kShow,true);
  pWipeout->setDisplayOpt(OdDbRasterImage::kClip,true);
  pWipeout->setDisplayOpt(OdDbRasterImage::kShowUnAligned,true);
  pWipeout->setDisplayOpt(OdDbRasterImage::kTransparent,false);
}

#endif  // HJJJHHBBVGF



  if (bClosed && bFilled){
    if (lcPropGetBool( hPline, LC_PROP_PLINE_ODSOLID )){
      OdDbSolidPtr pOdSolid = OdDbSolid::createObject();
//      pOdSolid->setDatabaseDefaults(bBTR->database());
//      bBTR->appendOdDbEntity(pSolid);
      pOdBlock->appendOdDbEntity( pOdSolid );
      iVer = 0;
      hVer = lcPlineGetFirstVer( hPline );
      while (hVer){
        X = lcPropGetFloat( hVer, LC_PROP_VER_X );
        Y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
        pOdSolid->setPointAt( iVer, OdGePoint3d( X, Y, 0.0 ));
        ++iVer;
        hVer = lcPlineGetNextVer( hPline, hVer );
      }
      if (iVer == 3){
        pOdSolid->setPointAt( iVer, OdGePoint3d( X, Y, 0.0 ));
      }
      WriteEntityBase( hPline, pOdDb, pOdSolid ); 
      return;
    }

    if (lcPropGetBool( hPline, LC_PROP_PLINE_WIPEOUT )){
      double Xmin, Ymin, Xmax, Ymax, Xc, Yc, x, y, dx, dy, W, H, Size;
      OdGePoint2dArray Bndr; //ptArray;
//      OdGePoint3dArray Bndr;
      OdGePoint3dArray boundary;
      OdDbWipeoutPtr pOdWipe = OdDbWipeout::createObject();
      pOdWipe->setDatabaseDefaults( pOdDb );
//      OdPt3d.z = 0.0;
//      iVer = 0;

      Xmin = lcPropGetFloat( hPline, LC_PROP_ENT_XMIN );
      Ymin = lcPropGetFloat( hPline, LC_PROP_ENT_YMIN );
      Xmax = lcPropGetFloat( hPline, LC_PROP_ENT_XMAX );
      Ymax = lcPropGetFloat( hPline, LC_PROP_ENT_YMAX );
      W = Xmax - Xmin;
      H = Ymax - Ymin;
      boundary.append( OdGePoint3d( Xmin, Ymin, 0.0 ));
      boundary.append( OdGePoint3d( Xmax, Ymin, 0.0 ));
      boundary.append( OdGePoint3d( Xmin, Ymax, 0.0 )); 
      boundary.append( OdGePoint3d( Xmin, Ymin, 0.0 ));
      pOdWipe->setBoundary( boundary );
      pOdWipe->setDisplayOpt(OdDbRasterImage::kShow,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kClip,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kShowUnAligned,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kTransparent,false);

      Size = max( W, H );
      Xc = Xmin + (Size / 2.0);
      Yc = Ymin + (Size / 2.0);
      hVer = lcPlineGetFirstVer( hPline );
      while( hVer ){
        x = lcPropGetFloat( hVer, LC_PROP_VER_X );
        y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
        dx = x - Xc;
        dy = Yc - y;
        OdPt2d.x = dx / Size;
        OdPt2d.y = dy / Size;
        Bndr.append( OdPt2d );
        hVer = lcPlineGetNextVer( hPline, hVer );
      }
      pOdWipe->setClipBoundary( Bndr );


      pOdBlock->appendOdDbEntity( pOdWipe );
      WriteEntityBase( hPline, pOdDb, pOdWipe ); 

/*
      pOdWipe->setBoundary( Bndr );
      pOdWipe->setDisplayOpt(OdDbRasterImage::kShow,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kClip,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kShowUnAligned,true);
      pOdWipe->setDisplayOpt(OdDbRasterImage::kTransparent,false);
*/

//      pOdWipe->setElevation( Zglob );
      return;
    }
//    i = lcPropGetInt( hPline, LC_PROP_ENT_ID );
//    wcscpy( szColor, lcPropGetStr( hPline, LC_PROP_ENT_FCOLOR ) );
  }

  //--------------------------------------------- LC_PLFIT_BULGE
  if (FitType == LC_PLFIT_BULGE){
    if (bConstZ){
      OdDbPolylinePtr pOdPolyline = OdDbPolyline::createObject();
      pOdPolyline->setDatabaseDefaults( pOdDb );
      pOdBlock->appendOdDbEntity( pOdPolyline );
      WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
      iVer = 0;
      hVer = lcPlineGetFirstVer( hPline );
      while( hVer ){
        OdPt2d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
        OdPt2d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
        Bulge  = lcPropGetFloat( hVer, LC_PROP_VER_BULGE );
        pOdPolyline->addVertexAt( iVer, OdPt2d, Bulge );
        ++iVer;
/*
        bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
        if (bEndPath){
          pOdPolyline->setElevation( Zglob );
          pOdPolyline->setClosed( bClosed?true:false );
          pOdPolyline->setPlinegen( true );
          pOdCurve[nPlines-1] = pOdPolyline;
          // next polyline
          pOdPolyline = OdDbPolyline::createObject();
          pOdPolyline->setDatabaseDefaults( pOdDb );
          pOdBlock->appendOdDbEntity( pOdPolyline );
          WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
          iVer = 0;
          nPlines++;
        }
*/
        hVer = lcPlineGetNextVer( hPline, hVer );
      }
      pOdPolyline->setElevation( Zglob );
      pOdPolyline->setClosed( bClosed?true:false );
/*
      if (pPline->HasConstWidth()){
        Width = lcPropGetFloat( hPline, LC_PROP_PLINE_WIDTH );
        pOdPolyline->setConstantWidth( Width );
      }
*/
      pOdPolyline->setPlinegen( true );
      pOdCurve[nPlines-1] = pOdPolyline;
    }else{
//      Write_3dPline( hPline, pOdDb, pOdBlock );

      OdDb3dPolylinePtr pOd3dPline = OdDb3dPolyline::createObject();
      pOd3dPline->setDatabaseDefaults( pOdDb );
      iVer = 0;
      hVer = lcPlineGetFirstVer( hPline );
      while( hVer ){
        pOdVer = OdDb3dPolylineVertex::createObject();
        pOd3dPline->appendVertex( pOdVer );
        OdPt3d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
        OdPt3d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
        OdPt3d.z = lcPropGetFloat( hVer, LC_PROP_VER_Z );
        pOdVer->setPosition( OdPt3d );
        pOdVer->setVertexType( OdDb::k3dSimpleVertex );
/*
        bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
        if (bEndPath){
          if (bClosed){
            pOd3dPline->makeClosed();
          }else{
            pOd3dPline->makeOpen();
          }
          pOdCurve[nPlines-1] = pOd3dPline;
          // next polyline
          pOd3dPline = OdDb3dPolyline::createObject();
          pOd3dPline->setDatabaseDefaults( pOdDb );
          pOdBlock->appendOdDbEntity( pOd3dPline );
          WriteEntityBase( hPline, pOdDb, pOd3dPline ); 
          iVer = 0;
          nPlines++;
        }
*/
        hVer = lcPlineGetNextVer( hPline, hVer );
      }
      pOdBlock->appendOdDbEntity( pOd3dPline );
      WriteEntityBase( hPline, pOdDb, pOd3dPline ); 
      if (bClosed){
        pOd3dPline->makeClosed();
      }else{
        pOd3dPline->makeOpen();
      }
      pOdCurve[nPlines-1] = pOd3dPline;
    }
    goto mHatch;
  }else

  //--------------------------------------------- LC_PLFIT_QUAD & LC_PLFIT_CUBIC
  if (FitType == LC_PLFIT_QUAD ||
      FitType == LC_PLFIT_CUBIC)
  {
    OdDb2dVertexPtr   pOdVertex;
    OdDb2dPolylinePtr pOdPolyline = OdDb2dPolyline::createObject();
    pOdPolyline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPolyline );
    WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
    OdPt3d.z = 0.;
    hVer = lcPlineGetFirstVer( hPline );
    while( hVer ){
      pOdVertex = OdDb2dVertex::createObject();
      pOdVertex->setDatabaseDefaults( pOdDb );
      pOdPolyline->appendVertex( pOdVertex );
      OdPt3d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
      OdPt3d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
      pOdVertex->setPosition( OdPt3d );
      pOdVertex->setVertexType( OdDb::k2dVertex ); //OdDb::k2dSplineCtlVertex );
/*
      bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
      if (bEndPath){
        pOdPolyline->setElevation( Zglob );
        if (bClosed){
          pOdPolyline->makeClosed();
        }else{
          pOdPolyline->makeOpen();
        }
        pOdPolyline->setLinetypeGenerationOn();
        switch( FitType ){
          case LC_PLFIT_QUAD:   pOdPolyline->splineFit( OdDb::k2dQuadSplinePoly, CurveQuality ); break;
          case LC_PLFIT_CUBIC:  pOdPolyline->splineFit( OdDb::k2dCubicSplinePoly, CurveQuality ); break;
        }
        pOdCurve[nPlines-1] = pOdPolyline;
        // next polyline
        pOdPolyline = OdDb2dPolyline::createObject();
        pOdPolyline->setDatabaseDefaults( pOdDb );
        pOdBlock->appendOdDbEntity( pOdPolyline );
        WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
        nPlines++;
      }
*/
      hVer = lcPlineGetNextVer( hPline, hVer );
    }
/*
    pVer = pPline->GetFirstVertex();
    while( pVer ){
      pOdVertex = OdDb2dVertex::createObject();
      pOdVertex->setDatabaseDefaults( pOdDb );
      pOdPolyline->appendVertex( pOdVertex );
      pos.x = pVer->m_Pnt.x;
      pos.y = pVer->m_Pnt.y;
      pOdVertex->setPosition( pos );
      pOdVertex->setVertexType( OdDb::k2dVertex ); //OdDb::k2dSplineCtlVertex );
      pVer = pPline->GetNextVertex( pVer );
    }
*/
    pOdPolyline->setElevation( Zglob );
/*

    switch( pPline->FitType() ){
      case LC_PLFIT_QUAD:   pOdPolyline->setPolyType( OdDb::k2dQuadSplinePoly );
      case LC_PLFIT_CUBIC:  pOdPolyline->setPolyType( OdDb::k2dCubicSplinePoly );
    }
*/
    if (bClosed){
      pOdPolyline->makeClosed();
    }else{
      pOdPolyline->makeOpen();
    }
    pOdPolyline->setLinetypeGenerationOn();
    switch( FitType ){
      case LC_PLFIT_QUAD:   pOdPolyline->splineFit( OdDb::k2dQuadSplinePoly, CurveQuality ); break;
      case LC_PLFIT_CUBIC:  pOdPolyline->splineFit( OdDb::k2dCubicSplinePoly, CurveQuality ); break;
    }
    pOdCurve[nPlines-1] = pOdPolyline;
    goto mHatch;
  }else

  //--------------------------------------------- LC_PLFIT_SPLINE
  if (FitType == LC_PLFIT_SPLINE){
    OdGePoint3dArray fitPoints;
    OdGeVector3d  TangStart, TangEnd;
    OdDbSplinePtr pOdSpline = OdDbSpline::createObject();
    pOdSpline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdSpline );
    WriteEntityBase( hPline, pOdDb, pOdSpline ); 
    OdPt3d.z = 0.;
    iVer = 0;
    hVer = lcPlineGetFirstVer( hPline );
    while( hVer ){
      OdPt3d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
      OdPt3d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
      fitPoints.push_back( OdPt3d );
      hVer = lcPlineGetNextVer( hPline, hVer );
    }
    BOOL bHasStartVector = lcPropGetBool( hPline, LC_PROP_PLINE_HASANG0 );
    BOOL bHasEndVector   = lcPropGetBool( hPline, LC_PROP_PLINE_HASANG2 );
    if (bHasStartVector){
      Angle = lcPropGetFloat( hPline, LC_PROP_PLINE_ANG0 );
      Pnt.Set( 0., 0. );
      Pnt.Polar( Angle, 1. );
      TangStart.x = Pnt.x;
      TangStart.y = Pnt.y;
    }
    if (bHasEndVector){
      Angle = lcPropGetFloat( hPline, LC_PROP_PLINE_ANG2 );
      Pnt.Set( 0., 0. );
      Pnt.Polar( Angle, 1. );
      TangEnd.x = Pnt.x;
      TangEnd.y = Pnt.y;
    }
    pOdSpline->setFitData(
      fitPoints,     // Fit Points
      3,             // Degree
      0.0,           // Fit tolerance
      TangStart,     // startTangent
      TangEnd        // endTangent
    );
    pOdCurve[nPlines-1] = pOdSpline;
    goto mHatch;
  }else

  //--------------------------------------------- LC_PLFIT_ROUND
  if (FitType == LC_PLFIT_ROUND){
    OdDbPolylinePtr pOdPolyline = OdDbPolyline::createObject();
    pOdPolyline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPolyline );
    WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
    bR0 = false;
    nPlines = 1;
    iVer = 0;
    hVer = lcPlineGetFirstVer( hPline );
    while( hVer ){
      OdPt2d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
      OdPt2d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
//      bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
      bLastVer = lcPropGetBool( hVer, LC_PROP_VER_LAST );
      R = lcPropGetFloat( hVer, LC_PROP_VER_RADIUS );
      if (R > 0.0){
        lcPlineGetRoundPrm( hPline, hVer, &x0, &y0, &Bulge, &x1, &y1 );
        OdPt2d.x = x0;
        OdPt2d.y = y0;
        if (iVer == 0){
          x00 = x0;
          y00 = y0;
          bR0 = true;  // vertex 0 is rounded
        }
        pOdPolyline->addVertexAt( iVer, OdPt2d, Bulge );
        ++iVer;
        OdPt2d.x = x1;
        OdPt2d.y = y1;
        pOdPolyline->addVertexAt( iVer, OdPt2d );
        ++iVer;
      }else{
        if (bLastVer && bR0){   // ((bEndPath || bLastVer) && bR0)
          OdPt2d.x = x00;
          OdPt2d.y = y00;
        }
        pOdPolyline->addVertexAt( iVer, OdPt2d );
        ++iVer;
      }
/*
      if (bEndPath){
        pOdPolyline->setElevation( Zglob );
        pOdPolyline->setClosed( bClosed?true:false );
        pOdPolyline->setPlinegen( true );
        pOdCurve[nPlines-1] = pOdPolyline;
        // next polyline
        pOdPolyline = OdDbPolyline::createObject();
        pOdPolyline->setDatabaseDefaults( pOdDb );
        pOdBlock->appendOdDbEntity( pOdPolyline );
        WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
        bR0 = false;
        iVer = 0;
        nPlines++;
      }
*/
      hVer = lcPlineGetNextVer( hPline, hVer );
    }
    pOdPolyline->setElevation( Zglob );
    pOdPolyline->setClosed( bClosed?true:false );
    pOdPolyline->setPlinegen( true );
    pOdCurve[nPlines-1] = pOdPolyline;
    goto mHatch;

/*
    CGePoint pt1, pt2;
//    LcDouble  bulge;
    pOdPolyline = OdDbPolyline::createObject();
    pOdPolyline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPolyline );
    WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
    iVer = 0;
    pVerPrev = NULL;
    pVer = pPline->GetFirstVertex();
    while( pVer ){
      pVerNext = pPline->GetNextVertex( pVer );
      if (pPline->IsClosed()){
        if (pVerPrev == NULL){
          pVerPrev = pPline->GetLastVertex();
        }
        if (pVerNext == NULL){
          pVerNext = pPline->GetFirstVertex();
        }
      }
      if (pVerPrev && pVerNext){
        if (lcuGetFilletBulge( pVerPrev->m_Pnt, pVer->m_Pnt, pVerNext->m_Pnt, pVer->Radius(), &pt1, &pt2, &bulge )){
          if (pPline->IsChamfer()){
            bulge = 0.;
          }
          odpt.x = pt1.x;
          odpt.y = pt1.y;
          pOdPolyline->addVertexAt( iVer++, odpt, bulge );
          odpt.x = pt2.x;
          odpt.y = pt2.y;
          pOdPolyline->addVertexAt( iVer++, odpt, 0. );
        }else{
          goto m1;
        }
      }else{
m1:
        odpt.x = pVer->m_Pnt.x;
        odpt.y = pVer->m_Pnt.y;
        pOdPolyline->addVertexAt( iVer++, odpt, 0. );
      }
      pVerPrev = pVer;
      pVer = pPline->GetNextVertex( pVer );
    }
    pOdPolyline->setElevation( pPline->Z() );
    pOdPolyline->setClosed( bClosed );
    if (pPline->HasConstWidth()){
      pOdPolyline->setConstantWidth( pPline->ConstWidth() );
    }
    pOdPolyline->setPlinegen( true );
    pOdCurve = pOdPolyline;
    goto mHatch;
*/
  }else{
/*
  //--------------------------------------------- LC_PLFIT_LINQUAD
  if (FitType == LC_PLFIT_LINQUAD){
    OdDbPolylinePtr pOdPolyline = OdDbPolyline::createObject();
    pOdPolyline->setDatabaseDefaults( pOdDb );
    pOdBlock->appendOdDbEntity( pOdPolyline );
    WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
    nPlines = 1;
    iVer = 0;
//    lcPlineSimple( hPline ); //, true, 0 );
    hVer = lcPlineGetFirstVer( hPline );
    while( hVer ){
      OdPt2d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
      OdPt2d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
      pOdPolyline->addVertexAt( iVer, OdPt2d );
      ++iVer;
      bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
      if (bEndPath){
        pOdPolyline->setElevation( Zglob );
        pOdPolyline->setClosed( bClosed?true:false );
        pOdPolyline->setPlinegen( true );
        pOdCurve[nPlines-1] = pOdPolyline;
        // next polyline
        pOdPolyline = OdDbPolyline::createObject();
        pOdPolyline->setDatabaseDefaults( pOdDb );
        pOdBlock->appendOdDbEntity( pOdPolyline );
        WriteEntityBase( hPline, pOdDb, pOdPolyline ); 
        iVer = 0;
        nPlines++;
      }
      hVer = lcPlineGetNextVer( hPline, hVer );
    }
//    lcPlineSimple( hPline ); //, false, 0 );
    pOdPolyline->setElevation( Zglob );
    pOdPolyline->setClosed( bClosed?true:false );
    pOdPolyline->setPlinegen( true );
    pOdCurve[nPlines-1] = pOdPolyline;
    goto mHatch;
  }else{
*/
  //--------------------------------------------- LC_PLFIT_others
    Write_Exploded( hPline, pOdDb, pOdBlock );
    return;
  }
  return;

mHatch:
  if (bClosed && bFilled){
    // apply inner filling (as solid hatch)
    pOdHatch = OdDbHatch::createObject();
    pOdBlock->appendOdDbEntity( pOdHatch );
    WriteEntityBase( hPline, pOdDb, pOdHatch, true ); 
    pOdHatch->setAssociative( true );
    pOdHatch->setElevation( Zglob );
    pOdHatch->setHatchStyle( OdDbHatch::kNormal );
    pOdHatch->setPattern( OdDbHatch::kPreDefined, L"SOLID" );
//    wcscpy( szColor, lcPropGetStr( hPline, LC_PROP_ENT_BCOLOR ) );
    wcscpy( szColor, lcPropGetStr( hPline, LC_PROP_ENT_FCOLOR ) );
    for (i=0; i<nPlines; i++){
      if (bBorder){
        MakeOdColor( szColor, &OdColor );
        pOdCurve[i]->setColor( OdColor, false );
      }
      loopIds.push_back( pOdCurve[i]->objectId() );
      pOdHatch->appendLoop( OdDbHatch::kDefault, loopIds );
      loopIds.clear();
    }
    // swap objects
    OdDbSortentsTablePtr pOdSort = pOdBlock->getSortentsTable();
    id1 = pOdCurve[0]->objectId();
    id2 = pOdHatch->objectId();
    pOdSort->swapOrder( id1, id2 );
  }
}

/*
//-----------------------------------------------
void Write_3dPline (HANDLE hPline, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock, int& nPlines, pOdCurve)
{
  int      iVer;
//  int      nPlines = 1;
  HANDLE   hVer;
  BOOL     bClosed = lcPropGetBool( hPline, LC_PROP_PLINE_CLOSED );
  BOOL     bEndPath;
  OdGePoint3d       OdPt3d;
  OdDb3dPolylineVertexPtr pOdVer;

  OdDb3dPolylinePtr pOd3dPline = OdDb3dPolyline::createObject();
  pOd3dPline->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOd3dPline );
  WriteEntityBase( hPline, pOdDb, pOd3dPline ); 
  iVer = 0;
  hVer = lcPlineGetFirstVer( hPline );
  while( hVer ){
    pOdVer = OdDb3dPolylineVertex::createObject();
    pOd3dPline->appendVertex( pOdVer );
    OdPt3d.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
    OdPt3d.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
    OdPt3d.y = lcPropGetFloat( hVer, LC_PROP_VER_Z );
    pOdVer->setPosition( OdPt3d );
    pOdVer->setVertexType( OdDb::k3dControlVertex );
    bEndPath = lcPropGetBool( hVer, LC_PROP_VER_ENDPATH );
    if (bEndPath){
      if (bClosed){
        pOd3dPline->makeClosed();
      }else{
        pOd3dPline->makeOpen();
      }
      pOdCurve[nPlines-1] = pOd3dPline;
      // next polyline
      pOd3dPline = OdDb3dPolyline::createObject();
      pOd3dPline->setDatabaseDefaults( pOdDb );
      pOdBlock->appendOdDbEntity( pOd3dPline );
      WriteEntityBase( hPline, pOdDb, pOd3dPline ); 
      iVer = 0;
      nPlines++;
    }
    hVer = lcPlineGetNextVer( hPline, hVer );
  }
  if (bClosed){
    pOd3dPline->makeClosed();
  }else{
    pOd3dPline->makeOpen();
  }
  pOdCurve[nPlines-1] = pOd3dPline;
}
*/

//-----------------------------------------------
void Write_Rect (HANDLE hRect, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szColor[32];
  double X[8], Y[8], Bulge[8], Z;
  BOOL   bFilled = false; //lcPropGetBool( hRect, LC_PROP_RECT_FILLED );
  BOOL   bBorder = false; //lcPropGetBool( hRect, LC_PROP_RECT_BORDER );
  int    nVers, i;
  OdCmColor       OdColor;
  OdGePoint2d     OdPnt; 
  OdDbPolylinePtr pOdPline = OdDbPolyline::createObject();

  pOdPline->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdPline );
  WriteEntityBase( hRect, pOdDb, pOdPline ); 
  nVers = lcRectGetPolyline( hRect, X, Y, Bulge );
  for (i=0; i<nVers; i++){
    OdPnt.set( X[i], Y[i] );
    pOdPline->addVertexAt( i, OdPnt, Bulge[i] );
  }
  pOdPline->setClosed( true );

  if (bFilled){
    if (bBorder){
//      wcscpy( szColor, lcPropGetStr( hRect, LC_PROP_ENT_BCOLOR ) );
      wcscpy( szColor, lcPropGetStr( hRect, LC_PROP_ENT_FCOLOR ) );
      MakeOdColor( szColor, &OdColor );
      pOdPline->setColor( OdColor, false );
    }
    // inside filling (as solid hatch)
    OdDbObjectIdArray loopIds;
    OdDbHatchPtr pOdHatch = OdDbHatch::createObject();
    pOdBlock->appendOdDbEntity( pOdHatch );
    WriteEntityBase( hRect, pOdDb, pOdHatch ); 
    loopIds.push_back( pOdPline->objectId() );
    pOdHatch->setAssociative( true );
    Z = lcPropGetFloat( hRect, LC_PROP_ENT_Z );
    pOdHatch->setElevation( Z );
    pOdHatch->setHatchStyle( OdDbHatch::kNormal );
    pOdHatch->setPattern( OdDbHatch::kPreDefined, "SOLID" );
    pOdHatch->appendLoop( OdDbHatch::kDefault, loopIds );
    // swap objects
    OdDbSortentsTablePtr pOdSort = pOdBlock->getSortentsTable();
    pOdSort->swapOrder( pOdPline->objectId(), pOdHatch->objectId() );
  }
}

//-----------------------------------------------
void Write_MLine (HANDLE hMLine, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
/*
  OdDbHandle   OdHandle;
  OdDbMlinePtr pOdMline;
  OdGePoint3d  odpt;
  OdDbDictionaryPtr pMLDic; // = pOdDb->getMLStyleDictionaryId().safeOpenObject( OdDb::kForWrite );
  OdDbObjectId      idMls;
  WCHAR  szOdDbHandle[32];
  BOOL   bClosed = lcPropGetBool( hMLine, LC_PROP_MLINE_CLOSED );
  int    Just = lcPropGetInt( hMLine, LC_PROP_MLINE_JUST );
  double Scale = lcPropGetFloat( hMLine, LC_PROP_MLINE_SCALE );
  HANDLE hStyle = lcPropGetHandle( hMLine, LC_PROP_MLINE_STYLE );
  HANDLE hVer;

  pOdMline = OdDbMline::createObject();
  pOdMline->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdMline );
  WriteEntityBase( hMLine, pOdDb, pOdMline ); 

  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdMline->setStyle( pOdDb->getOdDbObjectId( OdHandle ));
  }else{
    pMLDic = pOdDb->getMLStyleDictionaryId().safeOpenObject( OdDb::kForRead );
    idMls = pMLDic->getAt( L"Standard" );
    pOdMline->setStyle( idMls ); //idMls.safeOpenObject( OdDb::kForRead ));
  }
  pOdMline->setScale( Scale );
  pOdMline->setJustification( Just );
  pOdMline->setClosedMline( bClosed?true:false );
  hVer = lcMlineGetFirstVer( hMLine );
  while( hVer ){
    odpt.x = lcPropGetFloat( hVer, LC_PROP_VER_X );
    odpt.y = lcPropGetFloat( hVer, LC_PROP_VER_Y );
    odpt.z = lcPropGetFloat( hVer, LC_PROP_VER_Z );
    pOdMline->appendSeg( odpt );
    hVer = lcMlineGetNextVer( hMLine, hVer );
  }
*/
}


/*
//-----------------------------------------------
void Write_Bezier (const LcDbBezier* pBezier, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  OdGePoint3d      odpt;
  OdGePoint3dArray controlPoints;
  OdGeDoubleArray  knots;
  OdGeDoubleArray  weights;
  double           controlPtTol = 1.0e-10;
  double           knotTol = 1.0e-10;
  OdDbSplinePtr    pOdSpline = OdDbSpline::createObject();
  double           knotVal = 0.;
  LcBezVer*        pVer;
  LcBezVer*        pVerNext;

  pOdSpline->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdSpline );
  WriteEntityBase( pBezier, pOdDb, pOdSpline ); 
  weights.clear();
  knots.clear();
  controlPoints.clear();

  pVer = pBezier->GetFirstVertex();
  while( pVer ){
    pVerNext = pBezier->GetNextVertex( pVer );
    if (pVerNext){
      if (pVer->PrevVer() == NULL){
        odpt.x = pVer->Position().x;
        odpt.y = pVer->Position().y;
        odpt.z = 0.;
        controlPoints.append( odpt );
        knots.append( knotVal );
      }

      odpt.x = pVer->AbsPntB().x;
      odpt.y = pVer->AbsPntB().y;
      odpt.z = 0.;
      controlPoints.append( odpt );
      knots.append( knotVal );

      odpt.x = pVerNext->AbsPntA().x;
      odpt.y = pVerNext->AbsPntA().y;
      odpt.z = 0.;
      controlPoints.append( odpt );
      knots.append( knotVal );

      odpt.x = pVerNext->Position().x;
      odpt.y = pVerNext->Position().y;
      odpt.z = 0.;
      controlPoints.append( odpt );
      knots.append( knotVal );

      knotVal += 1.;
    }
    pVer = pVerNext;
  }
  knots.append( knotVal );
  knots.append( knotVal );
  knots.append( knotVal );
  knots.append( knotVal );

  pOdSpline->setNurbsData( 3, false, pBezier->IsClosed(), true, controlPoints, knots, weights, controlPtTol, knotTol ); 
}
*/

