/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Write AutoCAD drawing
* Text objects
*********************************************************************/
#include "pch.h"

//extern bool  _bDebugEnt;

extern COptions g_Options;

void Write_Exploded (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock);

//-----------------------------------------------
void Write_Text (HANDLE hText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[1024], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hText, LC_PROP_TEXT_STYLE );
  double Height  = lcPropGetFloat( hText, LC_PROP_TEXT_H ); 
  double WScale  = lcPropGetFloat( hText, LC_PROP_TEXT_WSCALE ); 
  double Angle   = lcPropGetFloat( hText, LC_PROP_TEXT_ANGLE ); 
  double Oblique = lcPropGetFloat( hText, LC_PROP_TEXT_OBLIQUE ); 
  BOOL   bBack   = lcPropGetBool( hText, LC_PROP_TEXT_BACKWARD ); 
  BOOL   bUpdown = lcPropGetBool( hText, LC_PROP_TEXT_UPDOWN ); 
  int    Align   = lcPropGetInt( hText, LC_PROP_TEXT_ALIGN ); 
  int    nChars  = lcPropGetInt( hText, LC_PROP_TEXT_LEN );
  BOOL   bHollow, bClosed;
  OdGePoint3d  pt0, pta;
  OdDbHandle   OdHandle;
  OdDbTextPtr  pOdText;

//if (_bDebugEnt){
//  _itow( nChars, szOdDbHandle, 10 );
//  ::MessageBox( ::GetActiveWindow(), szOdDbHandle, L"nChars of Export Text", MB_OK );
//}
  if (nChars > 1020){
    return;
  }
  if (nChars < 1){
    return;
  }
  bHollow = lcPropGetBool( hStyle, LC_PROP_TSTYLE_HOLLOW );
  bClosed = lcPropGetBool( hStyle, LC_PROP_TSTYLE_CLOSED );
  if (bHollow){
    Write_Exploded( hText, pOdDb, pOdBlock );
    return;
  }
  if (!bClosed && g_Options.m_Exp_bTextAsPlines){
    Write_Exploded( hText, pOdDb, pOdBlock );
    return;
  }

  if (Height == 0.0){
    Height = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_HEIGHT );
  }
  if (WScale == 0.0){
    WScale = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_WSCALE );
  }
  if (Oblique == 0.0){
    Oblique = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_OBLIQUE );
  }
  pOdText = OdDbText::createObject();
  pOdText->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdText );
  WriteEntityBase( hText, pOdDb, pOdText );
  pt0.x   = lcPropGetFloat( hText, LC_PROP_TEXT_X0 );  // X coordinate of left-bottom corner
  pt0.y   = lcPropGetFloat( hText, LC_PROP_TEXT_Y0 );  // Y coordinate of left-bottom corner
  pt0.z   = 0.0;  // ???
  pta.x   = lcPropGetFloat( hText, LC_PROP_TEXT_X );  // Insertion point X
  pta.y   = lcPropGetFloat( hText, LC_PROP_TEXT_Y );  // Insertion point Y
  pta.z   = lcPropGetFloat( hText, LC_PROP_ENT_Z );  // Insertion point Z
  szText[0] = 0;
//if (_bDebugEnt){
//  ::MessageBox( ::GetActiveWindow(), L"t4", L"Export Text", MB_OK );
//}
  wcscpy( szText, lcPropGetStr( hText, LC_PROP_TEXT_STR ));
//if (_bDebugEnt){
//  ::MessageBox( ::GetActiveWindow(), szText, L"Export Text", MB_OK );
//}
  pOdText->setTextString( szText );
  pOdText->setPosition( pt0 );
  pOdText->setAlignmentPoint( pta );
  pOdText->setHeight( Height );
  pOdText->setWidthFactor( WScale );
  pOdText->setOblique( Oblique );
  pOdText->setRotation( Angle );
  pOdText->mirrorInX( bBack? true : false );
  pOdText->mirrorInY( bUpdown? true : false );
  switch( Align ){
    case  LC_TA_LEFBOT:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_LEFCEN:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_LEFTOP:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_CENBOT:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_CENTER:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_CENTOP:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_RIGBOT:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_RIGCEN:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_RIGTOP:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
  }
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
//  pOdText->convertFieldToText();
//  pOdText->recordGraphicsModified();

//  lcEntGetNormal( hText, &x, &y, &z );
//  pOdText->setNormal( OdGeVector3d(x,y,z) );
//  pOdText->recordGraphicsModified( false );
}


//-----------------------------------------------
void Write_MText (HANDLE hMText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szOdDbHandle[32];
  WCHAR* szText = NULL;
  double Height    = lcPropGetFloat( hMText, LC_PROP_MTEXT_H ); 
  double WScale    = lcPropGetFloat( hMText, LC_PROP_MTEXT_WSCALE ); 
  double WrapWidth = lcPropGetFloat( hMText, LC_PROP_MTEXT_WRAPWIDTH ); 
  double Angle     = lcPropGetFloat( hMText, LC_PROP_MTEXT_ANGLE ); 
  double LineSpace = lcPropGetFloat( hMText, LC_PROP_MTEXT_LINESPACE );
  int    Align     = lcPropGetInt( hMText, LC_PROP_MTEXT_ALIGN ); 
  int    TextLen   = lcPropGetInt( hMText, LC_PROP_MTEXT_LEN ); 
  BOOL   bHollow, bClosed;
  OdGePoint3d    OdPnt;
  HANDLE         hStyle = lcPropGetHandle( hMText, LC_PROP_MTEXT_STYLE );
  OdDbHandle     OdHandle;
  OdDbMTextPtr   pOdMText = OdDbMText::createObject();

  bHollow = lcPropGetBool( hStyle, LC_PROP_TSTYLE_HOLLOW );
  bClosed = lcPropGetBool( hStyle, LC_PROP_TSTYLE_CLOSED );
  if (bHollow){
    Write_Exploded( hMText, pOdDb, pOdBlock );
    return;
  }
  if (!bClosed && g_Options.m_Exp_bTextAsPlines){
    Write_Exploded( hMText, pOdDb, pOdBlock );
    return;
  }

  pOdMText->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdMText );
  WriteEntityBase( hMText, pOdDb, pOdMText );
  OdPnt.x   = lcPropGetFloat( hMText, LC_PROP_MTEXT_X );  // Insertion point X
  OdPnt.y   = lcPropGetFloat( hMText, LC_PROP_MTEXT_Y );  // Insertion point Y
  OdPnt.z   = lcPropGetFloat( hMText, LC_PROP_ENT_Z );  // Insertion point Z
  pOdMText->setLocation( OdPnt );
  if (Height == 0.0){
    Height = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_HEIGHT );
  }
  if (Height <= 0.0){
    Height = 1.0;
  }
  if (WScale == 0.0){
    WScale = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_WSCALE );
  }
  pOdMText->setTextHeight( Height );
  pOdMText->setWidth( WrapWidth );
  pOdMText->setRotation( Angle );
  pOdMText->setLineSpacingStyle( OdDb::kExactly );  // kAtLeast 
  pOdMText->setLineSpacingFactor( LineSpace );
  switch( Align ){
    case LC_TA_LEFBOT:  pOdMText->setAttachment( OdDbMText::kBottomLeft ); break;
    case LC_TA_LEFCEN:  pOdMText->setAttachment( OdDbMText::kMiddleLeft ); break;
    case LC_TA_LEFTOP:  pOdMText->setAttachment( OdDbMText::kTopLeft ); break;
    case LC_TA_CENBOT:  pOdMText->setAttachment( OdDbMText::kBottomCenter ); break;
    case LC_TA_CENTER:  pOdMText->setAttachment( OdDbMText::kMiddleCenter ); break;
    case LC_TA_CENTOP:  pOdMText->setAttachment( OdDbMText::kTopCenter ); break;
    case LC_TA_RIGBOT:  pOdMText->setAttachment( OdDbMText::kBottomRight ); break;
    case LC_TA_RIGCEN:  pOdMText->setAttachment( OdDbMText::kMiddleRight ); break;
    case LC_TA_RIGTOP:  pOdMText->setAttachment( OdDbMText::kTopRight ); break;
  }
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdMText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
  if (TextLen > 0){
    szText = new WCHAR[TextLen+2];
    wcscpy( szText, lcPropGetStr( hMText, LC_PROP_MTEXT_STR ));
    pOdMText->setContents( szText );
    delete[] szText;
  }else{
    pOdMText->setContents( L"?" );
  }
}

//-----------------------------------------------
void Write_ArcText (HANDLE hArcText, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[256], szOdDbHandle[32];
  double Radius = lcPropGetFloat( hArcText, LC_PROP_ATEXT_RADIUS );
  double Height = lcPropGetFloat( hArcText, LC_PROP_ATEXT_H );
  double WScale = lcPropGetFloat( hArcText, LC_PROP_ATEXT_WSCALE );
  double Ang0   = lcPropGetFloat( hArcText, LC_PROP_ATEXT_ANGSTA );
  double Ang2   = lcPropGetFloat( hArcText, LC_PROP_ATEXT_ANGEND );
  BOOL   bCW    = lcPropGetBool( hArcText, LC_PROP_ATEXT_CW );
  int    nChars = lcPropGetInt( hArcText, LC_PROP_ATEXT_LEN );
  HANDLE hStyle = lcPropGetHandle( hArcText, LC_PROP_ATEXT_STYLE );
  BOOL   bHollow, bClosed;
  OdGePoint3d OdPnt;
  OdDbHandle  OdHandle;
  OdDbArcAlignedTextPtr pOdArcText;

  if (nChars > 250){
    return;
  }
  bHollow = lcPropGetBool( hStyle, LC_PROP_TSTYLE_HOLLOW );
  bClosed = lcPropGetBool( hStyle, LC_PROP_TSTYLE_CLOSED );
  if (bHollow){
    Write_Exploded( hArcText, pOdDb, pOdBlock );
    return;
  }
  if (!bClosed && g_Options.m_Exp_bTextAsPlines){
    Write_Exploded( hArcText, pOdDb, pOdBlock );
    return;
  }

  if (Height == 0.0){
    Height = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_HEIGHT );
  }
  if (WScale == 0.0){
    WScale = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_WSCALE );
  }
  pOdArcText = OdDbArcAlignedText::createObject();
  pOdArcText->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdArcText );
  WriteEntityBase( hArcText, pOdDb, pOdArcText );
  pOdArcText->setXScale( WScale );
//  Wsc = pOdArcText->xScale();
  pOdArcText->setTextSize( Height );
  OdPnt.x = lcPropGetFloat( hArcText, LC_PROP_ATEXT_X );
  OdPnt.y = lcPropGetFloat( hArcText, LC_PROP_ATEXT_Y );
  OdPnt.z = lcPropGetFloat( hArcText, LC_PROP_ENT_Z );
  pOdArcText->setCenter( OdPnt ); 
  if (bCW){
    pOdArcText->setRadius( Radius );
    pOdArcText->setTextDirection( kOutwardFromCenter );
    pOdArcText->setStartAngle( Ang2 );
    pOdArcText->setEndAngle( Ang0 );
  }else{
    pOdArcText->setRadius( Radius );
    pOdArcText->setTextDirection( kInwardToTheCenter );
    pOdArcText->setStartAngle( Ang0 );
    pOdArcText->setEndAngle( Ang2 );
  }
  szText[0] = 0;
  wcscpy( szText, lcPropGetStr( hArcText, LC_PROP_ATEXT_STRT ));
  pOdArcText->setTextString( szText );
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdArcText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
}


//-----------------------------------------------
void Write_TextWin (HANDLE hTextW, OdDbDatabasePtr pOdDb, OdDbBlockTableRecordPtr pOdBlock)
{
  WCHAR  szText[1024], szOdDbHandle[32];
  HANDLE hStyle = lcPropGetHandle( hTextW, LC_PROP_TEXTW_STYLE );
  double Height  = lcPropGetFloat( hTextW, LC_PROP_TEXTW_H ); 
  double WScale  = lcPropGetFloat( hTextW, LC_PROP_TEXTW_WSCALE ); 
  double Angle   = lcPropGetFloat( hTextW, LC_PROP_TEXTW_ANGLE ); 
  double Oblique = 0.0; //lcPropGetFloat( hTextW, LC_PROP_TEXTW_OBLIQUE ); 
  BOOL   bBack   = false; //lcPropGetBool( hTextW, LC_PROP_TEXTW_BACKWARD ); 
  BOOL   bUpdown = false; //lcPropGetBool( hTextW, LC_PROP_TEXTW_UPDOWN ); 
  int    Align   = lcPropGetInt( hTextW, LC_PROP_TEXTW_ALIGN ); 
  int    nChars  = lcPropGetInt( hTextW, LC_PROP_TEXTW_LEN );
  OdGePoint3d  pt0, pta;
  OdDbHandle   OdHandle;
  OdDbTextPtr  pOdText;

  if (nChars > 1020){
    return;
  }
/*
  if (lcPropGetBool( hTextW, LC_PROP_TEXTW_PIXH )){
    // unscalable text
    return;
  }
*/
  if (Height == 0.0){
    Height = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_HEIGHT );
  }
  if (WScale == 0.0){
    WScale = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_WSCALE );
  }
//  if (Oblique == 0.0){
//    Oblique = lcPropGetFloat( hStyle, LC_PROP_TSTYLE_OBLIQUE );
//  }

  pOdText = OdDbText::createObject();
  pOdText->setDatabaseDefaults( pOdDb );
  pOdBlock->appendOdDbEntity( pOdText );
  WriteEntityBase( hTextW, pOdDb, pOdText );
  pta.x   = lcPropGetFloat( hTextW, LC_PROP_TEXTW_X );  // Insertion point X
  pta.y   = lcPropGetFloat( hTextW, LC_PROP_TEXTW_Y );  // Insertion point Y
  pta.z   = lcPropGetFloat( hTextW, LC_PROP_ENT_Z );  // Insertion point Z
  szText[0] = 0;
  wcscpy( szText, lcPropGetStr( hTextW, LC_PROP_TEXTW_STRT ));
  pOdText->setTextString( szText );
//  pt0 = pta;
//  pOdText->setPosition( pt0 );
  pOdText->setAlignmentPoint( pta );
  pOdText->setHeight( Height );
  pOdText->setWidthFactor( WScale );
  pOdText->setOblique( Oblique );
  pOdText->setRotation( Angle );
  pOdText->mirrorInX( bBack? true : false );
  pOdText->mirrorInY( bUpdown? true : false );
  switch( Align ){
    case  LC_TA_LEFBOT:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextBase );
      pOdText->setPosition( pta );
      break;
    case  LC_TA_LEFCEN:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_LEFTOP:
      pOdText->setHorizontalMode( OdDb::kTextLeft );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_CENBOT:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_CENTER:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_CENTOP:
      pOdText->setHorizontalMode( OdDb::kTextCenter );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
    case  LC_TA_RIGBOT:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextBase );
      break;
    case  LC_TA_RIGCEN:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextVertMid );
      break;
    case  LC_TA_RIGTOP:
      pOdText->setHorizontalMode( OdDb::kTextRight );
      pOdText->setVerticalMode( OdDb::kTextTop );
      break;
  }
  wcscpy( szOdDbHandle, lcPropGetStr( hStyle, LC_PROP_TABLE_ODA_HANDLE ));
  if (szOdDbHandle[0]){
    OdHandle = OdDbHandle( szOdDbHandle );
    pOdText->setTextStyle( pOdDb->getOdDbObjectId( OdHandle ) );
  }
}

