/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2022, Oleg Kolbaskin
* All rights reserved.
*
* dllmain.cpp : Defines the entry point for the DLL application.
*********************************************************************/
#include "pch.h"

// Global variables
//-----------------------------------------------
HMODULE  g_hInst = 0;
WCHAR*   g_szMsgTitle = L"DWG/DXF plugin for LiteCAD";
COptions g_Options;
LC_FUN_PROGRESS g_pFunProgress = NULL;
HANDLE   g_hDrw = 0;   // used to pass with g_pFunProgress

BOOL Dg_Options (HWND hwParent, BOOL bSave);

#ifdef JJJJKKMNMNMNJHYTFD

#define ODRX_DECLARE_STATIC_MECHANICAL
#define ODRX_DEFINE_STATIC_MECHANICAL
//#define ODREADEX_STR L"OdReadEx"
//#define INIT_PROTOCOL_EXTENSION(theProtocolExtensions) ExProtocolExtension theProtocolExtensions;


/************************************************************************/
/* Define a module map for statically linked modules:                   */
/************************************************************************/

ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(ModelerModule);
ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(OdRecomputeDimBlockModule);
//ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(ExFieldEvaluatorModule);
//ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(TD_3DSolidHistoryModule);
//ODRX_DECLARE_STATIC_MECHANICAL;
ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(DbCryptModule);

ODRX_BEGIN_STATIC_MODULE_MAP()
  ODRX_DEFINE_STATIC_APPLICATION(OdModelerGeometryModuleName,   ModelerModule)
  ODRX_DEFINE_STATIC_APPLICATION(OdRecomputeDimBlockModuleName, OdRecomputeDimBlockModule)
//  ODRX_DEFINE_STATIC_APPMODULE(OdExFieldEvaluatorModuleName, ExFieldEvaluatorModule)
//  ODRX_DEFINE_STATIC_APPMODULE(Od3DSolidHistoryTxModuleName, TD_3DSolidHistoryModule)
//  ODRX_DEFINE_STATIC_MECHANICAL
  ODRX_DEFINE_STATIC_APPMODULE(DbCryptModuleName,            DbCryptModule)
ODRX_END_STATIC_MODULE_MAP()

#endif  // JJJJKKMNMNMNJHYTFD


//-----------------------------------------------
BOOL APIENTRY DllMain (HMODULE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
  g_hInst = hModule;
  switch( ul_reason_for_call ){
    case DLL_PROCESS_ATTACH:
    case DLL_THREAD_ATTACH:
    case DLL_THREAD_DETACH:
    case DLL_PROCESS_DETACH:
        break;
  }
  return TRUE;
}

// returns a number of interfaces
//-----------------------------------------------
int _stdcall LcPlug_CountIfaces ()
{
  return 4;  // read DWG/DXF, write DWG/DXF
}


// interface type
//-----------------------------------------------
int _stdcall LcPlug_GetIType (int index)
{
  switch( index ){
    case 0: return LC_PLUG_IMPDRW;   //  import a drawing DWG
    case 1: return LC_PLUG_IMPDRW;   //  import a drawing DXF
    case 2: return LC_PLUG_EXPDRW;   //  export a drawing DWG
    case 3: return LC_PLUG_EXPDRW;   //  export a drawing DXF
  }
  return 0;
}


// interface name, description and author
//-----------------------------------------------
void _stdcall LcPlug_GetIName (int Index, WCHAR* szIName, WCHAR* szIDesc, WCHAR* szAuthor)
{
  WCHAR* szAut = L"OKS";
  switch( Index ){
    case 0:
      wcscpy( szIName, L"Read AutoCAD drawing (DWG)" );
      wcscpy( szIDesc, L"Allows to read a drawing of AutoCAD DWG format" );
      wcscpy( szAuthor, szAut );
      break;
    case 1:
      wcscpy( szIName, L"Read AutoCAD drawing (DXF)" );
      wcscpy( szIDesc, L"Allows to read a drawing of AutoCAD DXF format" );
      wcscpy( szAuthor, szAut );
      break;
    case 2:
      wcscpy( szIName, L"Write AutoCAD drawing (DWG)" );
      wcscpy( szIDesc, L"Allows to write a drawing of AutoCAD DWG format" );
      wcscpy( szAuthor, szAut );
      break;
    case 3:
      wcscpy( szIName, L"Write AutoCAD drawing (DXF)" );
      wcscpy( szIDesc, L"Allows to write a drawing of AutoCAD DXF format" );
      wcscpy( szAuthor, szAut );
      break;
  }
}



// file extention for the specified interface
//-----------------------------------------------
void _stdcall LcPlug_GetFileExt (int Index, WCHAR* szExt, WCHAR* szInfo)
{
  switch( Index ){
    case 0:     // load AutoCAD DWG drawing
      wcscpy( szExt, L"dwg" );
      wcscpy( szInfo, L"AutoCAD Drawing (*.dwg)" );
      break;  
    case 1:     // load AutoCAD DXF drawing
      wcscpy( szExt, L"dxf" );
      wcscpy( szInfo, L"AutoCAD Drawing (*.dxf)" );
      break;  
    case 2:     // save AutoCAD DWG drawing
      wcscpy( szExt, L"dwg" );
      wcscpy( szInfo, L"AutoCAD Drawing (*.dwg)" );
      break;  
    case 3:     // save AutoCAD DXF drawing
      wcscpy( szExt, L"dxf" );
      wcscpy( szInfo, L"AutoCAD Drawing (*.dxf)" );
      break;  
  }
}


//-----------------------------------------------
int _stdcall LcPlug_GetPreview (LPCWSTR szFileName, BYTE* Buffer, int BufSize)
{
  int    nb=0;
  WCHAR  szExt[8], szBuf[256];
  OdStaticRxObject<MyServices> svcs;
  OdSmartPtr<OdBaseFileBuf> pFile;
  OdThumbnailImage Preview;
  OdBinaryData data;
  OdString msg;
  int PalSize=0;
  BITMAPFILEHEADER fh;
  BITMAPINFOHEADER* pih=NULL;
//  BYTE* Buffer = NULL;

  szBuf[0] = 0;
  okGetFileExt( szFileName, szExt );
  if (wcsicmp( szExt, L"dwg" )==0 || wcsicmp( szExt, L"dxf" )==0){
/*
    svcs.disableOutput(disableSvcsOutput);
    if (disableRecompute){
      svcs.setRecomputeDimBlocksRequired(false);
    }
*/
    try 
    {
      odInitialize( &svcs );
      pFile = OdRdFileBuf::createObject();
      pFile->open( szFileName, Oda::kShareDenyWrite, Oda::kFileRead, Oda::kOpenExisting );
      odDbGetPreviewBitmap( pFile, &Preview );
      if (Preview.hasBmp()){
//        Buffer = (BYTE*)::GlobalLock( hGlob );
        data = Preview.bmp;
        nb = data.physicalLength();
        if (nb < BufSize){
          pih = (BITMAPINFOHEADER*)data.getPtr();
          memcpy( Buffer+14, pih, nb );
          // create a header
          memset( &fh, 0, sizeof(fh) );
          fh.bfType = 0x4D42;   // BM
          PalSize = pih->biClrUsed * 4;
          // file size
          fh.bfSize = 14 + 40 + PalSize + pih->biSizeImage;
          // offset to bitmap bits
          fh.bfOffBits = 14 + 40 + PalSize;
          memcpy( Buffer, &fh, 14 );
          nb+=14;
        }else{
          nb = 0;
        }
//        ::GlobalUnlock( hGlob );
      }
      pFile->close();
    }
    catch (OdError &e)
    {
      msg = svcs.getErrorDescription(e.code());
      wcscpy( szBuf, (LPCWSTR)msg.c_str() );
      nb = 0;
    }
    catch (...)
    {
      wcscpy( szBuf, L"Unknown Error." );
      nb = 0;
    }
    odUninitialize();
  }
  return nb;  // number of bytes written to the buffer (hGlob)
}


//-----------------------------------------------
BOOL _stdcall LcPlug_Import (int Index, HANDLE hLcDb, LPCWSTR szFileName, void* pFunProgress, WCHAR* szError)
{
  if ((Index == 0) || (Index == 1)){  // Read DWG file,  Read DXF file
    g_pFunProgress = (LC_FUN_PROGRESS)pFunProgress;
    g_hDrw = hLcDb;
    if (ReadFile( szFileName, hLcDb, szError )){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
BOOL _stdcall LcPlug_Export (int Index, HANDLE hLcDb, LPCWSTR szFileName, const BYTE* ImgBuffer, int ImgSize, void* pFunProgress, WCHAR* szError)
{
  if ((Index == 2) || (Index == 3)){  // Write DWG file,  Write DXF file
    g_pFunProgress = (LC_FUN_PROGRESS)pFunProgress;
    g_hDrw = hLcDb;
    if (WriteFile( szFileName, ImgBuffer, ImgSize, hLcDb, szError )){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
void _stdcall LcPlug_LoadOptions (LPCWSTR szDir)
{
  WCHAR szFileName[256];
  wcscpy( szFileName, szDir );
  wcscat( szFileName, L"\\dwg_lcplug.cfg" );
  g_Options.Load( szFileName );
}

//-----------------------------------------------
void _stdcall LcPlug_DlgOptions (int Index, HWND hwParent, BOOL bSave)
{
  Dg_Options( hwParent, bSave );
}

//-----------------------------------------------
BOOL _stdcall LcPlug_GetOption (LPCWSTR szKey, WCHAR* szValue)
{
  if (_wcsicmp( szKey, L"EXP_FILEVER" ) == 0){
    _itow( g_Options.m_Exp_FileVer, szValue, 10 );
  }else
  if (_wcsicmp( szKey, L"EXP_DXFPREC" ) == 0){
    _itow( g_Options.m_Exp_DxfPrec, szValue, 10 );
  }else
  if (_wcsicmp( szKey, L"EXP_TEXTPLINES" ) == 0){ // export text with non-closed font as polylines
    _itow( g_Options.m_Exp_bTextAsPlines, szValue, 10 );
  }else
  if (_wcsicmp( szKey, L"IMP_DIMBLOCK" ) == 0){
    if (g_Options.m_Imp_bDimAsBlock){
      wcscpy( szValue, L"1" );
    }else{
      wcscpy( szValue, L"0" );
    }
  }else
  if (_wcsicmp( szKey, L"IMP_HATCHBLOCK" ) == 0){
    if (g_Options.GetImp_HatchAsBlock()){
      wcscpy( szValue, L"1" );
    }else{
      wcscpy( szValue, L"0" );
    }
  }else{
    szValue[0] = 0;
    return false;
  }
  return true;
}

//-----------------------------------------------
BOOL _stdcall LcPlug_SetOption (LPCWSTR szKey, LPCWSTR szValue, BOOL bSave)
{
  int ival;
  BOOL bRet = false;
  if (_wcsicmp( szKey, L"EXP_FILEVER" ) == 0){
    ival = _wtoi( szValue );
    if (ival == 13 ||   // Release 10 
        ival == 16 ||   // Release 11-12 
        ival == 19 ||   // Release 13 
        ival == 21 ||   // Release 14 
        ival == 23 ||   // R15 (2000) 
        ival == 25 ||   // R18 (2004) 
        ival == 27 ||   // R21 (2007)
        ival == 29)     // R24 (2010)
    {
      g_Options.m_Exp_FileVer = ival;
      bRet = true;
    }
  }else
  if (_wcsicmp( szKey, L"EXP_DXFPREC" ) == 0){
    ival = _wtoi( szValue );
    if (0 <= ival && ival <= 16){
      g_Options.m_Exp_DxfPrec = ival;
      bRet = true;
    }
  }else
  if (_wcsicmp( szKey, L"EXP_TEXTPLINES" ) == 0){ // export text with non-closed font as polylines
    ival = _wtoi( szValue );
    g_Options.m_Exp_bTextAsPlines = (ival==0)? false : true;
  }else
  if (_wcsicmp( szKey, L"IMP_DIMBLOCK" ) == 0){
    ival = _wtoi( szValue );
    g_Options.m_Imp_bDimAsBlock = ival?true:false;
    bRet = true;
  }else
  if (_wcsicmp( szKey, L"IMP_HATCHBLOCK" ) == 0){
    ival = _wtoi( szValue );
    g_Options.SetImp_HatchAsBlock( ival?true:false );
    bRet = true;
  }else
  if (_wcsicmp( szKey, L"IMP_PLINECIRCLE" ) == 0){
    ival = _wtoi( szValue );
    g_Options.m_Imp_bPlineToCircle = ival?true:false;
    bRet = true;
  }
  if (bRet && bSave){
    // save options in a file
    g_Options.Save();
  }
  return bRet;
}




//*****************************************************************************
COptions::COptions () 
{
  memset( m_szFileName, 0, sizeof(m_szFileName) );
  m_Exp_FileVer = OdDb::vAC21;  // 2007
  m_Exp_DxfPrec = 10;
  m_Exp_bTextAsPlines = false;
  m_Imp_bDimAsBlock = true;
  m_Imp_bHatchAsBlock = false;
  m_Imp_bPlineToCircle = false;
  m_bNoWarnings = true;
}

//-----------------------------------------------
void COptions::Load (LPCWSTR szFilename)
{
  COkConfig Cfg;
  int ival;

  wcscpy( m_szFileName, szFilename );
  if (Cfg.Read( m_szFileName )){
    ival = Cfg.GetValueInt( L"EXP_FILEVER", m_Exp_FileVer );
    if (ival == 13 ||   // OdDb::vAC10 Release 10 
        ival == 16 ||   // OdDb::vAC12 Release 11-12 
        ival == 19 ||   // OdDb::vAC13 Release 13 
        ival == 21 ||   // OdDb::vAC14 Release 14 
        ival == 23 ||   // OdDb::vAC15 R15 (2000-2002) 
        ival == 25 ||   // OdDb::vAC18 R18 (2004-2006) 
        ival == 27 ||   // OdDb::vAC21 R21 (2007-2009) release
        ival == 29 ||   // OdDb::vAC24  2010-2012 
        ival == 31 ||   // OdDb::vAC27  2013-2016
        ival == 33)     // OdDb::vAC32  2018
    {
      m_Exp_FileVer = ival;
    }
    ival = Cfg.GetValueInt( L"EXP_DXFPREC", m_Exp_DxfPrec );
    if (0 <= ival && ival <= 16){
      m_Exp_DxfPrec = ival;
    }
    ival = Cfg.GetValueInt( L"EXP_TEXTPLINES", 0 );
    m_Exp_bTextAsPlines = (ival==0)?false:true;
    m_Imp_bDimAsBlock = Cfg.GetValueBool( L"IMP_DIMBLOCK", m_Imp_bDimAsBlock );
    SetImp_HatchAsBlock( Cfg.GetValueBool( L"IMP_HATCHBLOCK", GetImp_HatchAsBlock() ));
    m_Imp_bPlineToCircle = Cfg.GetValueBool( L"IMP_PLINECIRCLE", m_Imp_bPlineToCircle );
    m_bNoWarnings = Cfg.GetValueBool( L"NOWARNINGS", m_bNoWarnings );
  }
}

//-----------------------------------------------
void COptions::Save ()
{
  FILE* df = _wfopen( m_szFileName, L"wt" );
  if (df){
    fprintf( df, "EXP_FILEVER=%d\n"    , m_Exp_FileVer ); 
    fprintf( df, "EXP_DXFPREC=%d\n"    , m_Exp_DxfPrec ); 
    fprintf( df, "EXP_TEXTPLINES=%d\n" , m_Exp_bTextAsPlines? 1 : 0 ); 
    fprintf( df, "IMP_DIMBLOCK=%d\n"   , m_Imp_bDimAsBlock? 1 : 0 );
    fprintf( df, "IMP_HATCHBLOCK=%d\n" , GetImp_HatchAsBlock()? 1 : 0 );
    fprintf( df, "IMP_PLINECIRCLE=%d\n", m_Imp_bPlineToCircle? 1 : 0 );
    fprintf( df, "NOWARNINGS=%d\n"     , m_bNoWarnings? 1 : 0 );

    fclose( df );
  }
}

//-----------------------------------------------
void COptions::SetImp_HatchAsBlock (bool b)
{
  m_Imp_bHatchAsBlock = b;
}



