// pch.h: This is a precompiled header file.
// Files listed below are compiled only once, improving build performance for future builds.
// This also affects IntelliSense performance, including code completion and many code browsing features.
// However, files listed here are ALL re-compiled if any one of them is updated between builds.
// Do not add files here that you will be updating frequently as this negates the performance advantage.

#ifndef PCH_H
#define PCH_H

#define WIN32_LEAN_AND_MEAN             // Exclude rarely-used stuff from Windows headers
  #define  WINVER        0x0500   // Windows 2000
  #define _WIN32_WINNT   0x0500   // Windows 2000
  #define _WIN32_WINDOWS 0x0500   // Windows Me
  #define _WIN32_IE      0x0500   // Internet Explorer 5.0
  #include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "LiteCAD.h"
#include "ListObj.h"
#include "Geom.h"

#include "OdaCommon.h"
#include "diagnostics.h"
#include "DbDatabase.h"
#include "DbEntity.h"
#include "DbDimAssoc.h"

#include "OdCharMapper.h"
#include "RxObjectImpl.h"

#include "ExSystemServices.h"
#include "ExHostAppServices.h"
//#include "ExProtocolExtension.h"
#include "OdFileBuf.h"
#include "RxDynamicModule.h"
#include "FdField.h"

#include "Db2LineAngularDimension.h"
#include "Db2dPolyline.h"
#include "Db3PointAngularDimension.h"
#include "Db3dPolyline.h"
#include "Db3dPolylineVertex.h"
#include "Db3dSolid.h"
#include "DbAlignedDimension.h"
#include "DbArc.h"
#include "DbArcAlignedText.h"
#include "DbArcDimension.h"
#include "DbAttribute.h"
#include "DbAttributeDefinition.h"
#include "DbBlockReference.h"
#include "DbBlockTable.h"
#include "DbBlockTableRecord.h"
#include "DbCircle.h"
#include "DbDiametricDimension.h"
#include "DbEllipse.h"
#include "DbEntity.h"
#include "DbFace.h"
#include "DbHatch.h"
#include "DbLayerTable.h"
#include "DbLayerTableRecord.h"
#include "DbLayout.h"
#include "DbLeader.h"
#include "DbMLeader.h"
#include "DbLine.h"
#include "DbLinetypeTable.h"
#include "DbLinetypeTableRecord.h"
#include "DbMInsertBlock.h"
#include "DbMText.h"
#include "DbMline.h"
#include "DbOrdinateDimension.h"
#include "DbPoint.h"
#include "DbPolyFaceMesh.h"
#include "DbPolyFaceMeshVertex.h"
#include "DbPolygonMesh.h"
#include "DbPolygonMeshVertex.h"
#include "DbPolyline.h"
#include "DbRadialDimension.h"
#include "DbRasterImage.h"
#include "DbRasterImageDef.h"
#include "DbRay.h"
#include "DbRotatedDimension.h"
#include "DbShape.h"
#include "DbSolid.h"
#include "DbSpline.h"
#include "DbTable.h"
#include "DbText.h"
#include "DbTextStyleTable.h"
#include "DbTextStyleTableRecord.h"
#include "DbTrace.h"
#include "DbViewport.h"
#include "DbViewportTable.h"
#include "DbViewportTableRecord.h"
#include "DbWipeout.h"
#include "DbXline.h"
#include "DbSurface.h"
#include "DbBody.h"
//#include "DbRegion.h"
#include "DbDictionary.h"
#include "DbMlineStyle.h"
#include "DbSortentsTable.h"
#include "ThumbnailImage.h"
#include "DbDimStyleTable.h"
#include "Ge\GeKnotVector.h"
#include "Ge\GeEllipArc2d.h"
#include "Ge\GeLinearEnt2d.h"
#include "Ge\GeCircArc2d.h"
#include "Ge\GeEllipArc2d.h"
#include "Ge\GeSplineEnt2d.h"
#include "Ge\GeNurbCurve2d.h"
#include "Gi\GiRasterWrappers.h"
#include "summinfo.h"
#include "Br\BrBrep.h"


class COptions {
  WCHAR m_szFileName[256];
  bool m_Imp_bHatchAsBlock;
public:
  int  m_Exp_FileVer;  
  int  m_Exp_DxfPrec;
  bool m_Exp_bTextAsPlines;    // export text with non-closed font as polylines
  bool m_Imp_bDimAsBlock;
  bool m_Imp_bPlineToCircle;
  bool m_bNoWarnings;
public:
    COptions ();

  void Load (LPCWSTR szFilename);
  void Save ();
  void SetImp_HatchAsBlock (bool b);
  bool GetImp_HatchAsBlock () const {return m_Imp_bHatchAsBlock;}
};

//-------------------------------------
class CConfigItem {
  WCHAR* m_szKey;
  WCHAR* m_szValue;
public:
    CConfigItem ();
    ~CConfigItem ();

  void    Set      (LPCWSTR szKey, LPCWSTR szValue);
  void    Get      (WCHAR* szKey, WCHAR* szValue) const;
  LPCWSTR Key      () const {return m_szKey;}
  LPCWSTR Value    () const {return m_szValue;}
  bool    GetValue (LPCWSTR szKey, WCHAR* szValue, UINT_PTR nMaxChars) const;
};


//-------------------------------------
class COkConfig {
  WCHAR        m_szFileName[256];
  CConfigItem* m_Item;
  UINT_PTR     m_nItems;    
public:
    COkConfig ();
    ~COkConfig ();

  void    SetFileName (LPCWSTR szFileName);
  LPCWSTR FileName    () const;

  void    Clear        ();
  bool    Read         (LPCWSTR szFileName=0, const char* szSignature=0);
  bool    GetValue     (LPCWSTR szKey, WCHAR* szValue, UINT_PTR nMaxChars) const;
  bool    GetValueBool (LPCWSTR szKey, bool DefVal) const;
  int     GetValueInt  (LPCWSTR szKey, int DefVal, int MinVal=0, int MaxVal=0) const;
  double  GetValueDbl  (LPCWSTR szKey, double DefVal, double MinVal=0.0, double MaxVal=0.0) const;
};
inline LPCWSTR COkConfig::FileName () const {return m_szFileName;}




typedef void (_stdcall* LC_FUN_PROGRESS) (int Mode, int IntVal, HANDLE hDrw);

bool ReadFile  (LPCWSTR szFileName, HANDLE hLcDb, WCHAR* szError);
bool WriteFile (LPCWSTR szFileName, const BYTE* ImgBuffer, int ImgSize, HANDLE hLcDb, WCHAR* szError);

void   SetObjHandle (const OdDbObjectId& objectId, HANDLE hObj, int IdProp);
HANDLE DrwGetObjectByHandle (const OdDbObjectId& objectId, HANDLE hLcDb, int ObjType);
void   WriteEntityBase (HANDLE hEnt, OdDbDatabasePtr pOdDb, OdDbEntityPtr pOdEnt, bool bSolid=false);
void   Read_EntBase (const OdDbEntityPtr pOdEnt, HANDLE hEnt, HANDLE hLcDb);
void   MakeOdColor (LPCWSTR szColor, OdCmColor* pOdColor);
void   MakeOdColor (int iColor, OdCmColor* pOdColor);
bool   OdColorToLcad (const OdCmColor& OdColor, WCHAR* szColor);
double GetAcadEllipseAngle (double Ang0, double Ratio);
int LWeight_DwgToLcd (OdDb::LineWeight LwDwg);
OdDb::LineWeight LWeight_LcdToDwg (int LwLcd);


// utility functions
LPCWSTR okGetFileName (LPCWSTR szFileName, bool bWithExtention=true);
LPCWSTR okGetFileExt (LPCWSTR szFileName, WCHAR* szExt);
LPCWSTR okGetFileDir (LPCWSTR szFileName, bool bIncludeBackslash=true);
INT_PTR okSearchFile (LPCWSTR szFileName, WCHAR* szFoundName=0, LPCWSTR szSearchDir=0);
void    okNumToStr (double Val, WCHAR* szBuf, UINT_PTR MaxDec=0, bool bTrimZero=true, bool bRemainPoint=false, UINT_PTR Separator='.', UINT_PTR TypeChar='f');
void    okDblToStr (double Val, WCHAR* szBuf, UINT_PTR MaxDec=5, bool bRemainPoint=true); 
void    okDelLastBlanks (WCHAR* szText); 



/************************************************************************/
/* Define a Custom Services class.                                      */
/*                                                                      */
/* Combines the platform dependent functionality of                     */
/* ExSystemServices and ExHostAppServices                               */ 
/************************************************************************/
class MyServices : public ExSystemServices, public ExHostAppServices
{
protected:
  ODRX_USING_HEAP_OPERATORS(ExSystemServices);

private:

#ifdef _MSC_VER

  /**********************************************************************/
  /* Define a console interface for password support.                   */
  /**********************************************************************/
  bool getPassword(const OdString& dwgName, bool /*isXref*/, OdPassword& password);

#endif

  virtual void OK_setLimit (int max);
  virtual void OK_meterProgress ();
  virtual void OK_warning (const OdString& str);
};



#endif //PCH_H
